/*
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * espda_ccf.c
 *
 *  Created on: Nov 10, 2015
 *      Author: Christophe Lovis
 */

/*
 * $Author: clovis $
 * $Date: 2015-11-10 14:48:13 $
 * $Revision: 1.16 $
 * $Name: $
 */


/*----------------------------------------------------------------------------
                                Includes
 ----------------------------------------------------------------------------*/

#include <espda_ccf.h>
#include <gsl/gsl_version.h>
/*----------------------------------------------------------------------------
                              Functions code
 ----------------------------------------------------------------------------*/



/*----------------------------------------------------------------------------*/
/**
 @brief     Compute the Cross-Correlation Function
 @param     ll                S2D wavelength matrix in barycentric frame
 @param     dll               S2D matrix of pixel sizes in wavelength units
 @param     flux              Input S2D spectrum (with blaze)
 @param     error             Input S2D error array
 @param     quality           Input S2D quality map
 @param     RV_table          Array of RV values on which to sample CCF (km/s)
 @param     mask              Line list for cross-correlation (wavelengths and weights)
 @param     mask_width        Width of mask holes in km/s
 @param     berv              Barycentric correction in km/s
 @param     bervmax           Maximum barycentric correction over the year (km/s)
 @param     CCF_flux_RE       CCF (returned)
 @param     CCF_error_RE      CCF error array (returned)
 @param     CCF_quality_RE    CCF quality map (returned)
 @return    CPL_ERROR_NONE if OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_compute_CCF(cpl_image *ll,
                                 cpl_image *dll,
                                 cpl_image *flux,
                                 cpl_image *error,
                                 cpl_image *blaze,
                                 cpl_image *quality,
                                 cpl_array *RV_table,
                                 cpl_table *mask,
                                 double mask_width,
                                 double berv,
                                 double bervmax,
                                 cpl_image **CCF_flux_RE,
                                 cpl_image **CCF_error_RE,
                                 cpl_image **CCF_quality_RE ) {
    
    int i, j, k, order, pir, imin, imax, index, index1, index2, index3, index_center;
    double c = 299792.458;
    double llmin, llmax, llcenter, llstart, llstop;
    double Sc, Nc, contrast, w;
    double *ccf_flux, *ccf_error;
    int *ccf_quality;
    int nx_s2d = cpl_image_get_size_x(flux);
    int ny_s2d = cpl_image_get_size_y(flux);
    int n_mask = cpl_table_get_nrow(mask);
    int nx_ccf = cpl_array_get_size(RV_table);
    
    cpl_image *ll2, *dll2;
    
    cpl_error_code my_error;
    
    ccf_flux = cpl_image_get_data_double(*CCF_flux_RE);
    ccf_error = cpl_image_get_data_double(*CCF_error_RE);
    ccf_quality = cpl_image_get_data_int(*CCF_quality_RE);
    
    espda_msg("Starting compute_CCF: mask_width = %f, berv = %f, bervmax = %f",
              mask_width, berv, bervmax);
    espda_ensure(flux == NULL, CPL_ERROR_NULL_INPUT, "flux input image is NULL");
    espda_ensure(ll == NULL, CPL_ERROR_NULL_INPUT, "ll input image is NULL");
    espda_ensure(dll == NULL, CPL_ERROR_NULL_INPUT, "dll input image is NULL");
    espda_ensure(error == NULL, CPL_ERROR_NULL_INPUT, "error input image is NULL");
    espda_ensure(blaze == NULL, CPL_ERROR_NULL_INPUT, "blaze input image is NULL");
    espda_ensure(quality == NULL, CPL_ERROR_NULL_INPUT, "quality input image is NULL");
    espda_ensure(RV_table == NULL, CPL_ERROR_NULL_INPUT, "RV_table input array is NULL");
    espda_ensure(mask == NULL, CPL_ERROR_NULL_INPUT, "input mask table is NULL");
    espda_ensure(n_mask == 0, CPL_ERROR_INCOMPATIBLE_INPUT, "input mask has no data");
    
    /*
     espda_msg("ASE DEBUG : ll  size_x=%lld size_y=%lld",
     cpl_image_get_size_x(ll),
     cpl_image_get_size_y(ll));
     espda_msg("ASE DEBUG : dll  size_x=%lld size_y=%lld",
     cpl_image_get_size_x(dll),
     cpl_image_get_size_y(dll));
     espda_msg("ASE DEBUG : flux  size_x=%lld size_y=%lld",
     cpl_image_get_size_x(flux),
     cpl_image_get_size_y(flux));
     espda_msg("ASE DEBUG : error size_x=%lld size_y=%lld",
     cpl_image_get_size_x(error),
     cpl_image_get_size_y(error));
     espda_msg("ASE DEBUG : blaze size_x=%lld size_y=%lld",
     cpl_image_get_size_x(blaze),
     cpl_image_get_size_y(blaze));
     espda_msg("ASE DEBUG : qual  size_x=%lld size_y=%lld",
     cpl_image_get_size_x(quality),
     cpl_image_get_size_y(quality));
     */
    
    dll2 = cpl_image_divide_scalar_create(dll,2.);
    ll2 = cpl_image_subtract_create(ll,dll2);
    
    for (order = 1; order <= ny_s2d; order++) {
        
        imin = 1; imax = nx_s2d;
        
        while ((imin < nx_s2d) && (cpl_image_get(quality,imin,order,&pir) != 0)) imin++;
        while ((imax > 1) && (cpl_image_get(quality,imax,order,&pir) != 0)) imax--;
        
        if (imin >= imax) continue;
        
        llmin = cpl_image_get(ll,imin+1,order,&pir)/(1.+berv/c)*
        (1.+bervmax/c)/(1.+cpl_array_get(RV_table,0,NULL)/c);
        llmax = cpl_image_get(ll,imax-1,order,&pir)/(1.+berv/c)*
        (1.-bervmax/c)/(1.+cpl_array_get(RV_table,nx_ccf-1,NULL)/c);
        
        //Sc = fmax(cpl_image_get_median_window(flux,imin,order,imax,order),0.);
        
        imin = 0; imax = n_mask-1;
        
        if (!cpl_table_has_column(mask, "lambda")) {
            espda_msg_warning("The input mask does not have the column 'lambda', exiting");
            return (CPL_ERROR_INCOMPATIBLE_INPUT);
        }
        
        if (!cpl_table_has_column(mask, "contrast")) {
            espda_msg_warning("The input mask does not have the column 'contrast', exiting");
            return (CPL_ERROR_INCOMPATIBLE_INPUT);
        }
        
        while ((imin < n_mask) &&
               (cpl_table_get(mask,"lambda",imin,NULL) < (llmin+0.5*mask_width/c*llmin))) {
            imin++;
        }
        
        while ((imax >= 0) &&
               (cpl_table_get(mask,"lambda",imax,NULL) > (llmax-0.5*mask_width/c*llmax))) {
            imax--;
        }
        
        for (i = imin; i <= imax; i++) {
            
            /*
             llstart = cpl_table_get(mask,"lambda",i,NULL)*(1.+cpl_array_get(RV_table,0,NULL)/c);
             llstop = cpl_table_get(mask,"lambda",i,NULL)*(1.+cpl_array_get(RV_table,nx_ccf-1,NULL)/c);
             index1 = 1;
             while (cpl_image_get(ll,index1,order,&pir) < llstart) index1++;
             index2 = index1;
             while (cpl_image_get(ll,index2,order,&pir) < llstop) index2++;
             Nc = cpl_image_get_median_window(error,index1,order,index2,order);
             */
            
            llcenter = cpl_table_get(mask,"lambda",i,NULL)*
            (1.+cpl_array_get(RV_table,nx_ccf/2,NULL)/c);
            
            index_center = 1;
            
            while (cpl_image_get(ll,index_center,order,&pir) < llcenter) index_center++;
            
            contrast = cpl_table_get(mask,"contrast",i,NULL);
            
            /*if (Nc > 0.) {
             //w = Sc/Nc/Nc*contrast*contrast;
             w = 1.0/Nc/Nc*contrast*contrast;
             } else {
             w = 0.;
             }*/
            
            w = contrast*contrast;
            
            for (j = 0; j < nx_ccf; j++) {
                
                llcenter = cpl_table_get(mask,"lambda",i,NULL)*
                (1.+cpl_array_get(RV_table,j,NULL)/c);
                llstart = llcenter - 0.5 * mask_width/c * llcenter;
                llstop = llcenter + 0.5 * mask_width/c  *llcenter;
                index1 = 1;
                while (cpl_image_get(ll2,index1,order,&pir) < llstart) index1++;
                
                index2 = index1;
                while (cpl_image_get(ll2,index2,order,&pir) < llcenter) index2++;
                
                index3 = index2;
                while (cpl_image_get(ll2,index3,order,&pir) < llstop) index3++;
                
                k = nx_ccf*(order-1)+j;
                
                for (index = index1; index < index3; index++) {
                    ccf_flux[k] += w*cpl_image_get(flux,index,order,&pir)/
                    cpl_image_get(blaze,index,order,&pir)*
                    cpl_image_get(blaze,index_center,order,&pir);
                }
                
                ccf_flux[k] += w * cpl_image_get(flux,index1-1,order,&pir)*
                (cpl_image_get(ll2,index1,order,&pir)-llstart)/
                cpl_image_get(dll,index1-1,order,&pir)/
                cpl_image_get(blaze,index1-1,order,&pir)*
                cpl_image_get(blaze,index_center,order,&pir);
                
                ccf_flux[k] -= w * cpl_image_get(flux,index3-1,order,&pir)*
                (cpl_image_get(ll2,index3,order,&pir)-llstop)/
                cpl_image_get(dll,index3-1,order,&pir)/
                cpl_image_get(blaze,index3-1,order,&pir)*
                cpl_image_get(blaze,index_center,order,&pir);
                
                ccf_error[k] += w * w * cpl_image_get(error,index2-1,order,&pir)*
                cpl_image_get(error,index2-1,order,&pir);
                
                ccf_quality[k] = ccf_quality[k] | (int)(cpl_image_get(quality,index2-1,order,&pir));
            }
        }
    }
    
    for (j = 0; j < nx_ccf; j++) {
        for (order = 0; order < ny_s2d; order++) {
            ccf_flux[nx_ccf*ny_s2d+j] += ccf_flux[nx_ccf*order+j];
            ccf_error[nx_ccf*ny_s2d+j] += ccf_error[nx_ccf*order+j];
            ccf_quality[nx_ccf*ny_s2d+j] += ccf_quality[nx_ccf*order+j];
        }
    }
    
    my_error = cpl_image_power(*CCF_error_RE,0.5);
    
    cpl_image_delete(ll2);
    cpl_image_delete(dll2);
    
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_fit_ccf(cpl_array *RV_table,
                             cpl_image *CCF_flux,
                             cpl_image *CCF_error,
                             cpl_image *CCF_quality,
                             double *asym_RE,
                             double *asym_err_RE,
                             double *bispan_RE,
                             double *bispan_err_RE,
                             cpl_table **residuals_RE,
                             espda_ngauss_result *g_res_RE) {
    
    cpl_error_code my_error = CPL_ERROR_NONE;
    
    cpl_image *CCF_last_order_flux;
    cpl_image *CCF_last_order_error;
    cpl_image *CCF_last_order_quality;
    
    double *CCF_last_order_data_flux;
    double *CCF_last_order_data_error;
    double *CCF_slope;
    int *CCF_last_order_data_quality;
    
    double *x0;
    double CCF_weight_sum=0.0;
    
    int CCF_x,CCF_y;
    
    int RV_length = cpl_array_get_size(RV_table);
    size_t min_index;
    
    espda_ngauss_data *g_data = NULL;
    
    g_data = (espda_ngauss_data *)cpl_malloc(sizeof(espda_ngauss_data));
    g_data->n = RV_length;
    g_data->m = 1;
    g_data->delta = 0.0;
    g_data->x = (double *)cpl_calloc (RV_length , sizeof(double));
    g_data->y = (double *)cpl_calloc (RV_length , sizeof(double));
    g_data->err = (double *)cpl_calloc (RV_length , sizeof(double));
    g_data->fit = (double *)cpl_calloc (RV_length , sizeof(double));
    
    x0 = (double *)cpl_malloc(sizeof(double));
    
    double sigma = 0.;
    double *model = (double *)cpl_calloc (RV_length , sizeof(double));
    
    *residuals_RE = cpl_table_new(RV_length);
    
    my_error = cpl_table_new_column(*residuals_RE,
                                    "rv",
                                    CPL_TYPE_DOUBLE);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "cpl_table_new_column failed: %s",
                 cpl_error_get_message_default(my_error));
    my_error = cpl_table_set_column_unit(*residuals_RE,
                                         "rv",
                                         "[km/s]");
    
    my_error = cpl_table_new_column(*residuals_RE,
                                    "residuals",
                                    CPL_TYPE_DOUBLE);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "cpl_table_new_column failed: %s",
                 cpl_error_get_message_default(my_error));
    my_error = cpl_table_set_column_unit(*residuals_RE,
                                         "residuals",
                                         "[km/s]");
    
    my_error = cpl_table_new_column(*residuals_RE,
                                    "error",
                                    CPL_TYPE_DOUBLE);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "cpl_table_new_column failed: %s",
                 cpl_error_get_message_default(my_error));
    my_error = cpl_table_set_column_unit(*residuals_RE,
                                         "error",
                                         "[km/s]");
    
    my_error = cpl_table_new_column(*residuals_RE,
                                    "quality",
                                    CPL_TYPE_INT);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "cpl_table_new_column failed: %s",
                 cpl_error_get_message_default(my_error));
    my_error = cpl_table_set_column_unit(*residuals_RE,
                                         "quality",
                                         "");
    
    g_res_RE->k = (double *)cpl_malloc (sizeof(double));
    g_res_RE->sig_k = (double *)cpl_malloc (sizeof(double));
    g_res_RE->x0 = (double *)cpl_malloc (sizeof(double));
    g_res_RE->sig_x0 = (double *)cpl_malloc (sizeof(double));
    g_res_RE->fwhm = (double *)cpl_malloc (sizeof(double));
    g_res_RE->sig_fwhm = (double *)cpl_malloc (sizeof(double));
    
    espda_msg("Fitting the CCF...");
    
    if (CCF_flux != NULL && CCF_error != NULL) {
        
        CCF_x = cpl_image_get_size_x(CCF_flux);
        CCF_y = cpl_image_get_size_y(CCF_flux);
        
        CCF_last_order_flux = cpl_image_extract(CCF_flux,
                                                1,CCF_y,
                                                CCF_x,CCF_y);
        
        CCF_last_order_data_flux =
        cpl_image_get_data_double(CCF_last_order_flux);
        
        CCF_last_order_error = cpl_image_extract(CCF_error,
                                                 1,CCF_y,
                                                 CCF_x,CCF_y);
        
        CCF_last_order_data_error =
        cpl_image_get_data_double(CCF_last_order_error);
        
        CCF_last_order_quality = cpl_image_extract(CCF_quality,
                                                   1,CCF_y,
                                                   CCF_x,CCF_y);
        
        CCF_last_order_data_quality =
        cpl_image_get_data_int(CCF_last_order_quality);
        
        for(int i = 0; i < RV_length; i++) {
            g_data->x[i]    = cpl_array_get(RV_table,i,NULL);
            g_data->y[i]    = CCF_last_order_data_flux[i];
            //g_data->err[i]  = CCF_last_order_data_error[i];
            g_data->err[i]  = 1.0;       // Fitting CCF with uniform weights
        }
        
        min_index = gsl_stats_min_index(g_data->y,1,RV_length);
        
        x0[0] = g_data->x[(int)min_index];
        
        my_error = espda_fit_Ngauss(g_data, x0, g_res_RE, 0, 0);
        
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "espda_fit_Ngauss failed for fibre %s",
                     cpl_error_get_message_default(my_error));
        
        g_res_RE->fwhm[0] = fabs(g_res_RE->fwhm[0]);    // Forcing positivity of FWHM
        
        /* CCF slope calculation */
        
        CCF_slope = (double *)cpl_calloc (RV_length , sizeof(double));
        
        for(int i = 1; i < RV_length-1; i++) {
            CCF_slope[i] =
            (CCF_last_order_data_flux[i+1]-CCF_last_order_data_flux[i-1])/
            (cpl_array_get(RV_table,i+1,NULL)-cpl_array_get(RV_table,i-1,NULL));
        }
        
        /* CCF error calculation */
        
        for(int i = 1; i < RV_length-1; i++) {
            CCF_weight_sum += pow(CCF_slope[i]/CCF_last_order_data_error[i],2.0);
        }
        
        cpl_free(CCF_slope);
        
        g_res_RE->sig_x0[0] = 1.0/sqrt(CCF_weight_sum);
        g_res_RE->sig_fwhm[0] = 1.0/sqrt(CCF_weight_sum)*2.0;
        if (g_res_RE->fwhm[0] != 0.) {
            g_res_RE->sig_k[0] = fabs(1.0/sqrt(CCF_weight_sum)*2.0/g_res_RE->fwhm[0]*g_res_RE->k[0]);
        } else {
            g_res_RE->sig_k[0] = 0.;
        }
        
        /* model and residuals table calculation */
        
        if ( g_res_RE->fwhm[0] != 0. && !isnan(g_res_RE->fwhm[0]) && !isinf(g_res_RE->fwhm[0]) ) {
            sigma = g_res_RE->fwhm[0]/2./sqrt(2.*log(2.));
            for (int i = 0; i < RV_length; i++)
                model[i] = g_res_RE->c + g_res_RE->k[0]*exp(-1.*(g_data->x[i]-g_res_RE->x0[0])*
                                                            (g_data->x[i]-g_res_RE->x0[0])/2./sigma/sigma);
        } else {
            for (int i = 0; i < RV_length; i++)
                model[i] = 0.;
        }
        
        for (int i = 0; i < RV_length; i++) {
            cpl_table_set(*residuals_RE,
                          "rv",i,g_data->x[i]);
            
            cpl_table_set(*residuals_RE,
                          "residuals",i,g_data->y[i]-model[i]);
            
            cpl_table_set(*residuals_RE,
                          "error",i,CCF_last_order_data_error[i]);
            
            cpl_table_set(*residuals_RE,
                          "quality",i,CCF_last_order_data_quality[i]);
        }
        
        espda_msg("CCF fitted, RV = %.10f", g_res_RE->x0[0]);
        
        /* Compute the CCF flux asymmetry */
        
        espda_msg("Computing the flux assymetry...");
        
        int il = 0, im = 0, ir = 0;
        double dx = (g_data->x[1] - g_data->x[0])/2.;
        while ((il < (RV_length-1)) && (g_data->x[il] < (g_res_RE->x0[0] - 1.5*g_res_RE->fwhm[0] + dx))) il++;
        while ((im < (RV_length-1)) && (g_data->x[im] < (g_res_RE->x0[0] + dx))) im++;
        while ((ir < (RV_length-1)) && (g_data->x[ir] < (g_res_RE->x0[0] + 1.5*g_res_RE->fwhm[0] + dx))) ir++;
        
        double suml = 0., sumr = 0., sumlerr = 0., sumrerr = 0.;
        if ( (il > 0) && (im > il) && (ir > im) && (ir < (RV_length-1)) ) {
            
            for (int i = il; i < (im-1); i++) {
                suml += g_data->y[i]-model[i];
                sumlerr += CCF_last_order_data_error[i]*CCF_last_order_data_error[i];
            }
            suml += (g_data->y[il-1]-model[il-1])*(g_data->x[il-1] + dx - g_res_RE->x0[0] + 1.5*g_res_RE->fwhm[0])/2./dx;
            suml += (g_data->y[im-1]-model[im-1])*(g_res_RE->x0[0] - g_data->x[im-1] + dx)/2./dx;
            
            for (int i = im; i < (ir-1); i++) {
                sumr += g_data->y[i]-model[i];
                sumrerr += CCF_last_order_data_error[i]*CCF_last_order_data_error[i];
            }
            sumr += (g_data->y[im-1]-model[im-1])*(g_data->x[im-1] + dx - g_res_RE->x0[0])/2./dx;
            sumr += (g_data->y[ir-1]-model[ir-1])*(g_res_RE->x0[0] + 1.5*g_res_RE->fwhm[0] - g_data->x[ir-1] + dx)/2./dx;
        }
        
        *asym_RE = (sumr - suml) * 2. * dx / g_res_RE->c;
        *asym_err_RE = sqrt(sumlerr + sumrerr) * 2. * dx / g_res_RE->c;
        
        espda_msg("Flux assymetry computed: %.10f (err: %.10f)",
                  *asym_RE, *asym_err_RE);
        
        /* Compute the CCF bisector span */
        
        espda_msg("Computing the bissector span...");
        
        espda_ngauss_result *g_res_bis = (espda_ngauss_result *)cpl_malloc (sizeof(espda_ngauss_result));
        g_res_bis->k = (double *)cpl_malloc (sizeof(double));
        g_res_bis->sig_k = (double *)cpl_malloc (sizeof(double));
        g_res_bis->x0 = (double *)cpl_malloc (sizeof(double));
        g_res_bis->sig_x0 = (double *)cpl_malloc (sizeof(double));
        g_res_bis->fwhm = (double *)cpl_malloc (sizeof(double));
        g_res_bis->sig_fwhm = (double *)cpl_malloc (sizeof(double));
        int i;
        
        // Higher errors are set on the bottom of the ccf fit, to fit only the top part
        for (int j = 0; j < RV_length; j++) {
            double arg_err = (model[j] - g_res_RE->c)/g_res_RE->k[0];
            double sigma_err = 0.4 / 2.3548;
            double gauss_err = 1.0 + 50.0 * exp(-(arg_err-1.0)*(arg_err-1.0)/2.0/sigma_err/sigma_err);
            //espda_msg("RV[%d] = %f\tmodel[%d] = %f\t(model-c)/k0 = %.10f\tgauss: %f",
            //          j, cpl_array_get(RV_table,j,NULL), j, model[j], arg_err, gauss_err);
            
            g_data->err[j] = gauss_err;
        }
        
        my_error = espda_fit_Ngauss(g_data, x0, g_res_bis, 0, 0);
        
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "espda_fit_Ngauss failed for fibre %s",
                     cpl_error_get_message_default(my_error));
        
        double RV_top = g_res_bis->x0[0];
        
        // Higher errors are set on the top of the ccf fit, to fit only the bottom part
        // The continuum (< 0.01 of contrast) is forced to 1.0
        for (int j = 0; j < RV_length; j++) {
            double arg_err = (model[j] - g_res_RE->c)/g_res_RE->k[0];
            double sigma_err = 0.25 / 2.3548;
            double gauss_err = 1.0 + 50.0 * exp(-(arg_err-0.25)*(arg_err-0.25)/2.0/sigma_err/sigma_err);
            
            g_data->err[j] = gauss_err;
            
            if (arg_err < 0.01) {
                g_data->err[j] = 1.0;
            }
            //espda_msg("RV[%d] = %f\tmodel[%d] = %f\t(model-c)/k0 = %.10f\tgauss: %f\tg_data_err = %.10f",
            //          j, cpl_array_get(RV_table,j,NULL), j, model[j], arg_err, gauss_err, g_data->err[j]);
        }
        
        my_error = espda_fit_Ngauss(g_data, x0, g_res_bis, 0, 0);
        
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "espda_fit_Ngauss failed for fibre %s",
                     cpl_error_get_message_default(my_error));
        
        double RV_bottom = g_res_bis->x0[0];
        
        if (!isnan(RV_top) && !isinf(RV_top) && !isnan(RV_bottom) && !isinf(RV_bottom)) {
            *bispan_RE = RV_top - RV_bottom;
            *bispan_err_RE = 2.0 * g_res_RE->sig_x0[0];
        } else {
            *bispan_RE = 0.;
            *bispan_err_RE = 0.;
        }
        
        espda_msg("Bissector span computed: %.10f (err: %.10f)",
                  *bispan_RE, *bispan_err_RE);
        
        cpl_free(g_res_bis->k);
        cpl_free(g_res_bis->sig_k);
        cpl_free(g_res_bis->x0);
        cpl_free(g_res_bis->sig_x0);
        cpl_free(g_res_bis->fwhm);
        cpl_free(g_res_bis->sig_fwhm);
        cpl_free(g_res_bis);
        
    }
    
    cpl_free(x0);
    cpl_free(model);
    cpl_free(g_data->x);
    cpl_free(g_data->y);
    cpl_free(g_data->err);
    cpl_free(g_data->fit);
    cpl_free(g_data);
    
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

const char *espda_get_mask_name(const char *spec_type_id,
                                const cpl_table *mask_lut) {
    
    int mask_lut_length = cpl_table_get_nrow(mask_lut);
    
    for(int i = 0 ; i < mask_lut_length ; i++) {
        
        //espda_msg("get_mask_name: %d SPEC_TYPE: %s, MASK_TABLE: %s", i,
        //          cpl_table_get_string(mask_lut, "SPEC_TYPE", i),
        //          cpl_table_get_string(mask_lut, "MASK_TABLE", i));
        
        if (strcmp(cpl_table_get_string(mask_lut, "SPEC_TYPE", i),
                   spec_type_id) == 0) {
            return (cpl_table_get_string(mask_lut, "MASK_TABLE", i));
        }
        
    }
    return "NOT_FOUND";
}

/*----------------------------------------------------------------------------*/
/**
 @brief     compute the Radial Velocity (RV) table
 @param     RV_range        RV range
 @param     RV_step         RV step
 @param     RV_center       RV center
 @param     RV_table_RE     RV table calculated (returned).
 @return    CPL_ERROR_NONE if OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_compute_rv_table(double RV_range,
                                      double RV_step,
                                      double RV_center,
                                      cpl_array **RV_table_RE) {
    
    int RV_end = round(RV_range/RV_step);
    int RV_ini = - RV_end;
    double RV = 0.0;
    
    cpl_size length = RV_end - RV_ini +1;
    cpl_size array_index = 0;
    
    *RV_table_RE = cpl_array_new(length,CPL_TYPE_DOUBLE);
    
    for (int i = RV_ini; i <= RV_end; i++) {
        RV = RV_center + i*RV_step;
        cpl_array_set_double(*RV_table_RE,array_index,RV);
        array_index++;
    }
    
    return cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

static int n_gauss_f (const gsl_vector *v, void *data, gsl_vector *f) {
    
    double *x = ((espda_ngauss_data *) data)->x;
    double *y = ((espda_ngauss_data *) data)->y;
    double *err = ((espda_ngauss_data *) data)->err;
    double *fit = ((espda_ngauss_data *) data)->fit;
    int n = ((espda_ngauss_data *) data)->n;
    int m = ((espda_ngauss_data *) data)->m;
    int i, j;
    
    double c = gsl_vector_get(v, 0);
    double *k = (double *)cpl_calloc(m, sizeof(double));
    double *x0 = (double *)cpl_calloc(m, sizeof(double));
    double *fwhm = (double *)cpl_calloc(m, sizeof(double));
    double *sigma = (double *)cpl_calloc(m, sizeof(double));
    
    for (j = 0; j < m; j++) {
        
        k[j] = gsl_vector_get(v, j+1);
        x0[j] = gsl_vector_get(v, m+j+1);
        fwhm[j] = gsl_vector_get(v, 2*m+j+1);
        sigma[j] = fwhm[j]/2/sqrt(2*log(2));
    }
    
    for (i = 0; i < n; i++) {
        
        fit[i] = c;
        
        for (j = 0; j < m; j++)
            fit[i] += k[j]*exp(-(x[i]-x0[j])*(x[i]-x0[j])/2/sigma[j]/sigma[j]);
        
        gsl_vector_set (f, i, (fit[i]-y[i])/err[i]);
    }
    
    cpl_free(k);
    cpl_free(x0);
    cpl_free(fwhm);
    cpl_free(sigma);
    
    return GSL_SUCCESS;
}

/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

static int n_gauss_df (const gsl_vector *v, void *data, gsl_matrix *J) {
    
    double *x = ((espda_ngauss_data *) data)->x;
    //double *y = ((espda_ngauss_data *) data)->y;
    double *err = ((espda_ngauss_data *) data)->err;
    //double *fit = ((espda_ngauss_data *) data)->fit;
    int n = ((espda_ngauss_data *) data)->n;
    int m = ((espda_ngauss_data *) data)->m;
    int i, j;
    
    //double c = gsl_vector_get(v, 0);
    double *k = (double *)cpl_calloc(m, sizeof(double));
    double *x0 = (double *)cpl_calloc(m, sizeof(double));
    double *fwhm = (double *)cpl_calloc(m, sizeof(double));
    double *sigma = (double *)cpl_calloc(m, sizeof(double));
    double *e = (double *)cpl_calloc(m, sizeof(double));
    
    for (j = 0; j < m; j++) {
        
        k[j] = gsl_vector_get(v, j+1);
        x0[j] = gsl_vector_get(v, m+j+1);
        fwhm[j] = gsl_vector_get(v, 2*m+j+1);
        sigma[j] = fwhm[j]/2/sqrt(2*log(2));
    }
    
    for (i = 0; i < n; i++) {
        
        gsl_matrix_set (J, i, 0, 1/err[i]);
        
        for (j = 0; j < m; j++) {
            
            e[j] = exp(-(x[i]-x0[j])*(x[i]-x0[j])/2/sigma[j]/sigma[j]);
            gsl_matrix_set (J, i, j+1, e[j]/err[i]);
            gsl_matrix_set (J, i, m+j+1,
                            k[j]/err[i]*e[j]*(x[i]-x0[j])/sigma[j]/sigma[j]);
            gsl_matrix_set (J, i, 2*m+j+1,
                            k[j]/err[i]*e[j]*(x[i]-x0[j])*(x[i]-x0[j])/sigma[j]/sigma[j]/sigma[j]/2/sqrt(2*log(2)));
        }
    }
    
    cpl_free(k);
    cpl_free(x0);
    cpl_free(fwhm);
    cpl_free(sigma);
    cpl_free(e);
    
    return GSL_SUCCESS;
}

/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

static int n_gauss_fdf (const gsl_vector *v, void *data,
                        gsl_vector *f, gsl_matrix *J) {
    
    n_gauss_f (v, data, f);
    n_gauss_df (v, data, J);
    
    return GSL_SUCCESS;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Fit n-gauss on the data
 @param data        data to be fit
 @param x0          first guess = lines peaks
 @param x0_RE       gaussian peaks returned
 @param sig_x0_RE   error on gaussian peaks
 @param spectrum_type       1 - emisison, 0 - absorption
 @param print_flag  1 - print, 0 - don't
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_fit_Ngauss (espda_ngauss_data *data,
                                 double *x0init,
                                 espda_ngauss_result *res,
                                 int spectrum_type,
                                 int print_flag) {
    
    
    const gsl_multifit_fdfsolver_type *T;
    gsl_multifit_fdfsolver *s;
    gsl_multifit_function_fdf f;
    int status;
    int i, j, iter = 0;
    int npar = GAUSS_NPAR;
    int vector_size = data->m * (npar-1) + 1;
    int n = data->n;
    int m = data->m;
    double c, *x0, *k, *fwhm;
    double sig_c, *sig_x0, *sig_k, *sig_fwhm;
    int *ind;
#if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
    gsl_matrix *J;
#endif
    
    gsl_vector *v = gsl_vector_alloc(vector_size);
    gsl_matrix *covar = gsl_matrix_alloc(vector_size, vector_size);
    
    x0 = (double *)cpl_calloc(m, sizeof(double));
    k = (double *)cpl_calloc(m, sizeof(double));
    ind = (int *)cpl_calloc(m, sizeof(int));
    fwhm = (double *)cpl_calloc(m, sizeof(double));
    sig_x0 = (double *)cpl_calloc(m, sizeof(double));
    sig_k = (double *)cpl_calloc(m, sizeof(double));
    sig_fwhm = (double *)cpl_calloc(m, sizeof(double));
    
    /* NOTE: The data->x and x0init vectors are used as provided, without subtracting the mean or first value */
    /* If needed (working with large x values), this should be taken care of outside of this function */
    
    /* Initialize the fit parameters properly */
    
    c = data->y[0];
    for (i = 0; i < n; i++) {
        if (spectrum_type == 1) { // emission spectrum
            if (data->y[i] < c) {
                c = data->y[i];
            }
        } else {    // absorption spectrum
            if (data->y[i] > c) {
                c = data->y[i];
            }
        }
    }
    gsl_vector_set(v, 0, c);
    
    for (j = 0; j < data->m; j++) {
        fwhm[j] = (data->x[n-1]-data->x[0])/3.0/data->m;
        x0[j] = x0init[j];
        ind[j] = (x0[j] - data->x[0])/(data->x[n-1] - data->x[0])*n;
        k[j] = data->y[ind[j]] - c;
        gsl_vector_set(v, j+1, k[j]);
        gsl_vector_set(v, m+j+1, x0[j]);
        gsl_vector_set(v, 2*m+j+1, fwhm[j]);
    }
    
    if (print_flag) {
        for (i = 0; i < data->m; i++) {
            espda_msg("NGAUSS: x0[%d] = %lf", i, x0[i]);
        }
    }
    
    f.f = &n_gauss_f;
    f.df = &n_gauss_df;
    f.fdf = &n_gauss_fdf;
    f.n = n;
    f.p = vector_size;
    f.params = data;
    
    T = gsl_multifit_fdfsolver_lmsder;
    s = gsl_multifit_fdfsolver_alloc(T, n, m*(npar-1)+1);
    gsl_multifit_fdfsolver_set(s, &f, v);
    
    //print_state (iter, s);
    
    do
    {
        iter++;
        status = gsl_multifit_fdfsolver_iterate(s);
        
        //printf ("status = %s\n", gsl_strerror (status));
        
        //print_state (iter, s);
        
        if (status)
            break;
        
        status = gsl_multifit_test_delta(s->dx, s->x, 0.0, 1e-7);
    }
    while (status == GSL_CONTINUE && iter < 1000);
    
#if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
    J = gsl_matrix_alloc(n, vector_size);
    gsl_multifit_fdfsolver_jac(s, J);
    gsl_multifit_covar (J, 0.0, covar);
    gsl_matrix_free (J);
#else
    gsl_multifit_covar (s->J, 0.0, covar);
#endif
    
    
    c = gsl_vector_get(s->x, 0);
    sig_c = sqrt(gsl_matrix_get(covar,0,0));
    for (j = 0; j < m; j++) {
        k[j] = gsl_vector_get(s->x, j+1);
        sig_k[j] = sqrt(gsl_matrix_get(covar,j+1,j+1));
        x0[j] = gsl_vector_get(s->x, m+j+1);
        sig_x0[j] = sqrt(gsl_matrix_get(covar, m+j+1, m+j+1));
        fwhm[j] = gsl_vector_get(s->x, 2*m+j+1);
        sig_fwhm[j] = sqrt(gsl_matrix_get(covar, 2*m+j+1, 2*m+j+1));
    }
    
    if (print_flag) {
        for (i = 0; i < data->m; i++) {
            espda_msg("NGAUSS after fit: x0[%d] = %lf", i, x0[i]);
            espda_msg("NGAUSS after fit: sig_x0[%d] = %lf", i, sig_x0[i]);
            espda_msg("NGAUSS after fit: fwhm[%d] = %lf", i, fwhm[i]);
            espda_msg("NGAUSS after fit: sig_fwhm[%d] = %lf", i, sig_fwhm[i]);
            espda_msg("NGAUSS after fit: k[%d] = %lf", i, k[i]);
            espda_msg("NGAUSS after fit: sig_k[%d] = %lf", i, sig_k[i]);
        }
    }
    
    res->n = n;
    res->m = m;
    res->c = c;
    res->sig_c = sig_c;
    for (j = 0; j < m; j++) {
        res->k[j] = k[j];
        res->sig_k[j] = sig_k[j];
        res->x0[j] = x0[j];
        res->sig_x0[j] = sig_x0[j];
        res->fwhm[j] = fwhm[j];
        res->sig_fwhm[j] = sig_fwhm[j];
    }
    
    gsl_multifit_fdfsolver_free(s);
    gsl_matrix_free(covar);
    gsl_vector_free(v);
    
    cpl_free(k);
    cpl_free(ind);
    cpl_free(fwhm);
    cpl_free(x0);
    cpl_free(sig_k);
    cpl_free(sig_x0);
    cpl_free(sig_fwhm);
    
    return cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/
/**
 @brief     polynomial fit (wrap for the gsl functions)
 @param         data            data to be fitted
 @param         deg             polynomial degree
 @param[out]    fit_RE          fit result
 @param[out]    coeffs_RE       coeffs of the fit
 @param[out]    coeffs_err_RE   error on coeffs of the fit
 @param[out]    chisq_RE        CHI2 of the fit
 @param         print_flag      1 - print, 0 - don't
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_fit_poly(espda_poly_data *data,
                              int deg,
                              double *fit_RE,
                              double *coeffs_RE,
                              double *coeffs_err_RE,
                              double *chisq_RE,
                              int print_flag) {

    int i, j, n;
    double err, chisq;
    gsl_vector *x, *y, *w, *p;
    gsl_matrix *X, *covar;
    gsl_multifit_linear_workspace *work;
    
    n = data->n;
    x = gsl_vector_alloc (n);
    y = gsl_vector_alloc (n);
    w = gsl_vector_alloc (n);
    p = gsl_vector_alloc (deg);
    X = gsl_matrix_alloc (n, deg);
    covar = gsl_matrix_alloc (deg, deg);
    
    if (print_flag) {
        espda_msg("n = %d, deg = %d", n, deg);
    }
    
    for (i = 0; i < n; i++) {
        gsl_vector_set (x, i, data->x[i]);
        gsl_vector_set (y, i, data->y[i]);
        err = data->err[i];
        gsl_vector_set (w, i, 1.0/err/err);
        
        for (j = 0; j < deg; j++)
            gsl_matrix_set (X, i, j, gsl_pow_int(gsl_vector_get(x, i), j));
    }
    
    if (print_flag) {
        espda_msg("vectors set");
    }
    
    work = gsl_multifit_linear_alloc (n, deg);
    gsl_multifit_wlinear (X, w, y, p, covar, &chisq, work);
    gsl_multifit_linear_free (work);
    
    if (print_flag) {
        espda_msg("Fit done");
    }
    
    for (i = 0; i < n; i++) {
        fit_RE[i] = 0.;
        for (j = 0; j < deg; j++)
            fit_RE[i] += gsl_matrix_get (X, i, j) * gsl_vector_get (p, j);
    }
    
    chisq = chisq/(n-deg);
    
    for (j = 0; j < deg; j++) {
        coeffs_RE[j] = gsl_vector_get (p, j);
        coeffs_err_RE[j] = sqrt(gsl_matrix_get (covar, j, j));
    }
    
    *chisq_RE = chisq;
    
    gsl_vector_free(x);
    gsl_vector_free(y);
    gsl_vector_free(w);
    gsl_vector_free(p);
    gsl_matrix_free(X);
    gsl_matrix_free(covar);
    
    return cpl_error_get_code();
}



/*----------------------------------------------------------------------------*/
/**
 @brief     Remove cosmics using the method of 4 bins histogram of flux
 @param     flux        flux to be cleaned
 @param     err         error on the flux
 @param     flux_RE     cleaned flux (returned)
 @param     cosmics_RE  nb of removed cosmics per order (returned)
 @return    CPL_ERROR_NONE if OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_remove_cosmics_via_histogram(cpl_image *flux,
                                                  cpl_image *err,
                                                  cpl_image *flux_RE,
                                                  int *cosmics_RE) {
    
    cpl_error_code my_error = CPL_ERROR_NONE;
    int order, pixel, nx, ny, index;
    cpl_image *flux_order = NULL;
    cpl_image *flux_order_corr = NULL;
    cpl_image *err_order = NULL;
    double *flux_order_data = NULL;
    double *flux_corr_data = cpl_image_get_data_double(flux_RE);
    double *flux_order_corr_data = NULL;
    double *err_order_data = NULL;
    double min_flux, max_flux, bin, detect_limit, order_median;
    int histogram[4];
    int *low_cosmics_nb, *high_cosmics_nb;
    int total_cosmics_nb = 0;
    
    nx = cpl_image_get_size_x(flux);
    ny = cpl_image_get_size_y(flux);
    //espda_msg("Image size: %dx%d", nx, ny);
    low_cosmics_nb = (int *)cpl_calloc(ny, sizeof(int));
    high_cosmics_nb = (int *)cpl_calloc(ny, sizeof(int));
    
    index = 0;
    for (order = 1; order <= ny; order++) {
        //espda_msg("Cosmics correction for order %d", order);
        flux_order = cpl_image_extract(flux, 1, order, nx, order);
        err_order = cpl_image_extract(err, 1, order, nx, order);
        flux_order_data = cpl_image_get_data_double(flux_order);
        err_order_data = cpl_image_get_data_double(err_order);
        flux_order_corr_data = (double *)cpl_calloc(nx, sizeof(double));
        
        // construct histogram
        //espda_msg("Constructing histogram for low cosmics order %d", order);
        for (int i = 0; i < 4; i++) {
            histogram[i] = 0;
        }
        min_flux = cpl_image_get_min(flux_order);
        max_flux = cpl_image_get_max(flux_order);
        bin = (max_flux - min_flux) / 4;
        for (pixel = 0; pixel < nx; pixel++) {
            if (flux_order_data[pixel] < (min_flux + bin)) {
                histogram[0]++;
            } else {
                if (flux_order_data[pixel] < (min_flux + 2*bin)) {
                    histogram[1]++;
                } else {
                    if (flux_order_data[pixel] < (min_flux + 3*bin)) {
                        histogram[2]++;
                    } else {
                        histogram[3]++;
                    }
                }
            }
        }
        //espda_msg("Low cosmics histogram: %d %d %d %d",
        //          histogram[0], histogram[1], histogram[2], histogram[3]);
        //espda_msg("Low cosmics histogram limits: min = %f %f %f %f %f max = %f",
        //          min_flux, min_flux+bin, min_flux+2*bin,
        //          min_flux+3*bin, min_flux+4*bin, max_flux);
        for (pixel = 0; pixel < nx; pixel++) {
            if (order == -1) {
                if (flux_order_data[pixel] < (min_flux+bin)) {
                    espda_msg("data[%d] = %f", pixel, flux_order_data[pixel]);
                } else {
                    if (flux_order_data[pixel] < (min_flux+2*bin)) {
                        espda_msg("data[%d] = %f", pixel, flux_order_data[pixel]);
                    }
                }
            }
        }
        // Histogram constructed, eliminate low cosmics
        //espda_msg("Eliminating low cosmics");
        detect_limit = min_flux;
        if ((histogram[1] == 0) && (histogram[0] > 0) &&
            ((histogram[2] + histogram[3]) > histogram[0])) {
            low_cosmics_nb[order-1] += histogram[0];
            detect_limit = min_flux + 2*bin;
        }
        for (pixel = 0; pixel < nx; pixel++) {
            if (flux_order_data[pixel] < detect_limit) {
                //flux_order_corr_data[pixel] = min_flux + 3*bin;
                flux_corr_data[index] = cpl_image_get_median(flux_order);
                //espda_msg("Low cosmic correction order %d, pixel %d", order, pixel);
            } else {
                flux_order_corr_data[pixel] = flux_order_data[pixel];
            }
        }
        flux_order_corr = cpl_image_wrap_double(nx, 1, flux_order_corr_data);
        
        // Eliminate the negative cosmics, which were not removed yet
        order_median = cpl_image_get_median(flux_order_corr);
        cpl_image_unwrap(flux_order_corr);
        for (pixel = 0; pixel < nx; pixel++) {
            if (flux_order_corr_data[pixel] < -10.0 * err_order_data[pixel]) {
                flux_order_corr_data[pixel] = order_median;
            }
        }
        flux_order_corr = cpl_image_wrap_double(nx, 1, flux_order_corr_data);
        
        // reconstruct the histogram
        //espda_msg("Constructing histogram for high cosmics order %d", order);
        for (int i = 0; i < 4; i++) {
            histogram[i] = 0;
        }
        min_flux = cpl_image_get_min(flux_order_corr);
        max_flux = cpl_image_get_max(flux_order_corr);
        bin = (max_flux - min_flux) / 4;
        for (pixel = 0; pixel < nx; pixel++) {
            if (flux_order_corr_data[pixel] < (min_flux + bin)) {
                histogram[0]++;
            } else {
                if (flux_order_corr_data[pixel] < (min_flux + 2*bin)) {
                    histogram[1]++;
                } else {
                    if (flux_order_corr_data[pixel] < (min_flux + 3*bin)) {
                        histogram[2]++;
                    } else {
                        histogram[3]++;
                    }
                }
            }
        }
        //espda_msg("High cosmics histogram: %d %d %d %d",
        //          histogram[0], histogram[1], histogram[2], histogram[3]);
        //espda_msg("High cosmics histogram limits: min = %f %f %f %f %f max = %f",
        //          min_flux, min_flux+bin, min_flux+2*bin,
        //          min_flux+3*bin, min_flux+4*bin, max_flux);
        // Histogram constructed, eliminate high cosmics
        //espda_msg("Eliminating high cosmics");
        for (pixel = 0; pixel < nx; pixel++) {
            if (order == -1) {
                if (flux_order_data[pixel] > (min_flux+3*bin)) {
                    espda_msg("data[%d] = %f", pixel, flux_order_data[pixel]);
                } else {
                    if (flux_order_data[pixel] > (min_flux+2*bin)) {
                        espda_msg("data[%d] = %f", pixel, flux_order_data[pixel]);
                    }
                }
            }
        }
        detect_limit = max_flux;
        if ((histogram[2] == 0) && (histogram[3] > 0) &&
            ((histogram[0] + histogram[1]) > histogram[3])) {
            high_cosmics_nb[order-1] += histogram[3];
            detect_limit = min_flux + 2*bin;
        }
        for (pixel = 0; pixel < nx; pixel++) {
            if (flux_order_corr_data[pixel] > detect_limit) {
                //flux_corr_data[index] = min_flux + 0.5 * bin;
                flux_corr_data[index] = cpl_image_get_median(flux_order_corr);
                //espda_msg("High cosmic correction order %d, pixel %d: %f",
                //          order, pixel, flux_corr_data[index]);
            } else {
                flux_corr_data[index] = flux_order_corr_data[pixel];
            }
            index++;
        }
        
        cpl_image_unwrap(flux_order_corr);
        cpl_free(flux_order_corr_data);
        cpl_image_delete(flux_order);
        cpl_image_delete(err_order);
    }
    
    total_cosmics_nb = 0;
    for (order = 0; order < ny; order++) {
        if ((low_cosmics_nb[order] != 0) || (high_cosmics_nb[order] != 0)) {
            //espda_msg("Removed cosmics for order %d: low = %d, high = %d",
            //          order+1, low_cosmics_nb[order], high_cosmics_nb[order]);
            total_cosmics_nb += low_cosmics_nb[order] + high_cosmics_nb[order];
            cosmics_RE[order] = low_cosmics_nb[order] + high_cosmics_nb[order];
        }
    }
    espda_msg("Total nb of cosmics removed: %d", total_cosmics_nb);
    cpl_free(low_cosmics_nb);
    cpl_free(high_cosmics_nb);
    
#if 0
    my_error = cpl_image_save(flux_RE, "ESPRESSO_FLUX_COSMICS_REMOVED.fits",
                              CPL_TYPE_DOUBLE, NULL, CPL_IO_CREATE);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_ccf_remove_cosmics failed to save image ESPRESSO_FLUX_COSMICS_REMOVED.fits: %s",
                 cpl_error_get_message_default(my_error));
    espda_msg("Image ESPRESSO_FLUX_COSMICS_REMOVED.fits saved");
#endif
    
    return cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/
/**
 @brief
 @param
 @param
 @param
 @param
 @return   CPL_ERROR_NONE iff OK
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_correct_flux(cpl_image *s2d_blaze,
                                  cpl_image *wave_matrix,
                                  cpl_table *flux_template,
                                  cpl_table *mask_table,
                                  char *spec_type,
                                  cpl_image *s2d_blaze_flux_corr_RE,
                                  double *flux_corr_RE) {
    
    cpl_error_code my_error = CPL_ERROR_NONE;
    
    int i, j, k, imin, fit_nb;
    double flux_sum, corr_min, corr_max, corr_model;
    int order_nb = cpl_image_get_size_y(s2d_blaze);
    int order_len = cpl_image_get_size_x(s2d_blaze);
    int flux_len = 0;
    int poly_deg = 8;
    
    double *flux_ref = NULL;
    double *s2d_blaze_data = cpl_image_get_data_double(s2d_blaze);
    double *wave_data = cpl_image_get_data_double(wave_matrix);
    double *s2d_blaze_corr_data = cpl_image_get_data_double(s2d_blaze_flux_corr_RE);
    double *flux = (double *) cpl_calloc(order_nb, sizeof(double));
    double *llc = (double *) cpl_calloc(order_nb, sizeof(double));
    double *corr = (double *) cpl_calloc(order_nb, sizeof(double));
    double chisq = 0.0;
    
    for (i = 0; i < order_nb; i++) {
        for (j = 0; j < order_len; j++) {
            s2d_blaze_corr_data[i*order_len+j] = s2d_blaze_data[i*order_len+j];
        }
        flux_corr_RE[i] = 0.;
    }
    
    if ((strcmp(spec_type, "NONE") == 0) ||
        (strcmp(spec_type, "none") == 0)) {
        
        espda_msg("Flux correction not performed: switched off by the user");
        
    } else {
        
        espda_msg("Performing flux correction for spectral type: %s", spec_type);
        
        flux_len = cpl_table_get_nrow(flux_template);
        espda_ensure(order_nb != flux_len, CPL_ERROR_INCOMPATIBLE_INPUT,
                     "Flux template length (%d) doesn't correspond to the number of orders in s2d - %d",
                     flux_len, order_nb);
        
        if (cpl_table_has_column(flux_template, spec_type)) {
            
            for (i = 0; i < order_nb; i++) {
                llc[i] = wave_data[i*order_len+order_len/2];
                flux_corr_RE[i] = 1.0;
            }
            
            imin = 0;
            while ((imin < order_nb-3) &&
                   (llc[imin+1] < cpl_table_get(mask_table,"lambda",0,NULL)) &&
                   (llc[imin+2] < cpl_table_get(mask_table,"lambda",0,NULL))) imin++;
            fit_nb = order_nb - imin;
            poly_deg = (int) round( (double) poly_deg * (double) fit_nb / (double) order_nb);
            espda_msg("Discarding %i orders and using polynomial of degree %i", imin, poly_deg);
            
            flux_ref = cpl_table_get_data_double(flux_template, spec_type);
            flux_sum = 0.;
            corr_min = 1000.;
            corr_max = -1000.;
            
            for (i = imin; i < order_nb; i++) {
                flux[i] = 0.;
                for (j = (order_len/4); j < (3*order_len/4); j++) {
                    flux[i] += s2d_blaze_data[i*order_len+j];
                }
                flux[i] = flux[i]/flux_ref[i];
                flux_sum += flux[i];
            }
            
            espda_poly_data *p_data = (espda_poly_data *) cpl_malloc(sizeof(espda_poly_data));
            p_data->x = (double *) cpl_calloc(fit_nb, sizeof(double));
            p_data->y = (double *) cpl_calloc(fit_nb, sizeof(double));
            p_data->err = (double *) cpl_calloc(fit_nb, sizeof(double));
            p_data->n = fit_nb;
            double *fit = (double *) cpl_calloc(fit_nb, sizeof(double));
            double *coeffs = (double *) cpl_calloc(poly_deg, sizeof(double));
            double *coeffs_err = (double *) cpl_calloc(poly_deg, sizeof(double));
            
            for (i = imin; i < order_nb; i++) {
                corr[i] = flux[i]/flux_sum*fit_nb;
                p_data->x[i-imin] = llc[i];
                p_data->y[i-imin] = corr[i];
                p_data->err[i-imin] = 1.0;
            }
            
            my_error = espda_fit_poly(p_data, poly_deg, fit, coeffs, coeffs_err, &chisq, 0);
            espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                         "espda_fit_poly failed: %s",
                         cpl_error_get_message_default(my_error));
            
            for (i = imin; i < order_nb; i++) {
                if (fit[i-imin] < corr_min) corr_min = fit[i-imin];
                if (fit[i-imin] > corr_max) corr_max = fit[i-imin];
                flux_corr_RE[i] = fit[i-imin];
            }
            
            if ((corr_min > 0.25) && (corr_max < 3.0)) {
                
                for (i = imin; i < order_nb; i++) {
                    for (j = 0; j < order_len; j++) {
                        corr_model = 0.;
                        for (k = 0; k < poly_deg; k++) {
                            corr_model += coeffs[k]*pow(wave_data[i*order_len+j],k);
                        }
                        s2d_blaze_corr_data[i*order_len+j] = s2d_blaze_data[i*order_len+j]/corr_model;
                    }
                }
                espda_msg("Flux correction performed with min/max values: %f/%f", corr_min, corr_max);
                
            } else {
                
                espda_msg("Flux correction not performed: min/max values out of range (%f/%f)",
                          corr_min, corr_max);
            }
            
            cpl_free(p_data->x);
            cpl_free(p_data->y);
            cpl_free(p_data->err);
            cpl_free(p_data);
            cpl_free(fit);
            cpl_free(coeffs);
            cpl_free(coeffs_err);
            
        } else {
            
            espda_msg_warning(ANSI_COLOR_RED"Flux correction not performed: no flux template available for spectral type %s"ANSI_COLOR_RESET, spec_type);
            
        }
    }
    
    cpl_free(flux);
    cpl_free(llc);
    cpl_free(corr);
    
    return cpl_error_get_code();
}



/*---------------------------------------------------------------------------*/
/**
 @brief        Save image with the complete FITS header and all the extensions
 @param     all_frames      recipe input frames
 @param        parameters      recipe parameters
 @param     used_frames     frames used for this product
 @param     recipe          recipe name
 @param     keywords        main header keywords
 @param     filename        product filename
 @param     images_to_save  list of images to be saved
 @return    CPL_ERROR_NONE iff OK
 
 */
/*---------------------------------------------------------------------------*/

cpl_error_code espda_dfs_save_data_err_qual(cpl_frameset *all_frames,
                                            const cpl_parameterlist *parameters,
                                            const cpl_frameset *used_frames,
                                            const char *recipe,
                                            cpl_propertylist *keywords,
                                            const char *filename,
                                            cpl_image **images_to_save) {

    int i;
    cpl_error_code my_error = CPL_ERROR_NONE;
    cpl_type image_type = CPL_TYPE_FLOAT;
    cpl_propertylist *keywords_ext = NULL;
    
    my_error = cpl_dfs_save_propertylist(all_frames, NULL, parameters,
                                         used_frames, NULL, recipe,
                                         keywords, NULL,
                                         PACKAGE "/" PACKAGE_VERSION,
                                         filename);

    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Error saving primary header: %s",
                 cpl_error_get_message());
    
    keywords_ext = cpl_propertylist_new();
    my_error = cpl_propertylist_append_string(keywords_ext,
                                              "EXTNAME", "SCIDATA");
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Error appending EXTNAME to extension header: %s",
                 cpl_error_get_message());
    
    for (i = 0; i < 3; i++) {
        if (i > 1) {
            image_type = CPL_TYPE_INT;
        }
        
        if (i == 1) {
            my_error = cpl_propertylist_update_string(keywords_ext,
                                                      "EXTNAME", "ERRDATA");
            espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                         "Error updating EXTNAME in extension header: %s",
                         cpl_error_get_message());
        }
        
        if (i == 2) {
            my_error = cpl_propertylist_update_string(keywords_ext,
                                                      "EXTNAME", "QUALDATA");
            espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                         "Error updating EXTNAME in extension header: %s",
                         cpl_error_get_message());
        }
        
        my_error = cpl_image_save(images_to_save[i],
                                  filename, image_type,
                                  keywords_ext, CPL_IO_EXTEND);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Error saving extension %d: %s",
                     i, cpl_error_get_message());
    }

    return cpl_error_get_code();

}



/*---------------------------------------------------------------------------*/
/**
 @brief    Check the sci_red QC
 @param    CCD_geom         CCD geometry
 @param    exposure_time    exposure time
 @param    inst_config      instrument configuration structure
 @param    qc_kws           structure holdin quality control, keys
 @param    keywords_RE      fits header to hold QC
 */
/*---------------------------------------------------------------------------*/

cpl_error_code espda_compu_radvel_QC(cpl_array *RV_table,
                                     double RV_step,
                                     int orders_nb,
                                     espda_ngauss_result *g_res,
                                     const char *ccf_mask_id,
                                     int *cosmics_nb_2,
                                     double *flux_correction,
                                     double asym,
                                     double asym_err,
                                     double bispan,
                                     double bispan_err,
                                     cpl_propertylist **keywords_RE) {

    cpl_error_code my_error = CPL_ERROR_NONE;
    
    int i;
    
    char *new_keyword = NULL;
    
    int flux_corr_QC = 1;
    int global_QC = 1;
    
    espda_msg("Saving QC KWs");
    
    cpl_propertylist_update_double(*keywords_RE,
                                   "ESO RV START",
                                   cpl_array_get_double(RV_table, 0, NULL));
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Update keyword ESO RV START to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));

    cpl_propertylist_update_double(*keywords_RE,
                                   "ESO RV STEP", RV_step);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Update keyword ESO RV STEP to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));

    espda_msg("Saving cosmics 2 KWs");
    for (i = 1; i <= orders_nb; i++) {
        new_keyword = (char *)cpl_malloc(32*sizeof(char));
        sprintf(new_keyword, "ESO QC ORDER%d COSMIC NB2", i);
        my_error = cpl_propertylist_update_int(*keywords_RE,
                                               new_keyword,
                                               cosmics_nb_2[i-1]);
        
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_ORDER_COSMICS_NB2 to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        cpl_free(new_keyword);
    }
    
    /* Flux correction KWs */
    
    espda_msg("Saving flux corr KWs");
    for (i = 1; i <= orders_nb; i++) {
        new_keyword = (char *)cpl_malloc(32*sizeof(char));
        sprintf(new_keyword, "ESO QC ORDER%d FLUX CORR", i);
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  new_keyword,
                                                  flux_correction[i-1]);
        
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_ORDER_FLUX_CORR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        cpl_free(new_keyword);
    }
    
    cpl_vector *flux_corr_vector = cpl_vector_wrap(orders_nb, flux_correction);
    double flux_correction_min = cpl_vector_get_min(flux_corr_vector);
    double flux_correction_max = cpl_vector_get_max(flux_corr_vector);
    cpl_vector_unwrap(flux_corr_vector);
    
    my_error = cpl_propertylist_update_double(*keywords_RE,
                                              "ESO QC FLUX CORR MIN",
                                              flux_correction_min);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Update keyword ESO QC FLUX CORR MIN to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));

    my_error = cpl_propertylist_update_double(*keywords_RE,
                                              "ESO QC FLUX CORR MAX",
                                              flux_correction_max);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Update keyword ESO QC FLUX CORR MAX to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));
    
    if ((flux_correction_min <= 0.25) || (flux_correction_max >= 3.0)) {
        flux_corr_QC = 0;
        /* CLO: global_QC not set to zero when flux correction fails or is not applied */
    }
        
    my_error = cpl_propertylist_update_int(*keywords_RE,
                                           "ESO QC FLUX CORR CHECK",
                                           flux_corr_QC);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Update keyword ESO QC FLUX CORR CHECK to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));
    
    espda_msg("Saving CCF KWs");
    /* ASE : verification needed when CCF fibre B is not calculated */
    if (g_res != NULL) {
        //espda_msg("Saving RV KWs");
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF RV",
                                                  g_res->x0[0]);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_RV to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        if ( g_res->sig_x0[0] != 0. && !isnan(g_res->sig_x0[0]) && !isinf(g_res->sig_x0[0])) {
            my_error = cpl_propertylist_update_double(*keywords_RE,
                                                      "ESO QC CCF RV ERROR",
                                                      g_res->sig_x0[0]);
        } else {
            my_error = cpl_propertylist_update_double(*keywords_RE,
                                                      "ESO QC CCF RV ERROR",
                                                      0.0);
        }
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_RV_ERROR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF FWHM",
                                                  g_res->fwhm[0]);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_FWHM to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        if ( g_res->sig_fwhm[0] != 0. && !isnan(g_res->sig_fwhm[0]) && !isinf(g_res->sig_fwhm[0])) {
            my_error = cpl_propertylist_update_double(*keywords_RE,
                                                      "ESO QC CCF FWHM ERROR",
                                                      g_res->sig_fwhm[0]);
        } else {
            my_error = cpl_propertylist_update_double(*keywords_RE,
                                                      "ESO QC CCF FWHM ERROR",
                                                      0.0);
        }
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_FWHM_ERROR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        double contrast;
        
        if (g_res->c != 0. && !isnan(g_res->c)) {
            contrast = (fabs(g_res->k[0]/g_res->c))*100.0;
        } else {
            espda_msg("WARNING : division by zero. g->res =%f",g_res->c);
            contrast = 0.0;
        }
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF CONTRAST",
                                                  contrast);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_CONTRAST to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        double contrast_error;
        
        if (g_res->c != 0. && !isnan(g_res->c)) {
            contrast_error = (fabs(g_res->sig_k[0]/g_res->c))*100.0;
        } else {
            espda_msg("WARNING : division by zero. g->res =%f",g_res->c);
            contrast_error = 0.0;
        }
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF CONTRAST ERROR",
                                                  contrast_error);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_CONTRAST_ERROR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF CONTINUUM",
                                                  g_res->c);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_CONTINUUM to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_string(*keywords_RE,
                                                  "ESO QC CCF MASK",
                                                  ccf_mask_id);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_MASK to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF FLUX ASYMMETRY",
                                                  asym);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_ASYMMETRY to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF FLUX ASYMMETRY ERROR",
                                                  asym_err);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_ASYMMETRY_ERROR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF BIS SPAN",
                                                  bispan);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_BIS_SPAN to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
        
        my_error = cpl_propertylist_update_double(*keywords_RE,
                                                  "ESO QC CCF BIS SPAN ERROR",
                                                  bispan_err);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Update keyword QC_CCF_BIS_SPAN_ERROR to the propertylist failed: %s",
                     cpl_error_get_message_default(my_error));
    }
    
    my_error = cpl_propertylist_append_int(*keywords_RE,
                                           "ESO QC COMPU RADVEL CHECK",
                                           global_QC);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Add keyword QC_COMPU_RADVEL_CHECK_KW to the propertylist failed: %s",
                 cpl_error_get_message_default(my_error));
    
    return cpl_error_get_code();
}



