/* $Id: espda_dfs.c 161574 2014-05-23 11:59:12Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2014-05-23 13:59:12 +0200 (Fri, 23 May 2014) $
 * $Revision: 161574 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_dfs.h"

int v_level;
int cycle;
char syst_id1 = 'a' - 1, syst_id2 = 'a';
char *qual;
char *grp_mode = "NOOVER";
char *grp_flag, *cycle_flag;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup   espda_dfs  DFS related functions
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Find the CPL_DFS_PRO_CATG value.
  @param    plist  The property list.
  @return   CPL_DFS_PRO_CATG value.

  This function finds the CPL_DFS_PRO_CATG value from a property list.
 */
/*----------------------------------------------------------------------------*/

const char *espda_dfs_catg(const cpl_propertylist *plist) {
    espda_util_error(plist == NULL, CPL_ERROR_NULL_INPUT, NULL);
    const char *value = cpl_propertylist_get_string(plist, CPL_DFS_PRO_CATG);
    cpl_ensure(value != NULL, cpl_error_get_code(), NULL);
    return value;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Group frames from a frame set.
  @param    set  A frame set.
  @return   CPL_ERROR_NONE iff OK.

  This function scans a frame set, checks the tags of the frames, and divides 
  them by group (RAW, CALIB) and type (IMAGE, TABLE).
*/
/*----------------------------------------------------------------------------*/

cpl_error_code espda_dfs_groups(cpl_frameset *set) {
    espda_util_error(set == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_errorstate prestate = cpl_errorstate_get();
    //extern int v_level;    
    cpl_frameset_iterator *iter  = cpl_frameset_iterator_new(set);
    cpl_frame             *frame = NULL;
    int                    i     = 0;
    for (frame = cpl_frameset_iterator_get(iter); frame != NULL;
        frame = cpl_frameset_iterator_get(iter), i++) {
        const char * tag = cpl_frame_get_tag(frame);
        if (tag == NULL) {
            cpl_msg_warning(cpl_func, "Frame %d has no tag.", i);
        } 

        // Calibrations
        else if (strstr(tag, "CALIB") != NULL || 
            strstr(tag, "FUNC") != NULL || strstr(tag, "ION") != NULL || 
            strstr(tag, "MAP") != NULL || strstr(tag, "PROF")) {
            cpl_frame_set_group(frame, CPL_FRAME_GROUP_CALIB);
       }

        // Reduced 1D images
        else if (strstr(tag, "1D") != NULL) {
            cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);
            cpl_frame_set_type(frame, CPL_FRAME_TYPE_ANY);            
        } 

        // Reduced 2D images
        else if (strstr(tag, "2D") != NULL || 
            strstr(tag, "MATRIX") != NULL) {
            cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);
            cpl_frame_set_type (frame, CPL_FRAME_TYPE_IMAGE);
        }

        // Reduced tables
        else if (strstr(tag, "SPEC") != NULL ||
            strstr(tag, "MASK") != NULL || strstr(tag, "LINE") != NULL ||
            strstr(tag, "TEFF") != NULL || strstr(tag, "FEH") != NULL) {
            cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);
            cpl_frame_set_type (frame, CPL_FRAME_TYPE_TABLE);
        } 

        // Unclassified
        else {
            cpl_frame_set_group(frame, CPL_FRAME_GROUP_NONE);
        }
        cpl_frameset_iterator_advance(iter, 1);
    }
    cpl_frameset_iterator_delete(iter);
    if (!cpl_errorstate_is_equal(prestate)) {
        cpl_frameset_delete(set);
        cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Unable to "
            "group frames: empty frame set.");
    }

    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Get the pipeline copyright and license.
  @return  The copyright and license string.

  This function returns a pointer to the statically allocated license string.
  This string should not be modified using the returned pointer.
 */
/*----------------------------------------------------------------------------*/

const char *espda_dfs_license(void)

{
    const char *espda_license = 
        "This file is part of the ESPDA Instrument Pipeline\n"
        "Copyright (C) 2002,2003 European Southern Observatory\n"
        "\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software\n"
        "Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, \n"
        "MA  02110-1301  USA" ;

    return espda_license ;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief
 *   Check if all SOF files exist
 *
 * @param frameset The input set-of-frames
 *
 * @return 1 if not all files exist, 0 if they all exist.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_error_code
espda_files_dont_exist(cpl_frameset *frameset)
{
    const char *func = "dfs_files_dont_exist";

    if (frameset == NULL ) {
        cpl_error_set(func, CPL_ERROR_NULL_INPUT);
        return CPL_ERROR_NULL_INPUT;
    }

    if (cpl_frameset_is_empty(frameset)) {
        return CPL_ERROR_NONE;
    }

    cpl_frameset_iterator* it = cpl_frameset_iterator_new(frameset);
    cpl_frame *frame = cpl_frameset_iterator_get(it);
    const char*  fname = NULL;
    cpl_propertylist* plist = NULL;
    while (frame) {
        fname = cpl_frame_get_filename(frame);
    	cpl_ensure_code(strcmp(fname, "/dev/null"), CPL_ERROR_NULL_INPUT);
    	plist = cpl_propertylist_load(fname, 0);
    	if(!cpl_propertylist_has(plist, "INSTRUME")) {
    		cpl_propertylist_delete(plist);
    		cpl_ensure_code(CPL_FALSE, CPL_ERROR_DATA_NOT_FOUND);
    	}
    	cpl_propertylist_delete(plist);
        if (access(cpl_frame_get_filename(frame), F_OK)) {
            cpl_msg_error(func, "File %s (%s) was not found",
                            cpl_frame_get_filename(frame),
                            cpl_frame_get_tag(frame));
            cpl_error_set(func, CPL_ERROR_FILE_NOT_FOUND);
        }
        cpl_frameset_iterator_advance(it, 1);
        frame = cpl_frameset_iterator_get(it);

    }

    cpl_frameset_iterator_delete(it);

    if (cpl_error_get_code())
        return cpl_error_get_code();

    return CPL_ERROR_NONE;
}
/**@}*/
