/* $Id: espda_fit_gauss.h 2018-05-02 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2018-05-02 14:15:10 +0000 (sat, 02 may 2018) $
 * $Revision: 198015 $
 * $Name:  $
 */
/* 38.11 Examples
 * Based on expfit.c -- model functions for exponential + background model :
 * example program fits a weighted exponential model with background to 
 * experimental data, Y = A \exp(-\lambda t) + b. The first part of the 
 * program sets up the functions expb_f and expb_df to calculate the model 
 * and its Jacobian. The appropriate fitting function is given by

     f_i = ((A \exp(-\lambda t_i) + b) - y_i)/\sigma_i

 * where we have chosen t_i = i. The Jacobian matrix J is the derivative of 
 * these functions with respect to the three parameters (A, \lambda, b).
 * It is given by,

     J_{ij} = d f_i / d x_j

 * where x_0 = A, x_1 = \lambda and x_2 = b.
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

struct data {
  size_t n;
  int npar;
  double * t;
  double * y;
  double * sig;
};

/*-----------------------------------------------------------------------------
               Gaussian function + polynomial background
 -----------------------------------------------------------------------------*/

/*
 * The following function compute residuals for gaussian function + background
 * model which can be a 0, 1, and 2 order polynomial function
 * Gaussian: f(x) = a0*EXP(-(z^2.)/2.) + a3 + a4 * x + a5 * x^2.
 * z=(x-a1)/a2
 * FWHM=2*SQRT(2*ALOG(2))*A2
 *
 */

/*-----------------------------------------------------------------------------
 -----------------------------------------------------------------------------*/

int gaussb_f (const gsl_vector * x, void *data,
        gsl_vector * f)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *y = ((struct data *)data)->y;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk] = gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
/* 
 * Model Yi = sum(i) : A_0 * exp(-((t-A_1)*(t-A_1))/(2*A_2*A_2)) +
 *                              A_3 + A_4 * t + A_5 * t * t
 */

    double ti = t[i];
	int j = 0;

	double Yi= ac[j] * exp (- (ti-ac[j+1])*(ti-ac[j+1])/
		   (2.*ac[j+2]*ac[j+2])) +
           ac[j+3] + ac[j+4] * ti + ac[j+5] * ti * ti ;
           //ac[j+3] + ac[j+4] * ti ;
	gsl_vector_set (f, i, (Yi - y[i])/sig[i]);

    }

  return GSL_SUCCESS;
}

int gaussb_df (const gsl_vector * x, void *data,
         gsl_matrix * J)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk]=gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sig[i],      */
      /* and the xj are the n parameters Ai  */
      
/* Yi= A_0 * exp(-((t-A_1)*(t-A_1))/(2*A_2*A_2)) +
       A_3 + A_4 * t + A_5 * t * t

z=exp(-((t-A_1)*(t-A_1))/(2*A_2*A_2))
dYi/dA_0 = z
dYi/dA_1 = A_0 * z * (t-A_1)/ (A_2*A_2)
dYi/dA_2 = A_0 * z * (t-A_1) * (t-A_1) / (A_2*A_2*A_2)
dYi/dA_3 = 1 
dYi/dA_4 = t 
dYi/dA_5 = t * t 
 */

      int j=0;
      double ti = t[i];
      double s = sig[i];

      double z = exp (- (t[i]-ac[j+1])*(t[i]-ac[j+1]) /
		       (2.*ac[j+2]*ac[j+2]));

      gsl_matrix_set (J, i, j, z / s);
      gsl_matrix_set (J, i, j+1, ac[j] * (ti-ac[j+1]) * 
		      z / s / ac[j+2] / ac[j+2] );
      gsl_matrix_set (J, i, j+2, ac[j] * (ti-ac[j+1]) * (ti-ac[j+1]) * 
                           z / s / ac[j+2] / ac[j+2] / ac[j+2] );
      gsl_matrix_set (J, i, j+3, 1 / s);
      gsl_matrix_set (J, i, j+4, ti / s);
      gsl_matrix_set (J, i, j+5, ( ti * ti ) / s);

    }
  return GSL_SUCCESS;
}

int gaussb_fdf (const gsl_vector * x, void *data,
          gsl_vector * f, gsl_matrix * J)
{
  gaussb_f (x, data, f);
  gaussb_df (x, data, J);

  return GSL_SUCCESS;
}

/*-----------------------------------------------------------------------------
               Lorentzian function + polynomial background
 -----------------------------------------------------------------------------*/

/*
 *
 * The following function compute residuals for Lorentzian function + background
 * model which can be a 0, 1, and 2 order polynomial function
 * Gaussian: f(x) = a0*EXP(-(z^2.)/2.) + a3 + a4 * x + a5 * x^2.
 * z=(x-a1)/a2
 * FWHM=2*SQRT(2*ALOG(2))*A2
 *
 * Lorentzian: f(x) = ((1/pi)*(1/2)*gamma)*(1/((x-x0)^2+(gamma/2)^2))
 * FWHM = gamma = 2 * SQRT(2*log(2)) * sigma = 2.3548 * sigma
 * y=A0*(1./((x-A1)^2.+((2.*SQRT(2.*log(2.))*A2)/2.)^2.))+A3+A4*x+A5*x^2.
 *
 */

/*-----------------------------------------------------------------------------
 -----------------------------------------------------------------------------*/


int lorntzb_f (const gsl_vector * x, void *data,
        gsl_vector * f)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *y = ((struct data *)data)->y;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk] = gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
/*
 *
 * Model Yi = sum(i) : A_0 * (1./ ( ( (t-A_1)*(t-A_1) ) +
 *		                     ( ( (CPL_MATH_FWHM_SIG * A_2) *
 *		                   (CPL_MATH_FWHM_SIG * A_2) ) / 4.) ) ) +
 *                         A_3 + A_4 * t + A_5 * t * t
 *
 */

    double ti = t[i];
	int j = 0;
	double b = CPL_MATH_FWHM_SIG ;

	double  Yi = ac[j] * (1./ ( ( (ti-ac[j+1])*(ti-ac[j+1]) ) +
					          ( ( (b * ac[j+2]) * (b * ac[j+2]) ) / 4.) ) ) +
			                  ac[j+3] + ac[j+4] * ti + ac[j+5] * ti * ti ;
							  //ac[j+3] + ac[j+4] * ti ;

	gsl_vector_set (f, i, (Yi - y[i])/sig[i]);

    }

  return GSL_SUCCESS;
}

int lorntzb_df (const gsl_vector * x, void *data,
         gsl_matrix * J)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk]=gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sig[i],      */
      /* and the xj are the n parameters Ai  */

/* Yi= A_0 * (1./ ( ( (t-A_1)*(t-A_1) ) +
   ( ( (CPL_MATH_FWHM_SIG * A_2) * (CPL_MATH_FWHM_SIG * A_2) ) / 4.) ) ) +
   A_3 + A_4 * t + A_5 * t * t

   	z = t-A_1
   	b = CPL_MATH_FWHM_SIG
   	h = b * A_2
   	k= ( ( z * z ) + ( ( h * h ) / 4.) )
	dYi/dA_0 = 1. / k
	dYi/dA_1 = A_0 * ( ( 2 * z ) / ( k * k) )
	dYi/dA_2 = (-A_0) * ( ( h * b ) / (2 * k * k) )
	dYi/dA_3 = 1
	dYi/dA_4 = t
	dYi/dA_5 = t * t
 */

      int j=0;
      double ti = t[i];
      double s = sig[i];

      double z = ti-ac[j+1] ;
	  double b = CPL_MATH_FWHM_SIG ;
	  double h = b * ac[j+2] ;
	  double k = ( ( z * z ) + ( ( h * h ) / 4.) ) ;
	  double l = (1. / k ) ;

      gsl_matrix_set (J, i, j, l / s);
      gsl_matrix_set (J, i, j+1, (ac[j] * 2. * z / (k  * k) ) /s);
      gsl_matrix_set (J, i, j+2, ( (-ac[j]) * ( ( h * b ) / (2 * k * k) ) ) / s );
      gsl_matrix_set (J, i, j+3, 1 / s);
      gsl_matrix_set (J, i, j+4, ti / s);
      gsl_matrix_set (J, i, j+5, ( ti * ti ) / s);

    }
  return GSL_SUCCESS;
}

int lorntzb_fdf (const gsl_vector * x, void *data,
          gsl_vector * f, gsl_matrix * J)
{
	lorntzb_f (x, data, f);
	lorntzb_df (x, data, J);

  return GSL_SUCCESS;
}

/*-----------------------------------------------------------------------------
                     Parabola function
 -----------------------------------------------------------------------------*/

/*
 *
 * The following function compute residuals for parabola function
 * Gaussian: f(x) = a0 + a1 * x + a2 * x^2.
 *
 */

/*-----------------------------------------------------------------------------
 -----------------------------------------------------------------------------*/


int parb_f (const gsl_vector * x, void *data,
        gsl_vector * f)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *y = ((struct data *)data)->y;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk] = gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
/*
 *
 * Model Yi = sum(i) :   A_0 +  ( ( t - A_1 ) * ( t - A_1 ) * A_2 )
 *
 */

    double ti = t[i];
	int j = 0;
	double z = ti-ac[j+1] ;
	double  Yi = ac[j] + ( z * z * ac[j+2] ) ;

	gsl_vector_set (f, i, (Yi - y[i])/sig[i]);

    }

  return GSL_SUCCESS;
}

int parb_df (const gsl_vector * x, void *data,
         gsl_matrix * J)
{
  size_t n = ((struct data *)data)->n;
  int npar = ((struct data *)data)->npar;
  double *t = ((struct data *)data)->t;
  double *sig = ((struct data *) data)->sig;

  double ac[npar];
  int ijk;
  for (ijk=0;ijk<npar;ijk++)
	ac[ijk]=gsl_vector_get (x, ijk);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sig[i],      */
      /* and the xj are the n parameters Ai  */

/* Yi = A_0 + ( ( t - A_1 ) * ( t - A_1 ) * A_2 )
    z = t - A1
    Yi = A_0 + z ^ 2. * A_2

	dYi/dA_0 = 1
	dYi/dA_1 = dYi/dz * dz/dA_1 = 2. * A_2 * z * (-1) = -2 * A_2 * z
	dYi/dA_2 = ( t - A_1 ) * ( t - A_1 )
 */

      int j=0;
      double ti = t[i];
      double s = sig[i];
      double z = ti-ac[j+1] ;

      gsl_matrix_set (J, i, j, 1. / s);
      gsl_matrix_set (J, i, j+1, (-1. * 2. * ac[j+2] * z) / s);
      gsl_matrix_set (J, i, j+2, ( z * z ) / s);

    }
  return GSL_SUCCESS;
}

int parb_fdf (const gsl_vector * x, void *data,
          gsl_vector * f, gsl_matrix * J)
{
	parb_f (x, data, f);
	parb_df (x, data, J);

  return GSL_SUCCESS;
}

