/*
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_frame.c 162070 2014-06-11 10:18:16Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-06-11 12:18:16 +0200 (Wed, 11 Jun 2014) $
 * $Revision: 162070 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_frame.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup   espda_frame  Frame I/O

  Utilities to handle frames.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Adjust the number of frames.
  @param   set      A frame set.
  @param   tag      A frame tag.
  @param   max      The maximum number of frames allowed for the given tag.
  @return  The number of frames adjusted to the maximum.

  This function counts how many frames in a frame set @a set have the tag
  @a tag and adjust the number of frames to be used to a maximum allowed value
  @a max.

  If @a set or @a tag are NULL, an error is issued. If the number of frames
  with the given tag exceeds @a max and the verbosity level is greater than 0, a
  warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_size espda_frame_adjust(cpl_frameset *set, const char *tag, cpl_size max) {
    extern int v_level;
    espda_util_error(set == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(tag == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_size num_in  = cpl_frameset_count_tags(set, tag);
    cpl_size num_out = num_in;
    if (num_in > max) {
        num_out = max;
        espda_print_verb(1, cpl_msg_warning(cpl_func, "too many frames with "
            "tag %s (%il): only the first %i will be used.", tag, (int)num_in,
            (int)num_out));
    }
    return num_out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Read a frame descriptor.
  @param   set   A frame set.
  @param   tag   A frame tag.
  @param   name  The descriptor's name.
  @param   ...   A pointer to the descriptor.
  @return  CPL error code.

  This function reads a descriptor @a name from a frame with tag @a tag in a
  frame set @a set. A pointer to the descriptor is issued and may be
  dereferenced to an integer, a double, or a string variable.

  If @a set, @a tag, or @a name are NULL, an error is issued. If the descriptor
  does not exist and the verbosity level is greater than 0, a warning is issued.

  N.B. This function may be demoted to a macro in a future release.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_desc(cpl_frameset *set, const char *tag,
    const char *name, ...) {
    extern int v_level;
    espda_util_error(set  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(tag  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(name == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_frame *frame = cpl_frameset_find(set, tag);
    cpl_propertylist *proplist = cpl_propertylist_load_regexp(
        cpl_frame_get_filename(frame), 0, name, 0);
    char *msg = (char *)malloc(sizeof(char) * LEN_STRING);
    sprintf(msg, "descriptor %s not present.", name);
    espda_util_warning(cpl_propertylist_is_empty(proplist), msg);
    cpl_free(msg);
    if (!cpl_propertylist_is_empty(proplist)) {
        cpl_type type = cpl_propertylist_get_type(proplist, name);
        va_list ap;
        va_start(ap, name);
        if (type == CPL_TYPE_INT) {
            int *value_int = va_arg(ap, int *);
            *value_int = cpl_propertylist_get_int(proplist, name);
        }
        else if (type == CPL_TYPE_DOUBLE) {
            double *value_double = va_arg(ap, double *);
            *value_double = cpl_propertylist_get_double(proplist, name);
        }
        else if (type == CPL_TYPE_STRING) {
            char **value_string = va_arg(ap, char **);
            const char *temp = cpl_propertylist_get_string(proplist, name);
            strcpy(*value_string, temp);
        }
        cpl_propertylist_delete(proplist);
        va_end(ap);
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  DEPRECATED!
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_desc_old(cpl_frame *frame,
                                char      *name,
                                double    *value)

{
    extern int v_level;

    if (frame != NULL) {

        // Find descriptor
        cpl_propertylist *proplist = cpl_propertylist_load_regexp(
            cpl_frame_get_filename(frame), 0, name, 0);
	espda_util_warning(cpl_propertylist_has(proplist, name) == 0,
            "descriptor not present.");

	if (cpl_propertylist_has(proplist, name)) {
            cpl_type type = cpl_propertylist_get_type(proplist, name);

            // Load value
            if (type == CPL_TYPE_DOUBLE) {
                *value = cpl_propertylist_get_double(proplist, name);
            }
            else if (type == CPL_TYPE_INT) {
                *value = (double)cpl_propertylist_get_int(proplist, name);
            }
        }
        cpl_propertylist_delete(proplist);
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Export a product from a table into a frame.
  @param    set       A frame set.
  @param    parlist   A list of parameters.
  @param    table     The table to be exported (may be NULL).
  @param    proplist  A property list with QC parameters.
  @param    recipe    The name of the recipe.
  @param    tag       The tag of the frame.
  @param    name      The filename of the frame.
  @return   CPL error code

  This function exports a table @a table into a FITS frame with tag @a tag and
  filename @a name. The user must provide the original frame set @a set, the
  list of parameters @a parlist used by the recipe, and the name of the recipe
  itself @a recipe, as required by @p cpl_dfs_save_table. The descriptors in the
  property list @a proplist are appended to the frame header, together with the
  DAS @p PRO.TYPE descriptor (@p "ANALYZED") and @p PRO.CATG descriptor (equal
  to @a tag).

  If @a set, @a parlist, @a proplist, @a recipe, @a tag, or @a name are NULL,
  an error is issued. If @a table is NULL and the verbosity level is greater
  than 0, a warning is issued.

  N.B. This function may be demoted to a macro in a future release.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_export(cpl_frameset *set,
    const cpl_parameterlist *parlist, cpl_table *table,
    cpl_propertylist *proplist, const char *recipe, const char *tag,
    const char *name) {
    extern int v_level;
    espda_util_error(set  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(parlist  == NULL, CPL_ERROR_NULL_INPUT,
        cpl_error_get_code());
    espda_util_error(proplist == NULL, CPL_ERROR_NULL_INPUT,
        cpl_error_get_code());
    espda_util_error(recipe   == NULL, CPL_ERROR_NULL_INPUT,
        cpl_error_get_code());
    espda_util_error(tag  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(name == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    if (table != NULL) {

        // Append DFS PRO.CATG
        cpl_propertylist_append_string(proplist, CPL_DFS_PRO_TYPE, "ANALYZED");
        cpl_propertylist_append_string(proplist, CPL_DFS_PRO_CATG, tag);

        // Export frame
        cpl_dfs_save_table(set, NULL, parlist, set, NULL, table, NULL, recipe,
            proplist, NULL, PACKAGE "/" PACKAGE_VERSION, name);
    }
    else {
        espda_print_verb(1, cpl_msg_warning(cpl_func, "Frame %s not produced.",
            tag));
    }
    cpl_table_delete(table);
    cpl_propertylist_delete(proplist);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_frame_collect(cpl_frameset  *set, cpl_recipeconfig *cfg,
			       const char *cfg_tag, cpl_size pos, int air_flag, cpl_array *cols,
			       cpl_array **rv, cpl_array **berv,
			       cpl_array **snr600) {
    extern int v_level;
    espda_util_error(set     == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(cfg     == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(cfg_tag == NULL, CPL_ERROR_NULL_INPUT, NULL);

    // Tag list is prepared
    char tag[LEN_STRING];
    char **tag_list = cpl_recipeconfig_get_inputs(cfg, cfg_tag);
    if (pos == -1) {
        strcpy(tag, cfg_tag);
    }
    else {
        strcpy(tag, tag_list[pos]);
    }
    int j = 0;
    while (tag_list[j] != NULL) {
        cpl_free(tag_list[j]);
        j++;
    }
    cpl_free(tag_list);

    // Frames are searched
    int max = cpl_recipeconfig_get_max_count(cfg, cfg_tag, tag);
    int num = espda_frame_adjust(set, tag, max);
    *rv = cpl_array_new(num, CPL_TYPE_DOUBLE);
    *berv = cpl_array_new(num, CPL_TYPE_DOUBLE);
    *snr600 = cpl_array_new(num, CPL_TYPE_DOUBLE);
    cpl_table *table = NULL;
    //espda_print_verb(1, cpl_msg_info(cpl_func, "loading %s...", tag));

    for (int f = 0; f < num; f++) {
        cpl_table *temp = cpl_table_new(0);
        cpl_frame *frame = cpl_frameset_find(set, tag);
        for (int i = 0; i < f; i++) {
            frame = cpl_frameset_find(set, NULL);
        }
        if (!strcmp(tag, S1D) || !strcmp(tag, S2D)) {
	    double rv_val = 0.0;
            espda_frame_desc(set, tag, RV, &rv_val);
	    if (rv_val != 0.0) {
	        cpl_array_set_double(*rv, f, rv_val);
	    }
	    double berv_val = 0.0;
            espda_frame_desc(set, tag, QC_BERV, &berv_val);
	    if (berv_val != 0.0) {
	        cpl_array_set_double(*berv, f, berv_val);
	    }
	    double snr600_val = 0.0;
            espda_frame_desc(set, tag, QC_ORDER102_SNR, &snr600_val);
	    if (snr600_val != 0.0) {
	        cpl_array_set_double(*snr600, f, snr600_val);
	    }
	    /*
	    cpl_propertylist *berv_proplist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(frame), 0, QC_BERV, 0);
            cpl_array_set_double(*berv, f,
				 cpl_propertylist_get_double(berv_proplist,
							     QC_BERV));
	    cpl_propertylist_delete(berv_proplist);
	    */
        }
        if (cpl_frame_get_type(frame) == CPL_FRAME_TYPE_IMAGE) {
	    if (!strcmp(tag, "WAVE_MATRIX")) {
	        cpl_image *wavel   = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 1);
                cpl_image_multiply_scalar(wavel, 0.1);
                cpl_size size_x = cpl_image_get_size_x(wavel);
                cpl_size size_y = cpl_image_get_size_y(wavel);
		/*
                cpl_table_set_size(temp, size_x * size_y);
                espda_spec_full(temp);
                cpl_table_copy_data_double(temp, "WAVEL",
                    cpl_image_get_data_double(wavel));
                cpl_table_copy_data_double(temp, "PIXSIZE",
                    cpl_image_get_data_double(wavel));
                cpl_table_duplicate_column(temp, "PIXSIZE2", temp, "PIXSIZE");
                cpl_table_shift_column(temp, "PIXSIZE2", 1);
                cpl_table_shift_column(temp, "PIXSIZE", -1);
                cpl_table_subtract_columns(temp, "PIXSIZE", "PIXSIZE2");
                cpl_table_multiply_scalar(temp, "PIXSIZE", 0.5);
		for (int j = 0; j < size_y; j++) {
		    cpl_table_set_double(temp, "PIXSIZE", j * size_x, 0.5
		        * (cpl_table_get_double(temp, "WAVEL", j * size_x + 1,
						NULL)
                        -  cpl_table_get_double(temp, "WAVEL", j * size_x,
						NULL)));
		    cpl_table_set_double(temp, "PIXSIZE", (j + 1)* size_x-1,
					 0.5
			* (cpl_table_get_double(temp, "WAVEL", (j + 1) * size_x
						- 1, NULL)
			- cpl_table_get_double(temp, "WAVEL", (j + 1) * size_x
					       - 2, NULL)));
		}
                cpl_table_erase_column(temp, "PIXSIZE2");
		*/
		cpl_array *prova = cpl_array_new(size_x * size_y,
						 CPL_TYPE_DOUBLE);
		cpl_array_copy_data_double(prova,
					   cpl_image_get_data_double(wavel));
            }
	    else if (!strcmp(tag, "DLL_MATRIX")) {
	        cpl_image *dll = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 1);
                cpl_image_multiply_scalar(dll, 0.1);
                cpl_size size_x = cpl_image_get_size_x(dll);
                cpl_size size_y = cpl_image_get_size_y(dll);
                cpl_table_set_size(temp, size_x * size_y);
                espda_spec_full(temp);
                cpl_table_copy_data_double(temp, "PIXSIZE",
                    cpl_image_get_data_double(dll));
	    }
            else {

                cpl_image *flux    = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 1);
                cpl_image *fluxerr = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 2);
                cpl_image *qualdata = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_INT, 0, 3);

                cpl_size size_x = cpl_image_get_size_x(flux);
                cpl_size size_y = cpl_image_get_size_y(flux);
                cpl_table_set_size(temp, size_x * size_y);
                espda_spec_full(temp);
                cpl_table_copy_data_double(temp, "FLUX",
                    cpl_image_get_data_double(flux));
                cpl_table_copy_data_double(temp, "FLUXERR",
                    cpl_image_get_data_double(fluxerr));
                cpl_table_copy_data_int(temp, "QUALDATA",
                    cpl_image_get_data_int(qualdata));

		int ext = 4;  // Vacuum wavelengths are in extension 4
		if (air_flag) {
		    ext = 5;  // Air wavelengths are in extension 5
		}
		cpl_image *wavel   = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, ext);
                cpl_image_multiply_scalar(wavel, 0.1);
		cpl_table_copy_data_double(temp, "WAVEL",
					   cpl_image_get_data_double(wavel));
		if (cpl_frame_get_nextensions(frame) > 5) {
  		    cpl_image *dll = cpl_image_load(cpl_frame_get_filename(
			frame), CPL_TYPE_DOUBLE, 0, ext+2);
		    cpl_image_multiply_scalar(dll, 0.1);
		    cpl_table_copy_data_double(temp, "PIXSIZE",
					       cpl_image_get_data_double(dll));
		}
		else {
                    cpl_table_copy_data_double(temp, "PIXSIZE",
                        cpl_image_get_data_double(wavel));
                    cpl_table_duplicate_column(temp, "PIXSIZE2", temp,
					       "PIXSIZE");
                    cpl_table_shift_column(temp, "PIXSIZE2", 1);
                    cpl_table_shift_column(temp, "PIXSIZE", -1);
                    cpl_table_subtract_columns(temp, "PIXSIZE", "PIXSIZE2");
                    cpl_table_multiply_scalar(temp, "PIXSIZE", 0.5);
		}

		if (cpl_frame_get_nextensions(frame) > 7) {
		    cpl_image *eff = cpl_image_load(cpl_frame_get_filename(
			frame), CPL_TYPE_DOUBLE, 0, 8);
                    cpl_table_duplicate_column(temp, "EFFICIENCY", temp,
					       "FLUX");
		    cpl_table_copy_data_double(temp, "EFFICIENCY",
					       cpl_image_get_data_double(eff));
		    cpl_table_multiply_columns(temp, "EFFICIENCY", "PIXSIZE");
		    //cpl_table_multiply_scalar(temp, "EFFICIENCY", 1e20);
		    cpl_table_multiply_columns(temp, "FLUX", "EFFICIENCY");
		    cpl_table_multiply_columns(temp, "FLUXERR", "EFFICIENCY");
		}

		if (cpl_frame_get_nextensions(frame) > 8) {
		    cpl_image *atm = cpl_image_load(cpl_frame_get_filename(
			frame), CPL_TYPE_INT, 0, 9);
                    cpl_table_duplicate_column(temp, "MASKATM", temp,
					       "SPECID");
		    cpl_table_copy_data_int(temp, "MASKATM",
					    cpl_image_get_data_int(atm));
		}

		/*
		cpl_table *test = cpl_table_duplicate(temp);
		cpl_table_unselect_all(test);
		cpl_table_or_selected_double(test, "WAVEL", CPL_GREATER_THAN, -1000.);
		cpl_table_and_selected_double(test, "WAVEL", CPL_LESS_THAN, -998.);
		cpl_table *extr = cpl_table_extract_selected(test);
		printf("%i\n", cpl_table_get_nrow(extr));
		*/
		double pixsize_med = cpl_table_get_column_median(temp,
								 "PIXSIZE");
		//printf("%f", pixsize_med);
		///*
		cpl_table_unselect_all(temp);
		cpl_table_or_selected_double(temp, "WAVEL", CPL_LESS_THAN, 0.);
		cpl_table_erase_selected(temp);

		for (int j = 1; j < size_y; j++) {
		    cpl_table_set_double(temp, "PIXSIZE", j * size_x, 0.5
		        * (cpl_table_get_double(temp, "WAVEL", j * size_x + 1,
						NULL)
                        -  cpl_table_get_double(temp, "WAVEL", j * size_x,
						NULL)));
		    cpl_table_set_double(temp, "PIXSIZE", (j + 1)* size_x-1,
					 0.5
			* (cpl_table_get_double(temp, "WAVEL", (j + 1) * size_x
						- 1, NULL)
			- cpl_table_get_double(temp, "WAVEL", (j + 1) * size_x
		   			       - 2, NULL)));
		}
		//*/
		///*

		for (int k = 1; k < cpl_table_get_nrow(temp); k++) {
		    if (fabs(cpl_table_get_double(temp, "PIXSIZE", k, NULL)) >
		        3*pixsize_med) {
		      /*
		      cpl_table_dump(temp, k, 1, NULL);
		      printf("%f %f %f %f %f\n",
			     cpl_table_get_double(temp, "WAVEL", k, NULL),
			     cpl_table_get_double(temp, "PIXSIZE", k, NULL),
			     fabs(cpl_table_get_double(temp, "PIXSIZE", k, NULL)),
			     pixsize_med,
			     cpl_table_get_double(temp, "PIXSIZE", k-1, NULL));
		      */
		      cpl_table_set_double(temp, "PIXSIZE", k,
					     cpl_table_get_double(temp, "PIXSIZE", k-1, NULL));
			     //pixsize_med);
		    }
		}
		//*/
		/*
		printf("after\n");
		cpl_table *test2 = cpl_table_duplicate(temp);
		cpl_table_unselect_all(test2);
		cpl_table_or_selected_double(test2, "WAVEL", CPL_GREATER_THAN, 487.2);
		cpl_table_and_selected_double(test2, "WAVEL", CPL_LESS_THAN, 489.7);
		cpl_table *extr2 = cpl_table_extract_selected(test2);
		cpl_table_dump(extr2, 0, 1000, NULL);
		*/
                cpl_table_fill_column_window_int(temp, "SPECID", 0,
                    cpl_table_get_nrow(temp), f + 1);
                for (int o = 0; o < size_y; o++) {
                    cpl_table_fill_column_window_int(temp, "ORDID", o * size_x,
                        size_x, o + 1);
                }

		cpl_image *wavel4   = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 4);
		cpl_image *wavel5   = cpl_image_load(cpl_frame_get_filename(
                        frame), CPL_TYPE_DOUBLE, 0, 5);
		cpl_array *prova4 = cpl_array_new(size_x * size_y,
						 CPL_TYPE_DOUBLE);
		cpl_array *prova5 = cpl_array_new(size_x * size_y,
						 CPL_TYPE_DOUBLE);
		cpl_array_copy_data_double(prova4,
					   cpl_image_get_data_double(wavel4));
		cpl_array_copy_data_double(prova5,
					   cpl_image_get_data_double(wavel5));

            }
        }
        else {

            cpl_table *temp_in = cpl_table_load(cpl_frame_get_filename(frame), 1, 0);
            if (cpl_table_get_nrow(temp_in)==1) {
              const char *wavel_col = cpl_array_get_string(cols, 0);
              const char *flux_col = cpl_array_get_string(cols, 1);
              const char *fluxerr_col = cpl_array_get_string(cols, 2);
              const char *qualdata_col = cpl_array_get_string(cols, 3);
              const cpl_array *wavel0 = NULL;
	      cpl_array *wavel = NULL;
	      const cpl_array *flux = NULL,
                  *fluxerr = NULL, *qualdata = NULL;
              if (cpl_table_has_column(temp_in, wavel_col)) {
                  wavel0 = cpl_table_get_array(temp_in, wavel_col, 0);
              }
              else {
                  cpl_msg_error(cpl_func, "Column %s does not exist.", wavel_col);
              }
              wavel = cpl_array_duplicate(wavel0);
              cpl_array_divide_scalar(wavel, 10.0);
              if (cpl_table_has_column(temp_in, flux_col)) {
                  flux = cpl_table_get_array(temp_in, flux_col, 0);
              }
              else {
                  cpl_msg_error(cpl_func, "Column %s does not exist.", flux_col);
              }
              if (cpl_table_has_column(temp_in, fluxerr_col)) {
                  fluxerr = cpl_table_get_array(temp_in, fluxerr_col, 0);
              }
              else {
                  cpl_msg_error(cpl_func, "Column %s does not exist.", fluxerr_col);
              }
              if (cpl_table_has_column(temp_in, qualdata_col)) {
                  qualdata = cpl_table_get_array(temp_in, qualdata_col, 0);
              }
              else {
                  cpl_msg_error(cpl_func, "Column %s does not exist.", qualdata_col);
              }
              cpl_size size_in = cpl_array_get_size(wavel);
              cpl_table_set_size(temp, size_in);
              espda_spec_full(temp);
              cpl_table_copy_data_double(temp, "WAVEL",
            		cpl_array_get_data_double(wavel));
              cpl_table_copy_data_double(temp, "PIXSIZE",
                cpl_array_get_data_double(wavel));
              cpl_table_duplicate_column(temp, "PIXSIZE2", temp, "PIXSIZE");
              cpl_table_shift_column(temp, "PIXSIZE2", 1);
              cpl_table_shift_column(temp, "PIXSIZE", -1);
              cpl_table_subtract_columns(temp, "PIXSIZE", "PIXSIZE2");
              cpl_table_multiply_scalar(temp, "PIXSIZE", 0.5);
              cpl_table_copy_data_double(temp, "FLUX",
              	cpl_array_get_data_double_const(flux));
              cpl_table_copy_data_double(temp, "FLUXERR",
                cpl_array_get_data_double_const(fluxerr));
                cpl_table_copy_data_int(temp, "QUALDATA",
                  cpl_array_get_data_int_const(qualdata));
              cpl_table_fill_column_window_int(temp, "SPECID", 0, size_in,
                                               f + 1);
              cpl_table_fill_column_window_int(temp, "ORDID", 0, size_in,
                                               INV_INT);
            }
            else {
            temp = cpl_table_load(cpl_frame_get_filename(frame), 1, 0);
            cpl_table_new_column(temp, "SPECID", CPL_TYPE_INT);
            cpl_table_new_column(temp, "ORDID", CPL_TYPE_INT);
            //espda_struct_init(temp, "ORDID", CPL_TYPE_INT);
            cpl_table_fill_column_window_int(temp, "SPECID", 0,
                cpl_table_get_nrow(temp), f + 1);
            cpl_table_fill_column_window_int(temp, "ORDID", 0,
                cpl_table_get_nrow(temp), INV_INT);
            espda_spec_drstodas(temp);
	    cpl_table_multiply_scalar(temp, "WAVEL", 0.1);
	    //cpl_table_multiply_scalar(temp, "WAVEL_AIR", 0.1);
            cpl_size size = cpl_table_get_nrow(temp);
            cpl_table_duplicate_column(temp, "PIXSIZE",  temp, "WAVEL");
	    cpl_table_duplicate_column(temp, "PIXSIZE2", temp, "PIXSIZE");
            cpl_table_shift_column(temp, "PIXSIZE2", 1);
            cpl_table_shift_column(temp, "PIXSIZE", -1);
            cpl_table_subtract_columns(temp, "PIXSIZE", "PIXSIZE2");
            cpl_table_multiply_scalar(temp, "PIXSIZE", 0.5);
            cpl_table_set_double(temp, "PIXSIZE", 0, 0.5
                * (cpl_table_get_double(temp, "WAVEL", 1, NULL)
                -  cpl_table_get_double(temp, "WAVEL", 0, NULL)));
            cpl_table_set_double(temp, "PIXSIZE", size - 1, 0.5
                * (cpl_table_get_double(temp, "WAVEL", size - 1,
                NULL) - cpl_table_get_double(temp, "WAVEL", size
                - 2, NULL)));
            cpl_table_erase_column(temp, "PIXSIZE2");
            cpl_table_dump(temp, 0, 10, NULL);
          }
        }
        if (f == 0) {
            table = cpl_table_duplicate(temp);
        }
        else {
            cpl_table_insert(table, temp, cpl_table_get_nrow(table));
        }
        cpl_table_delete(temp);

    }
    espda_print_verb(0, cpl_msg_info(cpl_func, "%s: %i collected.", tag, num));

    // If frames are wavelength matrixes, they are checked to be air or vacuum
    if (num != 0 &&
	//	(!strcmp(tag, "WAVE_MATRIX") || !strcmp(tag, "DLL_MATRIX"))) {
		 !strcmp(tag, "DLL_MATRIX")) {
        char *pro_catg = (char *)malloc(sizeof(char)*100);
	espda_frame_desc(set, tag, PRO_CATG, &pro_catg);
        char *pref = (char *)malloc(sizeof(char)*4);
	strncpy(pref, pro_catg, 4);
	if (!strcmp(pref, "AIR_") && air_flag==0) {
            cpl_msg_warning(cpl_func, "Wavelengths in %s are in air with "
			    "air-flag=FALSE. Please check your input.", tag);
	}
	if (strcmp(pref, "AIR_") && air_flag) {
            cpl_msg_warning(cpl_func, "Wavelengths in %s are in vacuum with "
			    "air-flag=TRUE. Please check your input.", tag);
	}
	cpl_free(pref);
    }
    return table;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Extract spectra from frames into tables
  @param    set        Frameset with spectra
  @param    wavel_min  Minimum wavelength [nm]
  @param    wavel_max  Maximum wavelength [nm]
  @param    fspec  Full spectrum
  @param    rspec  Rebinned spectrum
  @param    fspec_rv   RV values for the full spectrum
  @param    rspec_rv   RV values for the rebinned spectrum
  @return   CPL_ERROR_NONE iff OK

  This functions takes a set of frames with structure S1D and/or S2D and formats
  them into spectrum tables. Only the portion of spectrum in the range between
  the minimum and maximum wavelength is preserved.

  Frames are loaded by tag. Two sets are created,
  containing a given number of exposures of the same object (depending on the
  input, one of these sets can be empty):

    - a set of merged spectra: S1D(i) + RES_MAP_FIBER;
    - a set of order spectra:  S2D(i) + BLAZE_FIBER(i) +
                               WAVE_MATRIX_FIBER(i) + RES_MAP_FIBER.
    .

  All input merged spectra are concatenated into a single output merged
  spectrum, containing all the exposure; the same is done for the order spectra.
  In the output merged spectrum, columns are filled as follows:

    - WAVEL, FLUX, and FLUXERR are taken from the corresponding columns in
      S1D(i);
    - PIXSIZE is taken from FITS descriptor CDELT1 of S1D(i);
    - RESOL is interpolated from RES_MAP_FIBER;
    - SPECID is taken from the position of S1D(i) in the SOF file.
    .

  In the output order spectrum, columns are filled as follows:

    - WAVEL is taken from WAVE_MATRIX_FIBER(i);
    - PIXSIZE is computed from column WAVEL;
    - FLUX and FLUXERR are taken from the first two extensions of S2D(i),
      corrected for the blaze using BLAZE_FIBER(i);
    - RESOL is interpolated from RES_MAP_FIBER;
    - SPECID is taken from the position of S2D(i) in the SOF file;
    - ORDID is taken from the number of rows in S2D(i) (each row is a
      different order).
    .

  The RV values of the merged spectrum are read from the headers of S1D(i),
  while those of the order spectrum are read from the headers of FLUX_2D(i).

  NB: This function is still provisional. The description refers to the final
      version, which is targeted to handle the ESPRESSO DRS output; since the
      test datasets do not contain blaze frames or resolution frames,
      instructions about these quantities have not been included yet.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_extract_old(cpl_frameset  *set,
                                   double         wavel_min,
                                   double         wavel_max,
                                   cpl_table    **fspec,
                                   cpl_table    **rspec,
                                   cpl_array    **fspec_rv,
                                   cpl_array    **rspec_rv)
{
    cpl_errorstate    prestate = cpl_errorstate_get();
    extern int v_level;

/*
 * Frames are joined into framesets as follows:
 *
 * 1D set:  S1D + RES_MAP (optional);
 * 2D set:  WAVEL_2D + FLUX_2D + FLUXERR_2D + BLAZE_2D (optional) + RES_MAP
 *          (optional).
 *
 * Each frameset can contain more than one spectrum. The 1D set is used to
 * produce a single merged table while the 2D set is used to produce a single
 * order table.
 */

    cpl_msg_info(cpl_func, "Extracting merged and/or order spectra...");

    // Frame with resolution
    cpl_frameset *res_map_set = NULL;
    espda_dfs_call(0, espda_frame_load_old(set, RES_MAP, &res_map_set));

// 1D set

    // Frames with wavelength, flux, and error on flux
    cpl_frameset *spec_1D_set = NULL;
    espda_dfs_call(0, espda_frame_load_old(set, S1D, &spec_1D_set));

    // Loop over frames
    if (cpl_frameset_get_size(spec_1D_set) != 0) {

        cpl_size      rspec_num = cpl_frameset_get_size(spec_1D_set);
        cpl_frameset *rspec_set = cpl_frameset_new();
        cpl_frameset_join(rspec_set, spec_1D_set);
        cpl_frameset_join(rspec_set, res_map_set);
        *rspec_rv = cpl_array_new(rspec_num, CPL_TYPE_DOUBLE);
        *rspec = cpl_table_new(0);
        espda_spec_basic(*rspec);
        espda_struct_check(*rspec, espda_struct_initcol(*rspec, "SPECID",
            CPL_TYPE_INT));

        cpl_size rspec_size = espda_frame_size(rspec_set);

        // Redshift is needed to identify stellar spectra
        double zem = INV_DOUBLE;
        espda_frame_desc(rspec_set, "S1D", DAS_Z_EM, &zem);
	char *unit = (char *)malloc(sizeof(char) * 10);
        espda_frame_desc(rspec_set, "S1D", "CTYPE1", &unit);
        for (int i = 0; i < rspec_num; i++) {
            double rv = INV_DOUBLE;
            espda_dfs_call(0, espda_frame_desc_old(cpl_frameset_get_position(
                rspec_set, i), DAS_RV, &rv));

            cpl_array_set_double(*rspec_rv, i, rv);
            cpl_table *table = cpl_table_new(rspec_size);
            espda_spec_basic(table);
            espda_struct_check(table, espda_struct_initcol(table, "SPECID",
                CPL_TYPE_INT));

            // Fill columns of the merged table
            // N.B. Error on flux is filled only for quasar spectra. This is
            //      provisional (test HARPS data do not have error on flux).
	    espda_dfs_call(0, espda_spec_fill(rspec_set, i, rspec_size,
                "WAVEL",   table));
	    if (!strcmp(unit, "ANGSTROM")) {
	        cpl_table_divide_scalar(table, "WAVEL", 10);
	    }
            espda_dfs_call(0, espda_spec_fill(rspec_set, i, rspec_size,
                "PIXSIZE", table));
            espda_dfs_call(0, espda_spec_fill(rspec_set, i, rspec_size,
                "FLUX", table));
            if (zem > 0.0) {
                espda_dfs_call(0, espda_spec_fill(rspec_set, i, rspec_size,
                    "FLUXERR", table));
            }



            espda_dfs_call(0, espda_spec_fill(rspec_set, i, rspec_size,
                "SPECID", table));
            cpl_table_insert(*rspec, table, 0);
            cpl_table_delete(table);
        }
        cpl_frameset_delete(rspec_set);

        // Sort table by ascending wavelength
        espda_struct_sort_old(*rspec, "WAVEL", FALSE);

        cpl_msg_info(cpl_func, "Merged spectrum extracted.");

        // Apply wavelength range
        cpl_table_and_selected_double(*rspec, "WAVEL", CPL_LESS_THAN,
            wavel_min);
        cpl_table_or_selected_double (*rspec, "WAVEL", CPL_GREATER_THAN,
            wavel_max);
    	cpl_table_erase_selected(*rspec);
        if (cpl_table_get_nrow(*rspec) == 0) {
            cpl_msg_warning(cpl_func, "The merged spectrum is empty after "
                "applying the chosen wavelength range!");
            cpl_table_delete(*rspec);
            *rspec = NULL;
        }

    }
    else {
        *rspec = NULL;
    }
    cpl_frameset_delete(spec_1D_set);


// 2D set

    // Frames with flux
    cpl_frameset *flux_2D_set = NULL;
    espda_dfs_call(0, espda_frame_load_old(set, FLUX_2D, &flux_2D_set));

    if (cpl_frameset_get_size(flux_2D_set) != 0) {

        // Frames with wavelength
        cpl_frameset *wavel_2D_set = NULL;
        espda_dfs_call(0, espda_frame_load_old(set, WAVEL_2D, &wavel_2D_set));

        // Frames with error on flux
        cpl_frameset *fluxerr_2D_set = NULL;
        espda_dfs_call(0, espda_frame_load_old(set, FLUXERR_2D, &fluxerr_2D_set));

        // Frames with blaze function
        cpl_frameset *blaze_2D_set = NULL;
        espda_dfs_call(0, espda_frame_load_old(set, BLAZE_2D, &blaze_2D_set));

        cpl_size      fspec_num = cpl_frameset_get_size(wavel_2D_set);
        cpl_frameset *fspec_set = cpl_frameset_new();
        cpl_frameset_join(fspec_set, wavel_2D_set);
        cpl_frameset_join(fspec_set, flux_2D_set);
        cpl_frameset_join(fspec_set, fluxerr_2D_set);
        cpl_frameset_join(fspec_set, blaze_2D_set);
        cpl_frameset_join(fspec_set, res_map_set);
        *fspec_rv = cpl_array_new(fspec_num, CPL_TYPE_DOUBLE);
        *fspec = cpl_table_new(0);
        espda_spec_basic(*fspec);
        espda_struct_check(*fspec, espda_struct_initcol(*fspec, "SPECID",
            CPL_TYPE_INT));
        espda_struct_check(*fspec, espda_struct_initcol(*fspec, "ORDID",
            CPL_TYPE_INT));

        cpl_size fspec_size = espda_frame_size(fspec_set);

        // Loop over frames
        for (int i = 0; i < fspec_num; i++) {
            double rv;

            espda_dfs_call(0, espda_frame_desc_old(cpl_frameset_get_position(
                flux_2D_set, i), DAS_RV, &rv));
            cpl_array_set_double(*fspec_rv, i, rv);

            cpl_table *table = cpl_table_new(fspec_size);
            espda_spec_basic(table);
            espda_struct_check(table, espda_struct_initcol(table, "SPECID",
                CPL_TYPE_INT));
            espda_struct_check(table, espda_struct_initcol(table, "ORDID",
                CPL_TYPE_INT));

            // Fill columns of the order table
            espda_dfs_call(0, espda_spec_fill(wavel_2D_set,   i, fspec_size,
                "WAVEL",   table));
            espda_dfs_call(0, espda_spec_fill(wavel_2D_set,   i, fspec_size,
                "PIXSIZE", table));
	    espda_dfs_call(0, espda_spec_fill(flux_2D_set,    i, fspec_size,
                "FLUX",    table));
            espda_dfs_call(0, espda_spec_fill(fluxerr_2D_set, i, fspec_size,
                "FLUXERR", table));
	    espda_dfs_call(0, espda_spec_fill(fspec_set,    i, fspec_size,
                "SPECID",  table));
	    espda_dfs_call(0, espda_spec_fill(fspec_set,    i, fspec_size,
                "ORDID",   table));
            cpl_table_insert(*fspec, table, 0);
            cpl_table_delete(table);
        }

        // Erase invalid cells
        cpl_table_and_selected_double(*fspec, "WAVEL",   CPL_LESS_THAN,
            WAVEL_START);
        cpl_table_or_selected_double (*fspec, "WAVEL",   CPL_GREATER_THAN,
            WAVEL_END);
        cpl_table_or_selected_double (*fspec, "WAVEL",   CPL_EQUAL_TO,
            INV_DOUBLE);
        cpl_table_or_selected_double (*fspec, "FLUXERR", CPL_EQUAL_TO, 0.0);
        cpl_table_erase_selected(*fspec);
        cpl_frameset_delete(wavel_2D_set);
        cpl_frameset_delete(fluxerr_2D_set);
        cpl_frameset_delete(blaze_2D_set);
        cpl_frameset_delete(fspec_set);

        // Sort table by ascending wavelength
        espda_struct_sort_old(*fspec, "WAVEL", FALSE);

        cpl_msg_info(cpl_func, " Order spectrum extracted.");

        // Apply wavelength range
        cpl_table_and_selected_double(*fspec,  "WAVEL", CPL_LESS_THAN,
            wavel_min);
        cpl_table_or_selected_double (*fspec,  "WAVEL", CPL_GREATER_THAN,
            wavel_max);
    	cpl_table_erase_selected(*fspec);
        if (cpl_table_get_nrow(*fspec) == 0) {
            cpl_msg_warning(cpl_func, "The order spectrum is empty after "
                "applying the chosen wavelength range!");
            cpl_table_delete(*fspec);
            *fspec = NULL;
        }
    }
    else {
        *fspec = NULL;
    }

    cpl_frameset_delete(flux_2D_set);
    cpl_frameset_delete(res_map_set);


    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to extract spectra from frames into tables.");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Import frames into tables
  @param    set      Frameset
  @param    tag      Frame tag
  @param    table    Spectrum table (SPEC)
  @return   CPL_ERROR_NONE iff OK

  This functions takes a framesets, find the only frame with a given tag and
  import it as a spectrum table. If the frame does not exist, an empty table is
  created. The function is suited to DAS recipes accepting a single frame for
  each tag.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_import(cpl_frameset  *set,
                                  char          *tag,
                                  cpl_table    **table)

{
    //cpl_errorstate  prestate = cpl_errorstate_get();
    extern int v_level;

    cpl_frame *frame = cpl_frameset_find(set, tag);

    // Check if frame exist
    if (frame == NULL) {
        *table = NULL;
    }
    else {
        *table = cpl_table_load(cpl_frame_get_filename(frame), 1, 0);
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Load a frame from a frame set.
  @param   set      A frame set.
  @param   cfg      The recipe configuration object.
  @param   cfg_tag  The configuration tag.
  @param   pos      The position of the tag in the recipe configuration object.
  @return  A table containing the frame.

  This functions scans a frame set @a set, finds the only frame with tag @a tag
  and extracts it into a table. The tags are taken from the recipe configuration
  object @a cfg, defined with its configuration tag @a cfg_tag. The position
  @a pos of the tag ranges from -1 (corresponding to the configuration tag) to
  @a N, with @a N the number of tags in the recipe configuration object. The
  function is not suitable to scan frame sets containing multiple frames with
  the same tag.

  If @a set, @a cfg, or @cfg_tag are NULL, an error is issued. If @a set
  contains multiple frames with the same @a tag and the verbosity level is
  greater than 0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_frame_load(cpl_frameset *set, cpl_recipeconfig *cfg,
    const char *cfg_tag, cpl_size pos) {
    extern int v_level;
    espda_util_error(set     == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(cfg     == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(cfg_tag == NULL, CPL_ERROR_NULL_INPUT, NULL);

    // Tag list is prepared
    char tag[LEN_STRING];
    char **tag_list = cpl_recipeconfig_get_inputs(cfg, cfg_tag);
    if (pos == -1) {
        strcpy(tag, cfg_tag);
    }
    else {
        strcpy(tag, tag_list[pos]);
    }
    int j = 0;
    while (tag_list[j] != NULL) {
        cpl_free(tag_list[j]);
        j++;
    }
    cpl_free(tag_list);

    // Frames are searched
    int max = cpl_recipeconfig_get_max_count(cfg, cfg_tag, tag);
    int num = espda_frame_adjust(set, tag, max);
    cpl_table *table = NULL;
    if (num == 1) {
        espda_print_verb(1, cpl_msg_info(cpl_func, "loading %s...", tag));
        table = cpl_table_load(cpl_frame_get_filename(
            cpl_frameset_find(set, tag)), 1, 0);
    }
    else {
        espda_print_verb(1, cpl_msg_warning(cpl_func, "Found %i frames with "
            "tag %s: use espda_frame_collect instead. Frames will not be "
            "loaded.", num, tag));
    }
    return table;
}

/*
cpl_table *espda_get_table(cpl_frameset *set, char *tag) {
    extern int v_level;
    espda_util_error(set     == NULL, CPL_ERROR_NULL_INPUT, NULL);

    // Frames are searched
    cpl_table *table = NULL;
        cpl_msg_info(cpl_func, "Loading %s...", tag);
        table = cpl_table_load(cpl_frame_get_filename(
            cpl_frameset_find(set, tag)), 1, 0);
    return table;
}
*/
/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/**
  DEPRECATED!
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_load_old(cpl_frameset  *set_in,
                                char          *tag,
                                cpl_frameset **set_out)
{
    cpl_errorstate  prestate = cpl_errorstate_get();
    extern int v_level;

    if (set_in != NULL) {
        *set_out = cpl_frameset_new();
        cpl_size output_num = cpl_frameset_count_tags(set_in, tag);

        // Loop over frames
        for (int i = 0; i < (int)output_num; i++) {
            cpl_frame *frame = NULL;

            // First frame: tag is used
            if (i == 0) {
                frame = cpl_frame_duplicate(cpl_frameset_find(set_in, tag));
            }

            // Other frames: NULL is used
            else {
                frame = cpl_frame_duplicate(cpl_frameset_find(set_in, NULL));
            }
            cpl_frameset_insert(*set_out, frame);
        }
    }

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to load frames by tag.");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  DEPRECATED!
 */
/*----------------------------------------------------------------------------*/

/*
cpl_imagelist *espda_image_load(cpl_frameset *set,
			    cpl_type	     frame_type,
                            cpl_recipeconfig *cfg,
                            char             *cfg_tag,
                            int               pos)

{
    char tag[LEN_STRING];
    char **tag_list = cpl_recipeconfig_get_inputs(cfg, cfg_tag);
    if (pos == -1) {
        strcpy(tag, cfg_tag);
    }
    else {
        strcpy(tag, tag_list[pos]);
    }

    int j = 0;
    while (tag_list[j] != NULL) {
        cpl_free(tag_list[j]);
        j++;
    }
    cpl_free(tag_list);

    int max = cpl_recipeconfig_get_max_count(cfg, cfg_tag, tag);

    int num = espda_frame_adjust(set, tag, max);

    //int i,ext_nr;
    //cpl_image *curr_image = NULL;

    cpl_imagelist *extracted_images = NULL;
    extracted_images = cpl_imagelist_new();

    if (num == 1) {
        cpl_msg_info(cpl_func, "Loading %s...", tag);

	    int ext_nr = cpl_frame_get_nextensions(cpl_frameset_find(set,tag));

	    for (int i = 0; i < ext_nr; i++) {
		cpl_image *curr_image = cpl_image_load(cpl_frame_get_filename(
					cpl_frameset_find(set,tag)),
					    frame_type, 0, i+1);
		cpl_imagelist_set(extracted_images,
					cpl_image_duplicate(curr_image), i);
		cpl_image_delete(curr_image);
	    }
    }

    return extracted_images;
}
*/

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

/*
cpl_imagelist *espda_get_image(cpl_frameset *set,
			    cpl_type	     frame_type,
                            char             *tag)

{

    int i,ext_nr;
    //cpl_image *curr_image = NULL;

    cpl_imagelist *extracted_images = NULL;
    extracted_images = cpl_imagelist_new();

        cpl_msg_info(cpl_func, "Loading %s...", tag);

	    ext_nr = cpl_frame_get_nextensions(cpl_frameset_find(set,tag));

	    for (i = 0; i < ext_nr; i++) {

		cpl_image *curr_image = cpl_image_load(cpl_frame_get_filename(
					cpl_frameset_find(set,tag)),
					    frame_type, 0, i+1);
		cpl_imagelist_set(extracted_images,
					cpl_image_duplicate(curr_image), i);
		cpl_image_delete(curr_image);
	    }

    return extracted_images;
}
*/

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

cpl_error_code espda_frame_extract_by_tag(cpl_frameset *set,
                                          const char* tag,
                                          cpl_frameset* extracted_RE) {

        cpl_frame* frame = NULL;

        espda_util_error(set     == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());

    int nset = cpl_frameset_get_size(set);
    cpl_frameset_iterator* iter = cpl_frameset_iterator_new(set);
    frame = cpl_frameset_iterator_get(iter);

    for ( int iter_frame = 0; iter_frame < nset; iter_frame++ ) {

        const char* curr_tag = cpl_frame_get_tag(frame);

        if (!strcmp(curr_tag, tag)) {
            cpl_frameset_insert(extracted_RE, cpl_frame_duplicate(frame));
        }
        cpl_frameset_iterator_advance(iter, 1);
        frame = cpl_frameset_iterator_get(iter);
    }
    cpl_frameset_iterator_delete(iter);

        return cpl_error_get_code();
}

/**
  @brief    Find the size of frames.
  @param    set   A frame set.
  @return   The size of the frames.

  This functions scans a frame set @a set and returns the size of the frames.
  If the frames have different sizes, the smallest size is returned.

  If the @a set is NULL, an error is issued. If the size of a frame exceeds the
  returned size and the verbosity level is greater than 0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_size espda_frame_size(cpl_frameset *set) {
    extern int v_level;
    espda_util_error(set == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_size num  = cpl_frameset_get_size(set);
    cpl_size size = 0;
    cpl_size size_i[num];
    cpl_frameset_iterator *it1 = cpl_frameset_iterator_new(set);
    for (int i = 0; i < (int)num; i++) {
        cpl_frame *frame = cpl_frame_duplicate(cpl_frameset_iterator_get(it1));

        // 1D images
        if (cpl_frame_get_type(frame) == CPL_FRAME_TYPE_ANY) {
            size_i[i] = cpl_vector_get_size(cpl_vector_load(
                cpl_frame_get_filename(frame), 0));
        }

        // 2D images
        else if (cpl_frame_get_type(frame) == CPL_FRAME_TYPE_IMAGE) {
            cpl_image *image = cpl_image_load(cpl_frame_get_filename(frame),
                CPL_TYPE_DOUBLE, 0, 0);
            size_i[i] = cpl_image_get_size_x(image) *
                        cpl_image_get_size_y(image);
            cpl_image_delete(image);
        }
        if (i == 0 || size_i[i] <= size) {
            size = size_i[i];
        }
        cpl_frame_delete(frame);
        cpl_frameset_iterator_advance(it1, 1);
    }
    cpl_frameset_iterator_delete(it1);
    espda_print_verb(1, cpl_msg_info(cpl_func, "the reference size of the "
        "frames is %i pixels.", (int)size));
    cpl_frameset_iterator *it2 = cpl_frameset_iterator_new(set);
    for (int i = 0; i < (int)num; i++) {
        cpl_frame *frame = cpl_frame_duplicate(cpl_frameset_iterator_get(it2));
        if (size_i[i] != size) {
            espda_print_verb(1, cpl_msg_warning(cpl_func, "frame %s has more "
                "pixels than expected: %i instead of %i.",
                cpl_frame_get_filename(frame), (int)size_i[i], (int)size));
        }
        cpl_frame_delete(frame);
        cpl_frameset_iterator_advance(it2, 1);
    }
    cpl_frameset_iterator_delete(it2);
    return size;
}

/**@}*/
