/* $Id: espda_interp_synth.h 2017-02-25 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2017-05-11 14:15:10 +0000 (sat, 24 feb 2017) $
 * $Revision: 202524 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_interp_synth.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"
#include "espda_struct.h"
#include "espda_syst.h"

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_interp_synth
 *
 * Interpolation of synthetic spectra
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpolate within the cube of synthetic spectra
  @param    wobs     Array of wavelength of observed spectrum
  @param    wsyn     Array of wavelength of synthetic spectrum
  @param    fsyn1    Flux array 1 of the cube of synthetic spectra
  @param    fsyn2    Flux array 2 of the cube of synthetic spectra
  @param    fsyn3    Flux array 3 of the cube of synthetic spectra
  @param    fsyn4    Flux array 4 of the cube of synthetic spectra
  @param    fsyn5    Flux array 5 of the cube of synthetic spectra
  @param    fsyn6    Flux array 6 of the cube of synthetic spectra
  @param    fsyn7    Flux array 7 of the cube of synthetic spectra
  @param    fsyn8    Flux array 8 of the cube of synthetic spectra
  @param    fsyn     Flux of the interpolated synthetic spectrum
  @param    t1       Teff value of the lower corner of the cube
  @param    t2       Teff value of the upper corner of the cube
  @param    g1       Logg value of the lower corner of the cube
  @param    g2       Logg value of the upper corner of the cube
  @param    m1       [Fe/H] value of the lower corner of the cube
  @param    m2       [Fe/H] value of the upper corner of the cube
  @param    teff     Effective temperature (Teff) of the star
  @param    logg     Surface gravity (logg) of the star
  @param    feh      Metallity ([Fe/H]) of the star
  @return   CPL_ERROR_NONE iff OK

  @doc      
  This function interpolates the flux of synthetic spectra
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_interp_synth(cpl_array     *wobs,
			          cpl_array     *wsyn,
			          cpl_array     *fsyn1,
 			          cpl_array     *fsyn2,
 			          cpl_array     *fsyn3,
 			          cpl_array     *fsyn4,
 			          cpl_array     *fsyn5,
 			          cpl_array     *fsyn6,
 			          cpl_array     *fsyn7,
 			          cpl_array     *fsyn8,
 			          cpl_array     *fsyn,
   		                  double  t1,
   		                  double  t2,
   		                  double  g1,
   		                  double  g2,
   		                  double  m1,
   		                  double  m2,
   		                  double  teff,
		                  double  logg,
			              double  feh)

{
    cpl_errorstate prestate = cpl_errorstate_get();

/*
 * This function interpolate the eight flux arrays from the grid of        *
 * synthetic spectra to generate a single output spectrum, which is also   *
 * interpolated in the wavelength grid of the observed merged, normalized  * 
 * spectrum of the star                                                    *
 */

// Defining the number of elements of the arrays

    cpl_size nobs = cpl_array_get_size (wobs);
    cpl_size nsyn = cpl_array_get_size (wsyn);

// We compute the distances to the eight points of the cube within the stellar
// parameter space

    double dt = (teff-t1) / (t2-t1);
    double dg = (logg-g1) / (g2-g1);
    double dm = (feh-m1) / (m2-m1);

    printf("Number of points of observed spectrum: %i \n", (int)nobs);
    printf("Number of points of synthetic spectrum: %i \n", (int)nsyn);

    printf("Selected Teff [K] : %f \n", (double)teff);
    printf("Selected log (g/cm s2) : %f \n", (double)logg);
    printf("Selected [Fe/H] : %f \n", (double)feh);

    printf("Weights for interpolation \n");
    printf("Weight on Teff [K] : %f \n", (double)dt);
    printf("Weight on log (g/cm s2) : %f \n", (double)dg);
    printf("Weight on [Fe/H] : %f \n", (double)dm);

// We interpolate the flux arrays of the eight points the cube
    
    cpl_array *flxs  = cpl_array_duplicate (fsyn1);
    
    for (int j = 0; j < (int)nsyn; j++) {
      double cf1 = cpl_array_get_double (fsyn1, j, NULL);
      double cf2 = cpl_array_get_double (fsyn2, j, NULL);
      double cf3 = cpl_array_get_double (fsyn3, j, NULL);
      double cf4 = cpl_array_get_double (fsyn4, j, NULL);
      double cf5 = cpl_array_get_double (fsyn5, j, NULL);
      double cf6 = cpl_array_get_double (fsyn6, j, NULL);
      double cf7 = cpl_array_get_double (fsyn7, j, NULL);
      double cf8 = cpl_array_get_double (fsyn8, j, NULL);

      // sgrid(it-1,ig-1,im-1) -> mnam1= mt1 + mg1 + mm1;
      // sgrid(it-1,ig-1,im)   -> mnam2= mt1 + mg1 + mm2;
      // sgrid(it-1,ig,im-1)   -> mnam3= mt1 + mg2 + mm1;
      // sgrid(it-1,ig,im)     -> mnam4= mt1 + mg2 + mm2;
      // sgrid(it,ig-1,im-1)   -> mnam5= mt2 + mg1 + mm1;
      // sgrid(it,ig-1,im)     -> mnam6= mt2 + mg1 + mm2;
      // sgrid(it,ig,im-1)     -> mnam7= mt2 + mg2 + mm1;
      // sgrid(it,ig,im)       -> mnam8= mt2 + mg2 + mm2;

      double cfs1 = (1-dt) * (1-dg) * (1-dm) * cf1;
      double cfs2 = (1-dt) * (1-dg) * dm * cf2;
      double cfs3 = (1-dt) * dg * (1-dm) * cf3;
      double cfs4 = (1-dt) * dg * dm * cf4;
      double cfs5 = dt * (1-dg) * (1-dm) * cf5;
      double cfs6 = dt * (1-dg) * dm * cf6;
      double cfs7 = dt * dg * (1-dm) * cf7;
      double cfs8 = dt * dg * dm * cf8;
      double cfs = cfs1 + cfs2 + cfs3 + cfs4 + cfs5 + cfs6 + cfs7 + cfs8;

      cpl_array_set_double (flxs, j, cfs);

    }

    //Convolution of interpolated synthetic spectrum
    //      cpl_ensure_code(espda_cont_ord_eval(wax, xcf_in,  
    //                    ctx_out) == CPL_ERROR_NONE, cpl_error_get_code());


    // We interpolate the new flux array of the synthetic spectrum 
    // in the wavelength array of the observed spectrum 

//The interpolation function for a given dataset is stored in a gsl_interp object. These are created by the following functions.

//Function: gsl_interp * gsl_interp_alloc (const gsl_interp_type * T, size_t size)
//This function returns a pointer to a newly allocated interpolation object of type T for size data-points.

//Function: int gsl_interp_init (gsl_interp * interp, const double xa[], const double ya[], size_t size)
//This function initializes the interpolation object interp for the data (xa,ya) where xa and ya are arrays of size size. The interpolation object (gsl_interp) does not save the data arrays xa and ya and only stores the static state computed from the data. The xa data array is always assumed to be strictly ordered, with increasing x values; the behavior for other arrangements is not defined.

//Function: void gsl_interp_free (gsl_interp * interp)
//This function frees the interpolation object interp.

//Interpolation Type: gsl_interp_cspline
//Cubic spline with natural boundary conditions. The resulting curve is piecewise cubic on each interval, with matching first and second derivatives at the supplied data-points. The second derivative is chosen to be zero at the first point and last point.

    //printf("Initiate the interpolation \n");

    gsl_interp_accel *acc = gsl_interp_accel_alloc ();
    gsl_spline *spline = gsl_spline_alloc (gsl_interp_cspline, nsyn);

    double xsyn[nsyn] ;
    double ysyn[nsyn] ;
    for (int j = 0; j < (int)nsyn; j++) {
       double cxs = cpl_array_get_double (wsyn, j, NULL);
       double cys = cpl_array_get_double (flxs, j, NULL);
       xsyn[j] = cxs ;
       ysyn[j] = cys ;
    }

    //printf("Initiate the interpolation 2 \n");
    double xs_max= cpl_array_get_max (wsyn);
    double xs_min= cpl_array_get_min (wsyn);
    double xo_max= cpl_array_get_max (wobs);
    double xo_min= cpl_array_get_min (wobs);
    printf("xsyn_max = %f xsyn_min= %f \n",xs_max,xs_min);
    printf("xobs_max = %f xobs_min= %f \n",xo_max,xo_min);

    gsl_spline_init (spline, xsyn, ysyn, nsyn);

    for (int j = 0; j < (int)nobs; j++) {
       double cwobs = cpl_array_get_double (wobs, j, NULL);
       double cfobs = gsl_spline_eval (spline, cwobs, acc);
       cpl_array_set_double (fsyn, j, cfobs);
    }

    //printf("Finalize the interpolation \n");

    gsl_spline_free (spline);
    gsl_interp_accel_free (acc);

    if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to create the interpolated synthetic spectrum");
    }

    return CPL_ERROR_NONE;
}

/**@}*/
