/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_print.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_print.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_print  Print results
  
  Special messages (like tables with fitting results, etc.) are better printed
  through dedicated functions. These functions are meant to be called through
  espda_print_verb, to skip them when verbosity is low.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Print the covariance matrix.
  @param   covar     The covariance matrix.
  @param   par_vary  The varying parameters.
  @return  CPL error code.
  
  This function prints in the standard output the covariance matrix @a covar 
  (computed during the chi-square minimization of a Voigt profile based on the 
  varying parameters @a par_vary). Each table entry contains both the actual 
  value and the normalized value (in parentheses) of one matrix element.
  
  The covariance matrix is computed by the GSL package @p gsl_multifit. @a covar 
  is a @f$ M\times M @f$ matrix and @a par_vary is a @f$ M @f$-d vector, where 
  @f$ M=4L-C @f$, with L the number of lines (each fitted with 4 Voigt 
  parameters) and C the number of constraint among parameters. 
    
  If @a covar or @a par_vary are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_print_covar(gsl_matrix *covar, gsl_vector *par_vary) {
    //cpl_errorstate  prestate = cpl_errorstate_get();
    espda_util_error(covar == NULL,    CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(par_vary == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    cpl_msg_info(cpl_func, "covariance matrix (normalized values in "
        "parentheses):");
    int  vary_num = par_vary->size;
    for (int row = 0; row < vary_num; row++) {
        char row_string[LEN_STRING];
        char open [10] = " |";
        char close[10] = "| ";
        if (row == 0) {
            strcpy(open,  " |");
            strcpy(close, "| ");
        }
        else if (row == vary_num - 1) {
            strcpy(open,  " |");
            strcpy(close, "| ");       
        }                 
        for (int col = 0; col < vary_num; col++) {
            char col_string[20];
             char prefix    [10] = "  ";
             char suffix    [10] = "  ";
             if (col == row) {
                 strcpy(prefix, " {");
                 strcpy(suffix, "} ");
             }
             sprintf(col_string, "%+5.2e (%+5.3f)", 
                      gsl_matrix_get(covar, row, col), 
             	      gsl_matrix_get(covar, row, col) /
             	 sqrt(gsl_matrix_get(covar, row, row) *
             	      gsl_matrix_get(covar, col, col)));
            if (col == 0) {
                strcpy(row_string, open);
            }
            strcat(row_string, prefix);
            strcat(row_string, col_string);
            strcat(row_string, suffix);
            if (col == vary_num - 1) {
                strcat(row_string, close);
            }
        }
        cpl_msg_info(cpl_func, " %s", row_string);
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Print the result of Voigt-profile fitting.
  @param   line  A list of lines (LINE_VOIGT).
  @return  CPL error code.
  
  This function prints a table in the standard output listing the Voigt 
  parameters from a list of fitted lines @a line. Parameters are listed with 
  their fitting error and are taken from columns @a WAVEL and @a WAVELERR, 
  @a LINEZ and @a LINEZERR, @a COLDEN and @a COLDENERR, @a THERB and 
  @a THERBERR, and @a TURBB and @a TURBBERR of @a line.
  
  If @a line is NULL an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_print_line(cpl_table *line) {
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_size line_size = cpl_table_get_nrow(line);
    cpl_msg_info(cpl_func, "---------------------------------------------------"
        "------------------------------------------------------");
    cpl_msg_info(cpl_func, " Line            WAVEL               LINEZ         "
        "     COLDEN              THERB             TURBB");
    for (int i = 0; i < (int)line_size; i++) {
        cpl_msg_info(cpl_func, "  %2i       %7.3f +/-%6.3f    %8.5f +/-%8.5f    "
            "%6.3f +/-%7.3f    %6.2f +/-%7.2f   %6.2f +/-%7.2f", (int)i + 1,
            cpl_table_get_double(line, "WAVEL",     i, NULL), 
            cpl_table_get_double(line, "WAVELERR",  i, NULL), 
            cpl_table_get_double(line, "LINEZ",     i, NULL), 
            cpl_table_get_double(line, "LINEZERR",  i, NULL), 
            cpl_table_get_double(line, "COLDEN",    i, NULL), 
            cpl_table_get_double(line, "COLDENERR", i, NULL), 
            cpl_table_get_double(line, "THERB",     i, NULL), 
            cpl_table_get_double(line, "THERBERR",  i, NULL), 
            cpl_table_get_double(line, "TURBB",     i, NULL), 
            cpl_table_get_double(line, "TURBBERR",  i, NULL)); 
    }
    cpl_msg_info(cpl_func, "---------------------------------------------------"
                 "------------------------------------------------------");
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Print the Voigt parameters.
  @param   solverx   The solver vector.
  @param   par_vary  The varying parameters.
  @param   par_lim   Range of varying parameters.
  @return  CPL error code.
  
  This function prints in the standard output two vectors:
  
  -# @a par_vary, which contains the initial varying parameters to define a 
     Voigt profile; 
  -# @a solverx, which contains the fitting parameters computed during 
     chi-square minimization.

  Depending on the globally-defined transformation mode TRANSF_MODE, both 
  vectors may be transformed to correctly enforce ranges for parameters, as 
  defined by the columns @a MIN and @ MAX of @a par_lim (see function 
  @p espda_voigt_transf). In this case, the transformed values are also printed 
  in parentheses.  

  The solver vector is computed by the GSL package @p gsl_multifit. Both 
  @a solverx and @a par_vary are @f$ M @f$-d vectors, where @f$ M=4L-C @f$, 
  with L the number of lines (each fitted with 4 Voigt parameters) and C the 
  number of constraint among parameters. 

  If @a solverx, @a par_vary, or @a par_lim are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_print_par(gsl_vector *solverx, gsl_vector *par_vary,
    cpl_table *par_lim) {
    espda_util_error(solverx  == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(par_vary == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(par_lim  == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    if (strcmp(TRANSF_MODE, "T")) {
    	cpl_msg_info(cpl_func, "Voigt parameters:");
    }
    else {
    	cpl_msg_info(cpl_func, "Voigt parameters (transformed values in "
            "parentheses):");
    }
    char string_in[LEN_STRING];
    char string_out[LEN_STRING];
    char open [10] = " [";
    char close[10] = "] ";
    int  vary_num = par_vary->size;
    for (int elem = 0; elem < vary_num; elem++) {
        char elem_in[LEN_STRING], elem_out[LEN_STRING];
        double par_in  = gsl_vector_get(par_vary, elem);
        double par_out = gsl_vector_get(solverx,  elem);
        double min     = cpl_table_get_double(par_lim, "MIN", elem, NULL);
        double max     = cpl_table_get_double(par_lim, "MAX", elem, NULL);
        double par_in_proc,  err_in_proc;
        double par_out_proc, err_out_proc;
        cpl_ensure_code(espda_voigt_transf(par_in,  0, min, max, "ATRANSF", 
            &par_in_proc,  &err_in_proc)  == CPL_ERROR_NONE, 
            cpl_error_get_code());
        cpl_ensure_code(espda_voigt_transf(par_out, 0, min, max, "ATRANSF",
            &par_out_proc, &err_out_proc) == CPL_ERROR_NONE, 
            cpl_error_get_code());
        if (!strcmp(TRANSF_MODE, "NONE")) {
            sprintf(elem_in,  "  %+5.2e           ", par_in_proc);
            sprintf(elem_out, "  %+5.2e           ", par_out_proc);
        }
        else {
            sprintf(elem_in,  "  %+5.2e (%+5.3f)  ", par_in_proc, par_in);
            sprintf(elem_out, "  %+5.2e (%+5.3f)  ", par_out_proc, par_out);
        }
        if (elem == 0) {
            strcpy(string_in,  open);
            strcpy(string_out, open);
        }
        strcat(string_in,  elem_in);
        strcat(string_out, elem_out);
        if (elem == vary_num - 1) {
            strcat(string_in,  close);
            strcat(string_out, close);
        }
    }
    cpl_msg_info(cpl_func, " %s", string_in);
    cpl_msg_info(cpl_func, " %s", string_out);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Print the list of absorption systems.
  @param   line      A list of lines (LINE_SYST).
  @param   syst_num  Number of systems.
  @return  CPL error code.
  
  This function prints a table in the standard output listing @a syst_num 
  absorption systems from a list of identified lines @a line. The columns of the
  printed table are:
  
  -# ID: from column @a SYSTID; 
  -# Redshift: average of @a LINEZ values for all the lines in the system;
  -# Lines: number of lines in the system;
  -# Shared: number of lines that have been associated to at list another system
     in addition to the one considered.
  
  If @a line is NULL an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_print_syst(cpl_table *line, cpl_size syst_num) {
    extern int v_level;
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_struct_sort(line, 1, "SYSTID", FALSE);
    cpl_msg_info(cpl_func, "-------------------------------------");
    cpl_msg_info(cpl_func, " ID     Redshift      Lines   Shared");
    cpl_array *syst_z = espda_syst_ave(line, syst_num);
    for (int n = 0; n < syst_num; n++) {

        char syst_id[4];
        espda_syst_adv(syst_id);
        int row_start, row_num;//  = 0, row_num = 0;
        espda_dfs_call(0, espda_syst_find(line, syst_id, &row_start, 
            &row_num));
        int shr_num = 0;
        for (int m = row_start; m < row_start + row_num; m++) {
            cpl_table *temp = NULL;
            int mult;
            espda_struct_extract_old(line, temp, "WAVEL", CPL_TYPE_DOUBLE,
                CPL_EQUAL_TO, cpl_table_get_double(line, "WAVEL", m, NULL));
            cpl_table_and_selected_string(temp, "SYSTID", 
                CPL_NOT_EQUAL_TO, INV_STRING);
            cpl_table *entries = cpl_table_extract_selected(temp); 
            espda_struct_size(entries, mult, 0);
            if (mult > 1) {
                shr_num++;
            }            
	    espda_struct_delete(CPL_TABLE, 2, temp, entries);
        }
        if (row_num != 0) {
            cpl_msg_info(cpl_func, " %2s     %8.6f        %-2i       %-2i", 
                syst_id, cpl_array_get_double(syst_z, n, NULL), row_num, 
                shr_num);
        }        
    }
    cpl_msg_info(cpl_func, "-------------------------------------");
    espda_struct_sort(line, 1, "LINEZ", FALSE);
    cpl_array_delete(syst_z);
    return CPL_ERROR_NONE;
}

/**@}*/
