 /* $Id: espda_spec.c 285570 2023-11-14 15:04:59Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2023-11-14 15:04:59 +0000 (Tue, 14 Nov 2023) $
 * $Revision: 285570 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_spec.h"
#include <time.h>

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_spec  Spectra

  A spectrum (SPEC) is a table with information about the flux density as a
  function of wavelength. The rows of the table are also called pixels,
  especially when they are referred to non-rebinned spectra, where each row
  corresponds one-to-one to a pixel of the detector. Four types of spectra are
  used: basic spectrum (SPEC_BASIC), spectrum with information about flux RMS or
  "rebinned spectrum" (SPEC_RMS), spectrum with information about orders and
  exposures or "full spectrum" (SPEC_FULL), fitted spectrum (SPEC_FIT).

  The columns of SPEC_BASIC are:

  -# @a WAVEL: wavelength [nm];
  -# @a MASK: flag for masked rows (deprecated);
  -# @a PIXSIZE: pixel size [nm];
  -# @a FLUX: flux density [erg cm^-2 s^-1 nm^-1];
  -# @a FLUXERR: error on flux density [erg cm^-2 s^-1 nm^-1].

  SPEC_RMS has all the columns of SPEC_BASIC plus the following additional
  column:

  -# @a FLUXRMS: RMS of flux density [erg cm^-2 s^-1 nm^-1].

  SPEC_FULL has all the columns of SPEC_BASIC plus the following additional
  columns:

  -# @a SPECID: ID of the exposure;
  -# @a ORDID: ID of the order;
  -# @a EQFACT: equalizazion factor.

  SPEC_FIT has all all the columns of SPEC_RMS and SPEC_FULL plus the following
  additional columns:

  -# @a FIT: fitted flux density of the absorption lines [erg cm^-2 s^-1 nm^-1];
  -# @a CONT: fitted flux density of the emission continuum [erg cm^-2 s^-1
     nm^-1];
  -# @a NORM: flux density normalized to the emission continuum (@a FLUX/
     @a CONT);
  -# @a NORMERR: error on normalized flux density (@a FLUXERR/@a CONT).
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Average a spectrum.
  @param    spec         A spectrum (SPEC).
  @param    data_col     The column with the data to be averaged.
  @param    dataerr_col  The column with the errors (may be NULL).
  @param    pixfrac_col  The column with the pixel fraction (may be NULL).
  @param    row_start    The starting row.
  @param    row_end      The ending row.
  @param    data_ave     The (optionally weighted) average.
  @param    dataerr_ave  The error on the average.
  @param    datarms      The RMS.
  @return   CPL error code.

  This functions computes the average @a data_ave, its error @a dataerr_ave,
  and the RMS @a datarms of the data from a column @a data_col of a spectrum
  @a spec. The average is optionally weighted by the errors @a dataerr_col
  and/or by the pixel fraction covered by each data point @a pixfrac_col. Only
  the row range @a row_start and @a row_end is considered.

  The weight of the row @f$ r @f$ of @a spec is defined as @f[
    w_r=\frac{f_r}{\sigma_r^2},
  @f]
  where @f$ f_r @f$ is taken from column @a pixfrac_col (if provided, otherwise
  it is 1) and @f$ \sigma_r @f$ is taken from column @a dataerr_col (if
  provided, otherwise it is 1). @a data_ave, @a dataerr_ave, and @a datarms are
  thus defined as @f[
    \mathit{data\_ave}=\frac{\sum_r{d_r w_r}}{\sum_r{w_r}},
  @f]
  @f[
    \mathit{dataerr\_ave}=\frac{\sqrt{\sum_r{\sigma_r^2 w_r^2}}}{\sum_r{w_r}},
  @f]
  @f[
    \mathit{datarms}=\frac{\left(\sum_r{d_r^2 w_r}-\sum_r{d_r w_r}\times
      \mathit{data\_ave}^2\right)\sum_r{w_r^2}/\left(\sum_r{w_r}\right)^2}
      {\sum_r{w_r}-\sum_r{w_r}/(\mathit{row\_end}-\mathit{row\_start})},
  @f]
  where @f$ d_r @f$ is taken from column @a pixfrac and @f$ r @f$ spans from
  @a row_start to @a row_end.

  If @a spec or @a data_col are NULL, an error is issued. If @a dataerr_col or
  @a pixfrac_col are NULL and the verbosity level is greater than 0, a warning
  is issued. If @f$ \sum_r{w_r}\leq 0 @f$ or @f$ \sum_r{\sigma_r^2 w_r^2}<0 @f$
  and the verbosity level it greater than 1, a warning is also issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_spec_ave(const cpl_table *spec, const char *data_col,
    const char *dataerr_col, const char *pixfrac_col, const cpl_size row_start,
    const cpl_size row_end, double *data_ave, double *dataerr_ave,
    double *datarms) {
    extern int v_level;
    espda_util_error(spec          == NULL, CPL_ERROR_NULL_INPUT,
        cpl_error_get_code());
    espda_util_error(data_col      == NULL, CPL_ERROR_NULL_INPUT,
        cpl_error_get_code());
    //espda_util_warning(dataerr_col == NULL, "error column not provided.");
    //espda_util_warning(pixfrac_col == NULL, "pixel fraction column not "
    //    "provided.");
    double     wgh_sum = 0.0;
    double     wghsq_sum = 0.0;
    double     data_sum = 0.0;
    double     datanw_sum = 0.0;
    double     dataerr_sum = 0.0;
    double     datasq_sum = 0.0;
    double     datasqnw_sum = 0.0;

    // Loop over the row range to compute sums
    //cpl_table_dump(spec, 0, 10, NULL);
    for (int j = row_start; j < row_end; j++) {
        double data = cpl_table_get_double(spec, data_col, j, NULL);

        // If data errors are not provided, they are set equal to 1
        double dataerr;
        if (dataerr_col == NULL) {
            dataerr = 1.0;
        }
        else {
            dataerr = cpl_table_get_double(spec, dataerr_col, j, NULL);
        }

        // If pixel fractions are not provided, they set equal to 1
        double pixfrac;
        if (pixfrac_col == NULL) {
            pixfrac = 1.0;
        }
        else {
            pixfrac = cpl_table_get_double(spec, pixfrac_col, j, NULL);
	    if (pixfrac == INV_DOUBLE) {
	        pixfrac = 1.0;
	    }
        }

        if (dataerr != 0.0 && !isnan(data) && !isnan(dataerr)) {
	  //if (pixfrac < 0) printf("pixfrac %f\n", pixfrac);
            double wgh   = pixfrac / (dataerr * dataerr);
	    //wgh = 1;
            double wghsq = wgh * wgh;
            wgh_sum      = wgh_sum     + wgh;
            wghsq_sum    = wghsq_sum   + wghsq;
            data_sum     = data_sum    + wgh   * data;
	    datanw_sum   = datanw_sum + data;
            datasq_sum   = datasq_sum  + wgh   * data    * data;
	    datasqnw_sum = datasqnw_sum + data * data;
            dataerr_sum  = dataerr_sum + wghsq * dataerr * dataerr;
            //dataerr_sum  = dataerr_sum + 1 / (dataerr * dataerr);
        }
        else {
            espda_print_verb(2, cpl_msg_warning(cpl_func, "Row %i of %i "
                "(wavelength %7.3f) has null flux error! It will not be used.",
                j + 1, (int)(row_end - row_start),
                cpl_table_get_double(spec, "WAVEL", j, NULL)));
        }
    }

    // Compute average
    //double     datarmssq_den;
    //double     datarmssq_num;
    //if (wgh_sum > 0.0 && dataerr_sum >= 0.0) {
    if (wgh_sum != 0.0 && dataerr_sum >= 0.0) {
        *data_ave     =      data_sum     /  wgh_sum;
        *dataerr_ave  = sqrt(dataerr_sum) /  wgh_sum;
        //*dataerr_ave  = sqrt(dataerr_sum);
        double datarmssq_num = (datasq_sum - wgh_sum * *data_ave * *data_ave) *
	                       wghsq_sum / (wgh_sum * wgh_sum);
        double datarmssq_den =  wgh_sum - wgh_sum / (row_end - row_start);

	datarmssq_num = (datasq_sum + wgh_sum* *data_ave * *data_ave
			     // - 2 * wgh_sum* *data_ave * data_sum) *
			      - 2 * *data_ave * data_sum) *
	                       wghsq_sum / (wgh_sum * wgh_sum);
	datarmssq_den = wgh_sum - wgh_sum / (row_end - row_start);

	int N = row_end - row_start;
	//printf("%e %e\n", datasqnw_sum, datanw_sum);
	datarmssq_num = fabs(datasqnw_sum + N * *data_ave * *data_ave
			 - 2 * *data_ave * datanw_sum);
	datarmssq_den = N - 1;


        // Computer RMS
	//printf("%e %e %e %e %e %e %e\n", datasq_sum, wgh_sum* *data_ave * *data_ave,2 * wgh_sum* *data_ave * data_sum, wghsq_sum, wgh_sum, datarmssq_num, datarmssq_den);
        if (datarmssq_num > 0.0 && datarmssq_den > 0.0 &&
            row_end > row_start + 1) {
            *datarms = sqrt(datarmssq_num/datarmssq_den);
	    //printf("%e\n",*datarms);
        }

        // Inconsistent values: invalid results
        else {
            *datarms = INV_DOUBLE;
        }
    }

    // inconsistent values: invalid results
    else {
        *data_ave    = INV_DOUBLE;
        *dataerr_ave = INV_DOUBLE;
        *datarms     = INV_DOUBLE;
        espda_print_verb(2, cpl_msg_warning(cpl_func, "Invalid average: "
            "weights are too low. Sum of weights: %f.", wgh_sum));
    }
    //printf("%e %e %f %f\n", wgh_sum, dataerr_sum, *data_ave, *dataerr_ave);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Remove outliers from a spectrum
  @param    spec         Spectrum (SPEC)
  @param    data_col     Column with data points
  @param    dataerr_col  Column with data errors
  @param    kappa        Value to be used in kappa-sigma clipping
  @param    rej_num      Number of data points rejected
  @return   CPL_ERROR_NONE iff OK

  This functions uses a standard kappa-sigma clipping algorithn to reject
  outliers from data contained in a spectrum table column. Data are averaged
  using their errors as weights; the farthest outlier is removed and the
  procedure is iterated until no more outliers are found outside kappa standard
  deviations or a minimum number of points is reached.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_spec_clean(cpl_table *spec, cpl_table *ref,
				const char *data_col, const char *dataerr_col,
				int kappa, cpl_size *rej_num) {
    cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;

    // Initialize counters of selected rows
    cpl_size sel_num_in  = cpl_table_get_nrow(spec);
    cpl_size sel_num;
    int      sel_num_thr = 2;
    int count = 0;

    // Loop: stop iteration if the number of rows falls below a threshold
    //printf("before\n");
    //cpl_table_dump(spec, 0, 100, NULL);
    /*
    double data_ave;
    double dataerr_ave;
    double datarms;
    espda_util_func(espda_spec_ave(ref, data_col, dataerr_col, NULL,
        0, sel_num_in, &data_ave, &dataerr_ave, &datarms), SPEC_AVE, 0);
    */
    //printf("-----------------------\n");
    for (sel_num = sel_num_in; sel_num >= sel_num_thr; sel_num--) {
	*rej_num = count;

        // Averaging

	double data_ave;
        double dataerr_ave;
        double datarms;
        espda_util_func(espda_spec_ave(ref, data_col, dataerr_col, NULL,
            0, sel_num, &data_ave, &dataerr_ave, &datarms), SPEC_AVE, 0);

        // Find the farthest outlier
        double dev_max  = 0.0;
        int    j_max = INV_INT;
	//cpl_table_dump(spec, 0, 100, NULL);

	double *data_arr = cpl_table_get_data_double(spec, data_col);
	double *dataerr_arr = cpl_table_get_data_double(spec, dataerr_col);
	int count_old = count;
	cpl_table_unselect_all(spec);
	int sel_num_tmp = sel_num;
        for (int j = 0; j < sel_num_tmp; j++) {
	  //double dev     = fabs(data_ave - data_arr[j]) / sqrt(dataerr_ave *
          //                   dataerr_ave);
	  double dev     = fabs(data_ave - data_arr[j]) / sqrt(datarms *
                               datarms);
	  //printf("%e %e %e %e %e\n", data_arr[j], data_ave, dataerr_ave, datarms, dev);
	    if (sel_num <= sel_num_thr) break;
	    if (dev > kappa) {// && 1 == 1) {
		count++;
		//printf("caught!\n");
		sel_num--;
  	        cpl_table_or_selected_window(spec, j, 1);
	    }
        }
	//printf("after\n");
	//cpl_table_dump(spec, 0, 100, NULL);
	cpl_table_erase_selected(spec);
	if (count == count_old) {
	  //printf("break\n");
	    break;
	  //return CPL_ERROR_NONE;
	}
	//printf("dev_max %f\n", dev_max);

        // Stop iteration if the outlier is within kappa standard deviation
	/*
	if (dev_max <= kappa) {
            *rej_num = sel_num_in - sel_num;

            // Warning is commented to avoid a verbose printout
            //cpl_msg_info(cpl_func, "%i out of %i contributions rejected...",
            //             *rej_num, sel_num_in);
            return CPL_ERROR_NONE;
        }
	*/
        //cpl_table_unselect_all(spec);
        //cpl_table_or_selected_window(spec, j_max, 1);
        //cpl_table_erase_selected(spec);

	////cpl_table_erase_window(spec, j_max, 1);
    }

    // Compute number of rejected points
    *rej_num = sel_num_in - sel_num;
    *rej_num = count;

    // Warning is commented to avoid a verbose printout
    //cpl_msg_info(cpl_func, "%i out of %i contributions rejected...", *rej_num,
    //             sel_num_in);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to remove outliers from spectrum.");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Convolve a spectrum chunk with the instrument profile.
  @param   prof_chunk  The instrument profile in the spectrum chunk (INST_PROF).
  @param   wavel_over  Wavelengths of the spectrum over-chunk [nm].
  @param   fit_in      The Voigt profile in the spectrum over-chunk.
  @return  The convolved Voigt profile in the chunk.

  This function computes the pixel-by-pixel convolution of a Voigt profile
  @a fit_in fitted on a spectrum chunk with the instrument profile
  @a prof_chunk. The convolution uses a slightly-larger over-chunk, with
  wavelengths @a wavel_over, to avoid border effects.

  The convolution of the Voigt profile @f$ I @f$ and the instrument profile
  @f$ \Phi @f$ at the row @f$ r @f$ of the spectrum chunk is defined as @f[
    I_{r,\mathrm{conv}}=\sum_s\Phi_{r,s}I_s,
  @f]
  where @f$ r @f$ runs within the size of the chunk, while @f$ s @f$ runs within
  the width of the instrument profile. In practice, the convolution is computed
  at each row of the chunk within a wavelength range defined by the columns
  @a WAVELMIN and @a WAVELMAX of @a prof_chunk, while @f$ \Phi @f$ is an array
  entry from column @a PROFNORM of @a prof_chunk. If the range of the chunk is
  @f$ [\lambda_\mathrm{min},\lambda_\mathrm{max}] @f$, the range of the
  over-chunk is @f$ [\lambda_\mathrm{min}-\Delta\lambda_\mathrm{min},
  \lambda_\mathrm{max}+\Delta\lambda_\mathit{max}] @f$, where
  @f$ \Delta\lambda_\mathrm{min} @f$ (@f$ \Delta\lambda_\mathrm{max} @f$) is the
  absolute difference between @f$ \lambda_\mathrm{min} @f$
  (@f$ \lambda_\mathrm{max} @f$) and the first (last) entry of @a WAVELMIN
  (@a WAVELMAX) in @a prof_chunk.

  If @a prof_chunk, @a wavel_over, or @a fit_in are NULL, an error is issued.
*/
/*----------------------------------------------------------------------------*/

cpl_array *espda_spec_conv(cpl_table *prof_chunk, cpl_array *wavel_over,
    cpl_array *fit_in) {
    extern int v_level;
    espda_util_error(prof_chunk == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(wavel_over == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(fit_in     == NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_array *fit_out = NULL;
    cpl_size size_over = cpl_array_get_size(wavel_over);
    if (prof_chunk != NULL) {
        cpl_size size_chunk  = cpl_table_get_nrow(prof_chunk);
        fit_out              = cpl_array_new(size_chunk, CPL_TYPE_DOUBLE);
        cpl_table *temp_over = cpl_table_new(size_over);
        cpl_table_new_column(temp_over, "WAVEL", CPL_TYPE_DOUBLE);
        cpl_table_new_column(temp_over, "FIT",   CPL_TYPE_DOUBLE);
        cpl_table_copy_data_double(temp_over, "WAVEL",
                                   cpl_array_get_data_double(wavel_over));
        cpl_table_copy_data_double(temp_over, "FIT",
                                   cpl_array_get_data_double(fit_in));
        // Loop over rows of the output chunk
	double *wavel_min = cpl_table_get_data_double(prof_chunk, "WAVELMIN");
	double *wavel_max = cpl_table_get_data_double(prof_chunk, "WAVELMAX");
	int *ii_min = cpl_table_get_data_int(prof_chunk, "PROFMIN");
	int *ii_max = cpl_table_get_data_int(prof_chunk, "PROFMAX");
        for (int i = 0; i < (int)size_chunk; i++) {
	  //printf("%f %f\n", wavel_min[i], wavel_max[i]);
            double conv  = 0.0;

            // Select a slice of the input over-chunk as wide as the inst. prof.
            /*
	    double wavel_min = cpl_table_get_double(prof_chunk, "WAVELMIN", i,
                                                    NULL);
            double wavel_max = cpl_table_get_double(prof_chunk, "WAVELMAX", i,
                                                    NULL);
            int    ii_min    = cpl_table_get_int(prof_chunk, "PROFMIN", i,
                                                 NULL);
            int    ii_max    = cpl_table_get_int(prof_chunk, "PROFMAX", i,
                                                 NULL);
	    */
	    cpl_table_unselect_all(temp_over);
            cpl_table_or_selected_double (temp_over, "WAVEL",
                                          CPL_NOT_LESS_THAN, wavel_min[i]);
            cpl_table_and_selected_double(temp_over, "WAVEL",
                                          CPL_NOT_GREATER_THAN, wavel_max[i]);
            cpl_table *temp_sel = cpl_table_extract_selected(temp_over);
            cpl_size   size_sel = cpl_table_get_nrow(temp_sel);
            if (size_sel != ii_max[i] - ii_min[i]) {
                cpl_msg_warning(cpl_func, "Portion of fit (size = %i) not "
                                "compatible with instrument profile (size = "
                                "%i) at position %i!", (int)size_sel,
                                ii_max[i] - ii_min[i], i);
            }

            // Compute the convolution between the slice and the chunk
            for (int ii = ii_min[i]; ii < ii_max[i]; ii++) {
                conv += cpl_array_get_double(
                        cpl_table_get_array(prof_chunk, "PROFNORM", i),
                                            ii - ii_min[i], NULL) *
                        // If the vector is inverted, the convolution is not
                        // correct when the wavelength step is not constant.
                        //Investigate.
                        //cpl_table_get_double(temp_sel, "FIT", ii_max - ii - 1,
                        //                     NULL);
                        cpl_table_get_double(temp_sel, "FIT", ii - ii_min[i],
                                             NULL);
            }
            cpl_table_delete(temp_sel);

            // Save the convolution as a pixel of the output chunk
            cpl_array_set_double(fit_out, i, conv);
        }
        cpl_table_delete(temp_over);
    }
    else {
        fit_out = cpl_array_duplicate(fit_in);
    }
    return fit_out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Correct the spectral flux from HI absorption
  @param    spec_in   Input spectrum
  @param    lfit_in   Line list
  @param    zem       Emission redshift
  @param    spec_out  Output spectrum
  @return   CPL_ERROR_NONE iff OK

  This function estimates the unabsorbed flux level in the Lyman-alpha forest
  by integrating the optical depth contribution of lines below a given threshold
  in column density. The treshold is defined as the average column density from
  the list of fitted lines (assuming that all lines above this value have been
  properly fitted). The integral of the optical depth is used to compute a
  correction factor depending on redshift as follows:

     factor = exp(TAU_NORM * (1 + z)^TAU_INDEX * optical depth integral).

  A spectrum with the estimated unabsorbed flux is given in output.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_spec_corr(cpl_table  *spec_in, //cpl_table  *lfit,
    double zem, double colden_bar, int mode, cpl_table **spec_out) {
    cpl_errorstate  prestate = cpl_errorstate_get();

    if (spec_in != NULL) {
        cpl_size spec_size = cpl_table_get_nrow(spec_in);
        *spec_out          = cpl_table_duplicate(spec_in);


        double zprox = zem - (1.0 + zem) * VEL_PROX / LIGHTSPEED;

        double num = 1.0, den = 1.0;

        double fact_0 = 1.0;
        double fact_1 = 1e14 / log(1e14);
        double fact_2 = log(1e18) / log (1e14) * 1e5;
        if (colden_bar != 100) {
            /*
             cpl_propertylist *col = cpl_propertylist_new();
            cpl_propertylist_append_bool(col, "COLDEN", FALSE);
            cpl_table_sort(lfit, col);
            cpl_propertylist_delete(col);
            cpl_size lfit_size  = cpl_table_get_nrow(lfit);
            cpl_size lfit_hsize = (int)floor(0.5 * lfit_size);
            double colden_bar =
                0.5 * (cpl_table_get_double(lfit, "COLDEN", lfit_hsize, NULL) +
                       cpl_table_get_double(lfit, "COLDEN", lfit_hsize, NULL));
             */
            //double colden_bar = cpl_table_get_column_median(lfit, "COLDEN");
            if (colden_bar <= COLDEN_1) {
                num = 0;
            }
            if (colden_bar > COLDEN_1 && colden_bar <= COLDEN_2) {
                num = (pow(10, colden_bar * (2.0 + COLDEN_INDEX))  -
                       pow(10, COLDEN_1   * (2.0 + COLDEN_INDEX))) /
                                            (2.0 + COLDEN_INDEX);
            }
            if (colden_bar > COLDEN_2 && colden_bar <= COLDEN_3) {
                num = (pow(10, COLDEN_2   * (2.0 + COLDEN_INDEX))  -
                       pow(10, COLDEN_1   * (2.0 + COLDEN_INDEX))) /
                                            (2.0 + COLDEN_INDEX);
                num = num + fact_1                                 *
                      (pow(10, colden_bar * (1.0 + COLDEN_INDEX))) /
                                            (1.0 + COLDEN_INDEX)   *
                      (log(pow(10, colden_bar)) - 1.0 / (1.0 + COLDEN_INDEX));
                num = num - fact_1                                 *
                      (pow(10, COLDEN_2   * (1.0 + COLDEN_INDEX))) /
                                            (1.0 + COLDEN_INDEX)   *
                      (log(pow(10, COLDEN_2))   - 1.0 / (1.0 + COLDEN_INDEX));
            }
            if (colden_bar > COLDEN_3 && colden_bar <= COLDEN_4) {
                num = (pow(10, COLDEN_2   * (2.0 + COLDEN_INDEX))  -
                       pow(10, COLDEN_1   * (2.0 + COLDEN_INDEX))) /
                                            (2.0 + COLDEN_INDEX);
                num = num + fact_1                                 *
                      (pow(10, COLDEN_3   * (1.0 + COLDEN_INDEX))) /
                                            (1.0 + COLDEN_INDEX)   *
                      (log(pow(10, COLDEN_3))   - 1.0 / (1.0 + COLDEN_INDEX));
                num = num - fact_1                                 *
                      (pow(10, COLDEN_2   * (1.0 + COLDEN_INDEX))) /
                                            (1.0 + COLDEN_INDEX)   *
                      (log(pow(10, COLDEN_2))   - 1.0 / (1.0 + COLDEN_INDEX));
                num = num + fact_2                                 *
                      (pow(10, colden_bar * (1.5 + COLDEN_INDEX))  -
                       pow(10, COLDEN_3   * (1.5 + COLDEN_INDEX))) /
                                            (1.5 + COLDEN_INDEX);
            }
            if (colden_bar > COLDEN_4) {
                num = 1;
            }
            den = (pow(10, COLDEN_2 * (2.0 + COLDEN_INDEX))  -
                   pow(10, COLDEN_1 * (2.0 + COLDEN_INDEX))) /
                                      (2.0 + COLDEN_INDEX);
            den = den + fact_1 *
                  (pow(10, COLDEN_3 * (1.0 + COLDEN_INDEX))) /
                                      (1.0 + COLDEN_INDEX)   *
                  (log(pow(10, COLDEN_3))   - 1.0 / (1.0 + COLDEN_INDEX));
            den = den - fact_1 *
                  (pow(10, COLDEN_2 * (1.0 + COLDEN_INDEX))) /
                                      (1.0 + COLDEN_INDEX)   *
                  (log(pow(10, COLDEN_2))   - 1.0 / (1.0 + COLDEN_INDEX));
        }
	//printf("%f %f\n", num, den);
        for (int i = 0; i < spec_size; i++) {
            double zi   = cpl_table_get_double(*spec_out, "WAVEL", i, NULL) /
                          WAVEL_LY_A - 1.0;
            double corr;
            if (zi < zprox) {
                corr = exp(TAU_NORM * pow(1.0 + zi,    TAU_INDEX) * num / den)
                       * fact_0;
            }
            else if (zi < zem) {
                corr = (1.0 -
                       exp(TAU_NORM * pow(1.0 + zprox, TAU_INDEX) * num / den)
                       * fact_0) /
                       (zem - zprox) * (zi - zprox) +
                       exp(TAU_NORM * pow(1.0 + zprox, TAU_INDEX) * num / den)
                       * fact_0;
            }
            else {
                corr = 1.0;
            }
            //corr = 1.;
            if (mode == 1) {
                cpl_table_set_double(*spec_out, "FLUX",    i,
                    cpl_table_get_double(*spec_out, "FLUX",    i, NULL) * corr);
                cpl_table_set_double(*spec_out, "FLUXERR", i,
                    cpl_table_get_double(*spec_out, "FLUXERR", i, NULL) * corr);
            }
            else {
                cpl_table_set_double(*spec_out, "FLUX",    i,
                    cpl_table_get_double(*spec_out, "FLUX",    i, NULL) / corr);
                cpl_table_set_double(*spec_out, "FLUXERR", i,
                    cpl_table_get_double(*spec_out, "FLUXERR", i, NULL) / corr);
            }
	    //if (corr < 1) printf("%f\n", corr);
        }
    }

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to correct spectral flux from HI absorption.");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Equalize the flux in spectral orders and exposures.
  @param    fspec  A full spectrum (SPEC_FULL).
  @param    ref    ID number of the reference exposure.
  @return   The equalized full spectrum (SPEC_FULL)

  This function equalizes the flux in a spectrum @a fspec containing different
  exposures of the same object. The exposures are equalized order by order to
  one reference exposure @a ref.

  The exposures are selected using column @a SPECID of @a fspec. In each
  exposure, the spectral orders are selected using column @a ORDID. For each
  order @f$ o @f$ and exposure @f$ e\neq\mathit{ref} @f$, the equalization
  factor @f[
    f_{e,o}=\frac{\bar{I}_{e,o}}{\bar{I}_{\mathit{ref},o}}
  @f]
  is computed, @f$ \bar{I} @f$ being the error-weighted average of the
  flux density. The flux density and its error @f$ \sigma @f$ are then equalized
  as follows: @f[
    I_{e,o}^\mathrm{eq}=f_{e,o}I_{e,o},\qquad
      \sigma_{e,o}^\mathrm{eq}=f_{e,o}\sigma_{e,o}.
  @f]
  For details on the averaging, see @p espda_spec_ave.

  If @a fspec is NULL an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_spec_equal(cpl_table *fspec, int ref, double wavel_min,
			    double wavel_max) {
    extern int v_level;
    espda_util_error(fspec == NULL, CPL_ERROR_NULL_INPUT, NULL);


/*
 * The output spectrum is created and filled with values from the input
 * spectrum.
 */

    cpl_table_unselect_all(fspec);
    cpl_table_or_selected_double(fspec, "WAVEL", CPL_NOT_LESS_THAN, wavel_min);
    cpl_table_and_selected_double(fspec, "WAVEL", CPL_NOT_GREATER_THAN,
				  wavel_max);
    cpl_table *fspec_sel = cpl_table_extract_selected(fspec);

    cpl_table *fspec_eq = cpl_table_new(cpl_table_get_nrow(fspec_sel));
    espda_spec_full(fspec_eq);

    cpl_table_copy_data_double(fspec_eq, "WAVEL",
                               cpl_table_get_data_double(fspec_sel, "WAVEL"));
    cpl_table_copy_data_double(fspec_eq, "PIXSIZE",
                               cpl_table_get_data_double(fspec_sel, "PIXSIZE"));
    cpl_table_copy_data_double(fspec_eq, "FLUX",
                               cpl_table_get_data_double(fspec_sel, "FLUX"));
    cpl_table_copy_data_double(fspec_eq, "FLUXERR",
                               cpl_table_get_data_double(fspec_sel, "FLUXERR"));
    cpl_table_copy_data_int(fspec_eq, "SPECID",
                            cpl_table_get_data_int(fspec_sel, "SPECID"));
    cpl_table_copy_data_int(fspec_eq, "ORDID",
                            cpl_table_get_data_int(fspec_sel, "ORDID"));
    cpl_table_copy_data_int(fspec_eq, "QUALDATA",
                            cpl_table_get_data_int(fspec_sel, "QUALDATA"));
    cpl_table_delete(fspec_sel);
    if (ref != 0) {

/*
 * Flux in spectra is equalized order by order. The average flux of each order
 * is compared to the one from the reference spectrum and the resulting ratio is
 * used as an equalization factor (EF). FLUX and FLUXERR are equalized with the
 * formula
 *
 *     FLUX    (new) = FLUX    (old) * EF;
 *     FLUXERR (new) = FLUXERR (old) * sqrt(EF).
 *
 * The average EQF is computed and printed in the standard output for checking.
 */

    espda_print_verb(1, cpl_msg_info(cpl_func, "Equalizing flux..."));

    cpl_size spec_num = (int)cpl_table_get_column_max(fspec_eq, "SPECID") -
      (int)cpl_table_get_column_min(fspec_eq, "SPECID") + 1;
    cpl_size ord_num  = (int)cpl_table_get_column_max(fspec_eq, "ORDID") -
      (int)cpl_table_get_column_min(fspec_eq, "ORDID") + 1;
    //printf("%i %i %i\n", ord_num, (int)cpl_table_get_column_max(fspec_eq, "ORDID"),
    //   (int)cpl_table_get_column_min(fspec_eq, "ORDID"));
    double   eq_factor_ave[spec_num];

    // Create an array with the order boundaries in the table
    //espda_struct_sort(fspec_eq, 2, "ORDID", FALSE, "SPECID", FALSE);
    espda_struct_sort(fspec_eq, 2, "SPECID", FALSE, "ORDID", FALSE);
    cpl_array *ord_bound = cpl_array_new(1, CPL_TYPE_INT);
    cpl_array_set_int(ord_bound, 0, 0);
    int specid_old = cpl_table_get_int(fspec_eq, "SPECID", 0, FALSE);
    int ordid_old = cpl_table_get_int(fspec_eq, "ORDID", 0, FALSE);
    int ordnum_tot = 0;

    int *specid = cpl_table_get_data_int(fspec_eq, "SPECID");
    int *ordid = cpl_table_get_data_int(fspec_eq, "ORDID");

    for (int k = 1; k < cpl_table_get_nrow(fspec_eq); k++) {
      //if ((cpl_table_get_int(fspec_eq, "SPECID", k, FALSE) != specid_old) ||
      //(cpl_table_get_int(fspec_eq, "ORDID", k, FALSE) != ordid_old)) {
         if (specid[k-1] != specid[k] || ordid[k-1] != ordid[k]) {
	    ordnum_tot++;
	    //specid_old = cpl_table_get_int(fspec_eq, "SPECID", k, FALSE);
	    //ordid_old = cpl_table_get_int(fspec_eq, "ORDID", k, FALSE);
	    cpl_array_set_size(ord_bound, ordnum_tot + 1);
	    cpl_array_set_int(ord_bound, ordnum_tot, k);
	}
    }
    cpl_array_set_size(ord_bound, ordnum_tot + 2);
    cpl_array_set_int(ord_bound, ordnum_tot + 1, cpl_table_get_nrow(fspec_eq));

    // Loop over spectra, starting from reference spectrum
    //double flux_ref = 0.0;
    int l = -1;
    int lmod;
    //for (int i = 1; i <= (int)ord_num; i++) {
    double     flux_ref[ord_num];
    for (int j = ref; j < ref + (int)spec_num; j++) {
        double     eq_factor[ord_num];
        double     flux_ave[ord_num];
        double     null;
	int        jmod;

	double *eq_factor_all = cpl_table_get_data_double(fspec_eq, "EQFACT");

        if (j > (int)spec_num) {
            jmod = j % spec_num;
        }
        else {
            jmod = j;
        }
	eq_factor_ave[jmod-1] = 0;

	espda_print_verb(1,
			 cpl_msg_info(cpl_func, "Equalizing spectrum %i:",
				      jmod));

	// Loop over orders
	//cpl_array_dump(ord_bound, 0, 100, NULL);
        //for (int j = ref; j < ref + (int)spec_num; j++) {
        //for (int i = 1; i <= (int)ord_num; i++) {
	for (int i = 0; i < (int)ord_num; i++) {
            cpl_table *fspec_tmp = NULL;


	    //l = (jmod-1)*ord_num+i-1;
	    l = (jmod-1)*ord_num+i;
	    //printf("%i %i\n", l, ordnum_tot+1);
	    if (l > ordnum_tot) {
	        lmod = l % (ordnum_tot+1);
	    }
	    else {
	        lmod = l;
	    }
	    cpl_size start = cpl_array_get_int(ord_bound, lmod, NULL);
	    cpl_size count = cpl_array_get_int(ord_bound, lmod+1, NULL) - start;
            fspec_tmp = cpl_table_extract(fspec_eq, start, count);

            // EF for the reference spectrum
            if (jmod == ref) {
  	        /*
            espda_spec_ave(fspec_tmp, "FLUX", "FLUXERR", NULL, 0,
	      	               cpl_table_get_nrow(fspec_tmp) - 1, &flux_ref[i],
	      	               &null, &null);
            */
            flux_ref[i] = cpl_table_get_column_median(fspec_tmp, "FLUX");
		eq_factor[i] = 1.0;
            }

            // EF for the other spectra
            else {
	          /*
            espda_spec_ave(fspec_tmp, "FLUX", "FLUXERR", NULL, 0,
	      	               cpl_table_get_nrow(fspec_tmp) - 1, &flux_ave[i],
	      	               &null, &null);
            */
            flux_ave[i] = cpl_table_get_column_median(fspec_tmp, "FLUX");
                if (flux_ave[i] == 0) {
		    eq_factor[i] = 1.0;
                }
                else {
                    eq_factor[i] =  flux_ref[i] / flux_ave[i];
                }

                // Flux equalization
                cpl_table_multiply_scalar(fspec_tmp, "FLUX",    eq_factor[i]);
                cpl_table_multiply_scalar(fspec_tmp, "FLUXERR", eq_factor[i]);
            }
            cpl_table_fill_column_window_double(fspec_tmp, "EQFACT", 0,
                cpl_table_get_nrow(fspec_tmp), eq_factor[i]);

	    //printf("%f %f %f\n", flux_ref[i], flux_ave[i], eq_factor[i]);

            eq_factor_ave[jmod-1] = eq_factor_ave[jmod-1] + eq_factor[i] /
                                  (int)ord_num;
	    for (int h = start; h < start+count; h++) {
	        cpl_table_set_double(fspec_eq, "EQFACT", h, eq_factor[i]);
	        cpl_table_set_double(
		    fspec_eq, "FLUX", h,
		    cpl_table_get_double(fspec_eq, "FLUX", h, NULL) \
		    *eq_factor[i]);
	        cpl_table_set_double(
		    fspec_eq, "FLUXERR", h,
		    cpl_table_get_double(fspec_eq, "FLUXERR", h, NULL) \
		    *eq_factor[i]);
	    }
	    cpl_table_delete(fspec_tmp);

	    // Loop printout
	    ///*
	    if ((int)ord_num < 10) {
	        espda_print_verb(1, cpl_msg_info(cpl_func, " %-2i orders "
						 "equalized; %-2i left.", i+1,
						 (int)ord_num - i - 1));
	    }
	    else {
	        if (i % (int)round(ord_num/5.0) == 0 && i != ord_num) {
                    espda_print_verb(1, cpl_msg_info(cpl_func, " %-2i orders "
						     "equalized; %-2i left.", i+1,
						     (int)ord_num - i - 1));
		}
	    }
        }

	espda_print_verb(1, cpl_msg_info(cpl_func, "Average equalization "
					 "factor: %1.3f.", eq_factor_ave[jmod-1]));

    }
    }

    // Sort tables by ascending wavelength
    cpl_propertylist *col = cpl_propertylist_new();
    cpl_propertylist_append_bool(col, "WAVEL",  FALSE);
    //cpl_table_sort(fspec, col);
    cpl_table_sort(fspec_eq, col);
    cpl_propertylist_delete(col);

    return fspec_eq;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Fill a spectrum table column with data from a frame
  @param    set       Frameset
  @param    pos       Position of the frame in the frameset
  @param    size       Size of the frame
  @param    quantity  Physical quantity corresponding to a table column
  @param    table     Spectrum table (SPEC)
  @return   CPL_ERROR_NONE iff OK

  This functions copies the data from a frame with structure S1D or S2D into
  a column of a spectrum table. Different instructions are given for merged
  spectrum tables and order spectrum tables. Allowed columns are WAVEL, PIXSIZE,
  FLUX, FLUXERR, SPECID, ORDID (the last one just for order spectrum tables).

  NB: This function is still provisional. The description refers to the final
      version, which is targeted to handle the ESPRESSO DRS output; since the
      test datasets do not contain blaze frames or resolution frames,
      instructions about these quantities have not been included yet.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_spec_fill(cpl_frameset *set,
                                 cpl_size      pos,
                                 cpl_size      size,
                                 char         *quantity,
                                 cpl_table    *table)

{
    cpl_errorstate    prestate   = cpl_errorstate_get();

    // Load frame
    cpl_frame *frame = cpl_frame_duplicate(
                       cpl_frameset_get_position(set, pos));

    // Merged table
    if (cpl_frame_get_type(frame) == CPL_FRAME_TYPE_ANY) {
        cpl_vector *vector = NULL;
        // Spectrum #
        if (!strcmp(quantity, "SPECID")) {
	    for (int k = 0; k < (int)size; k++) {
                cpl_table_set_int(table, "SPECID", k, pos + 1);
	    }
        }

        // Wavelength and pixel size: from CRVAL1 and CDELT1
        else if (!strcmp(quantity, "WAVEL") || !strcmp(quantity, "PIXSIZE")) {
            double crval1, cdelt1;
            vector = cpl_vector_new(size);
            if (!strcmp(quantity, "WAVEL")) {
                espda_frame_desc_old(frame, "CRVAL1", &crval1);
            }
            espda_frame_desc_old(frame, "CDELT1", &cdelt1);
            for (int k = 0; k < (int)size; k++) {
                if (!strcmp(quantity, "WAVEL")) {
                    cpl_vector_set(vector, k, crval1 + k * cdelt1);
	        }
                else {
                    cpl_vector_set(vector, k, cdelt1);
	        }
            }
            cpl_table_copy_data_double(table, quantity,
                                       cpl_vector_get_data(vector));
        }

        // Flux and flux error: copied from extensions 0 and 1 of the frames
        else if (!strcmp(quantity, "FLUX") || !strcmp(quantity, "FLUXERR")) {
            if (!strcmp(quantity, "FLUX")) {
                vector = cpl_vector_load(cpl_frame_get_filename(frame), 0);
            }
            else {
                vector = cpl_vector_load(cpl_frame_get_filename(frame), 1);
            }
            cpl_table_copy_data_double(table, quantity,
                                       cpl_vector_get_data(vector));

        }
        cpl_vector_delete(vector);
    }

    // Order table
    else if (cpl_frame_get_type(frame) == CPL_FRAME_TYPE_IMAGE) {
        cpl_image *image = cpl_image_load(cpl_frame_get_filename(frame),
                                          CPL_TYPE_DOUBLE, 0, 0);
        cpl_size   size_x = cpl_image_get_size_x(image);
        cpl_size   size_y = cpl_image_get_size_y(image);
        if (size_x > size) {
            size_x = size;
        }
        if (size_y > size) {
            size_y = size;
        }

        // Spectrum #
        if (!strcmp(quantity, "SPECID")) {
	    for (int k = 0; k < (int)size; k++) {
                cpl_table_set_int(table, "SPECID", k, pos + 1);
	    }
	}

        // Order #
        else if (!strcmp(quantity, "ORDID")) {
            size_x = cpl_image_get_size_x(image);
            size_y = cpl_image_get_size_y(image);
            if (size_x > size) {
                size_x = size;
            }
            if (size_y > size) {
                size_y = size;
            }
	    for (int l = 0; l < (int)size_y; l++) {
                for (int k = 0; k < (int)size_x; k++) {
                    cpl_table_set_int(table, "ORDID", k + l * size_x, l + 1);
		}
	    }
	}

        // Pixel size: computed from wavelength frame
        else if (!strcmp(quantity, "PIXSIZE")) {
            double *double_arr = cpl_image_get_data_double(image);
            for (int l = 0; l < (int)size_y; l++) {
                for (int k = 0; k < (int)size_x; k++) {
                    double size_pix;
                    // Border pixels: twice the half distance from closest pixel
                    if (k == 0) {
                        size_pix = double_arr[k + 1 + l * size_x] -
                                   double_arr[k     + l * size_x];
                    }
                    else if (k == size_x - 1) {
                        size_pix = double_arr[k     + l * size_x] -
                                   double_arr[k - 1 + l * size_x];
                    }

                    // Other pixels: half distance between two closest pixels
                    else {
                        if (double_arr[k + 1 + l * size_x] <
                            double_arr[k     + l * size_x]) {
                            size_pix = double_arr[k     + l * size_x] -
                                       double_arr[k - 1 + l * size_x];
                        }
                        else {
                            size_pix = (double_arr[k + 1 + l * size_x] -
                                        double_arr[k - 1 + l * size_x]) * 0.5;
                        }
                    }
                    //size_pix = size_pix * 0.1;
                    cpl_table_set_double(table, "PIXSIZE", k + l * size_x,
                                         size_pix);
	        }
	    }
        }

        // Others (wavel, flux, and flux error): copied from respective frames
        else {
            double *double_arr = cpl_image_get_data_double(image);
            for (int l = 0; l < (int)size_y; l++) {
                for (int k = 0; k < (int)size_x; k++) {
                    cpl_table_set_double(table, quantity, k + l * size_x,
                                         double_arr[k + l * size_x]);
	        }
	    }
        }
        cpl_image_delete(image);
    }
    cpl_frame_delete(frame);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to fill column %s.", quantity);
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Rebin the flux in a spectrum.
  @param   fspec        A full spectrum (SPEC_FULL).
  @param   kappa        Number of standard deviations to clip flux outliers.
  @param   wavel_start  The starting wavelength [nm].
  @param   wavel_end    The ending wavelength [nm].
  @param   vel_step     The velocity step [km s^-1]
  @return  A rebinned spectrum (SPEC_RMS)

  This functions rebins the flux from a spectrum @a fspec into a new spectrum
  having a fixed wavelength grid with a range defined by @a wavel_start and
  @a wavel_end and a velocity step @a vel_step. The contributions to a row of
  the rebinned spectrum are computed from the pixels of @a fspec which overlap
  the row itself. The flux densities from column @a FLUX of @a fspec are
  weighted by their errors from column @a FLUXERR and by the overlap computed
  from column @a PIXSIZE; the average, error on average and RMS are saved in
  columns @a FLUX, @a FLUXERR, and @a FLUXRMS of the rebinned spectrum.
  Flux outliers are rejected via kappa-sigma clipping.

  The overlap @f$ f(\lambda) @f$ between the pixels @f$ p @f$ of @a fspec and
  the rows @f$ r @f$ of the rebinned spectrum is @f[
    f_r=\frac{
      \mathrm{min}(\lambda_p+\Delta\lambda_p/2,\lambda_r+\Delta\lambda_r/2)-
      \mathrm{max}(\lambda_p-\Delta\lambda_p/2,\lambda_r-\Delta\lambda_r/2)}
      {\Delta\lambda_r},
  @f]
  where @f$ \lambda @f$ is taken from column @a WAVEL and @f$ \Delta\lambda @f$
  from column @a PIXSIZE. This value is saved in a temporary column @a PIXFRAC
  and is used in averaging.
  For details on the averaging, see @p espda_spec_ave.
  For details on the clipping of outliers, see @a espda_spec_clean.

  If @a fspec is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_spec_rebin(cpl_table *fspec, double kappa, double wavel_start,
    double wavel_end, double vel_step) {
    //cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;
    espda_util_error(fspec == NULL, CPL_ERROR_NULL_INPUT, NULL);

/*
 * The sampling is checked against the spectrum size. If the ratio between the
 * wavelength range and the velocity step is less than one, the velocity step
 * is redefined.
 */

    //printf("%f %f %f %f\n", wavel_end, wavel_start, cpl_table_get_double(fspec, "WAVEL", 0, NULL), cpl_table_get_double(fspec, "WAVEL",
    //				     cpl_table_get_nrow(fspec) - 1, NULL));

    espda_print_verb(1, cpl_msg_info(cpl_func, "Rebinning spectrum to a "
        "velocity step of %6.1f km/s...", vel_step));

    if (wavel_start < cpl_table_get_double(fspec, "WAVEL", 0, NULL)) {
        wavel_start = cpl_table_get_double(fspec, "WAVEL", 0, NULL);
    }
    if (wavel_end > cpl_table_get_double(fspec, "WAVEL",
					 cpl_table_get_nrow(fspec) - 1, NULL)) {
        wavel_end = cpl_table_get_double(fspec, "WAVEL",
				         cpl_table_get_nrow(fspec) - 1, NULL);
    }
//printf("%f %f %f %f\n", wavel_end, wavel_start, cpl_table_get_double(fspec, "WAVEL", 0, NULL), cpl_table_get_double(fspec, "WAVEL",
    //			     cpl_table_get_nrow(fspec) - 1, NULL));
    /*
    double wavel_min, wavel_max;
    if (wavel_end > wavel_start) {
        wavel_min = wavel_start;
        wavel_max = wavel_end;
    }
    else {
        wavel_min = wavel_end;
        wavel_max = wavel_start;
    }
    espda_struct_sort(fspec, 1, "WAVEL", FALSE);
    */
/*
 * The output spectrum table is created. The wavelength grid is the same of the
 * S1D spectra produced by the DRS (see espda_utils.h). Temporary columns are
 * added to the input spectrum table to store information needed to rebin
 * spectra, i.e. half pixel size and start/end wavelength for all rows.
 */

    cpl_size row_num_in  = cpl_table_get_nrow(fspec);
    cpl_size row_num_out;
    if (wavel_end > wavel_start) {
        row_num_out = (int)ceil(LIGHTSPEED / vel_step * log(wavel_end /
            wavel_start));
    }
    else {
        row_num_out = (int)ceil(LIGHTSPEED / vel_step * log(wavel_start /
            wavel_end));
    }
    cpl_table *rspec = cpl_table_new(row_num_out);
    if (!cpl_table_has_column(fspec, "CONT")) {
        espda_spec_rms(rspec);
    }
    else {
        espda_spec_fit(rspec);
    }
    cpl_table_duplicate_column(fspec, "VELSTEP",  fspec, "PIXSIZE");
    cpl_table_duplicate_column(fspec, "PIXSTART", fspec, "WAVEL");
    cpl_table_duplicate_column(fspec, "PIXEND",   fspec, "WAVEL");
    cpl_table_divide_columns  (fspec, "VELSTEP", "WAVEL");
    cpl_table_multiply_scalar (fspec, "VELSTEP", LIGHTSPEED);
    for (int j = 0; j < row_num_in; j++) {
        double wavel_in    = cpl_table_get_double(fspec, "WAVEL",   j, NULL);
        double vel_step_in = cpl_table_get_double(fspec, "VELSTEP", j, NULL);
        cpl_table_set_double(fspec, "PIXSTART", j, wavel_in *
            exp(-0.5 * vel_step_in / LIGHTSPEED));
        cpl_table_set_double(fspec, "PIXEND",   j, wavel_in *
            exp( 0.5 * vel_step_in / LIGHTSPEED));
    }

/*
 * The contributions to the rows of the output spectrum table are computed and
 * the rows are filled. An input row contributes to an output row if there is an
 * overlap between the two. The size of the overlap is used to weight the
 * contributions. Contributing rows are cleaned to remove outliers via
 * kappa-sigma clipping. The final weighted averages of the flux and its error,
 * together with the RMS of the contributing rows, are used to fill the output
 * spectrum table. Temporary columns are erased from the input spectrum table.
 */

    espda_print_verb(1, cpl_msg_info(cpl_func, "Compute the contributions and "
        "fill the rows..."));
    cpl_size rej_num_tot = 0;
    cpl_size sel_num_tot = 0;

    double start_offset  = -1.5;  //-1.5
    double end_offset    = -0.5;  //-0.5
    double center_offset = 0.5;   //0.5
    if (wavel_end <= wavel_start) {
        start_offset  = 0.0;     //-0.5
        end_offset    = 1.0;      //0.5
        center_offset = -0.5;     //-0.5
        vel_step      = -vel_step;
    }
    double pix_start_out = wavel_start * exp(start_offset * vel_step /
        LIGHTSPEED);
    double pix_end_out   = wavel_start * exp(end_offset   * vel_step /
        LIGHTSPEED);
    //double pix_size      = 0;
    double pix_snr_in    = 0;
    double pix_snr_out   = 0;
    double pix_num_ave   = 0;

    // Loop over output rows
    int row_start = 0, row_end = row_num_in;

    double exp_vel_step = exp(vel_step / LIGHTSPEED);
    double exp_center_offset = exp(center_offset * vel_step / LIGHTSPEED);

    espda_struct_sort(fspec, 1, "WAVEL", FALSE);
    for (int i = 0; i < (int)row_num_out; i++) {
      //printf("%i %i\n", i, row_num_out);

        // Selection of contributing input rows
        double flux_out, fluxerr_out, fluxrms;
        if (wavel_end > wavel_start) {
            pix_start_out = pix_end_out;
            pix_end_out   = pix_start_out * exp_vel_step;
        }
        else {
            pix_end_out   = pix_start_out;
            pix_start_out = pix_end_out   * exp_vel_step;
        }
        double pix_size = pix_end_out - pix_start_out;

        cpl_table_set_double(rspec, "WAVEL", i,
			     pix_start_out * exp_center_offset);
        cpl_table_set_double(rspec, "PIXSIZE", i, pix_size);

	cpl_table_unselect_all(fspec);
        int j;
        if (wavel_end > wavel_start) {
	  /*
	    for (j = row_start; j < row_num_in && cpl_table_get_double(fspec,
                "PIXEND", j, NULL) < pix_start_out; j++) {
                row_start = j+1;
	    }
	  */
            for (j = row_end - 1; j >= 0 && cpl_table_get_double(fspec,
                "PIXEND", j, NULL) > pix_start_out; j--) {
	        row_start = j;
            }
            for (j = row_start; j < row_num_in && cpl_table_get_double(fspec,
                "PIXSTART", j, NULL) < pix_end_out; j++) {
                row_end = j+1;
            }
            // Case when there are no contributions
	    if (row_start == 0 && row_end == row_num_in) {
	        row_end = 0;
	    }
	    if (row_start > row_num_in && row_end == row_num_in) {
     ;
	    }
	    //printf("%i %i %i %f %f\n", row_start, row_end, row_num_in,
	       // cpl_table_get_double(fspec, "PIXSTART", j, NULL),
	       // cpl_table_get_double(fspec, "PIXEND", j, NULL),
	    //	 pix_start_out, pix_end_out);
	//cpl_table_dump(fspec, row_start, 1, NULL);
	//cpl_table_dump(fspec, row_end, 1, NULL);
        }
        else {

	  /*
            for (j = row_end - 1; j >= 0 && cpl_table_get_double(fspec,
                "PIXSTART", j, NULL) > pix_end_out; j--) {
                row_end = j;
            }
	  */
	  //	  printf("%i %i %i\n", row_start, row_end, row_num_in);
            for (j = row_start; j < row_num_in && cpl_table_get_double(fspec,
                "PIXSTART", j, NULL) < pix_end_out; j++) {
                row_end = j + 1;
            }
            for (j = row_end - 1; j >= 0 && cpl_table_get_double(fspec,
                "PIXEND", j, NULL) > pix_start_out; j--) {
                row_start = j;
            }
            // Case when there are no contributions
	    if (row_start == row_num_in && row_end == 0) {
	        row_end = row_num_in;
	    }
        }

        cpl_size sel_num = 0;
	if (row_end - row_start >= 0) {
            cpl_table_or_selected_window(fspec, row_start, row_end - row_start);
            sel_num = cpl_table_count_selected(fspec);
        }
	sel_num = row_end - row_start;


        // If contributing rows are found
        if ((int)sel_num > 0) {


	  //cpl_table *fspec_tmp  = cpl_table_extract_selected(fspec);
	  cpl_table *fspec_tmp  = cpl_table_extract(fspec, row_start,
						    row_end - row_start);
            cpl_table_new_column(fspec_tmp, "PIXFRAC", CPL_TYPE_DOUBLE);

            // Kappa-sigma clipping
            cpl_size rej_num = 0;
            //espda_util_func(espda_spec_clean(fspec_tmp, "FLUX", "FLUXERR",
            //    kappa, &rej_num), SPEC_CLEAN, 0);
	    cpl_size row_start_ref = espda_max(0, row_start-2);
	    cpl_size row_end_ref = espda_min(cpl_table_get_nrow(fspec),
					     row_end+2);
	    cpl_table *fspec_ref = cpl_table_extract(fspec, row_start_ref,
	    					     row_end_ref-row_start_ref);
	    //cpl_table *fspec_ref = cpl_table_duplicate(fspec_tmp);
	    //printf("before %i\n", cpl_table_get_nrow(fspec_tmp));
	    sel_num_tot += cpl_table_get_nrow(fspec_tmp);
	    espda_spec_clean(fspec_tmp, fspec_ref, "FLUX", "FLUXERR", kappa,
	    		     &rej_num);
	    //printf("after %i\n", cpl_table_get_nrow(fspec_tmp));
	    //printf("%i %i\n", rej_num, row_end-row_start);
            rej_num_tot += rej_num;
            if ((int)rej_num == -1) {
                cpl_msg_warning(cpl_func, "Contributions to row %i cannot be "
                                "averaged!", i + 1);
            }

            // Loop over non-rejected rows
	    //int pixfrac_new = 0;
            for (int j = 0; j < (int)sel_num - (int)rej_num; j++) {
                // Compute overlap
	        double pix_start_in = cpl_table_get_double(fspec_tmp, "PIXSTART",
                                                           j, NULL);
	        double pix_end_in   = cpl_table_get_double(fspec_tmp, "PIXEND",
                                                           j, NULL);
		double pix_size_in = pix_end_in - pix_start_in;
                double pixfrac = INV_DOUBLE;
                double pixnorm = INV_DOUBLE;
                if (pix_start_in <= pix_start_out &&
                    pix_end_in   >= pix_start_out &&
                    pix_end_in   <= pix_end_out) {
		  pixfrac = (pix_end_in  - pix_start_out) / pix_size;
		    pixnorm = (pix_end_in  - pix_start_out) / pix_size_in;
                }
                if (pix_start_in <= pix_end_out &&
                    pix_end_in   >= pix_end_out &&
                    pix_start_in >= pix_start_out) {
		  pixfrac = (pix_end_out - pix_start_in)  / pix_size;
		    pixnorm = (pix_end_out - pix_start_in) / pix_size_in;
                }
                if (pix_start_in >= pix_start_out &&
                    pix_end_in   <= pix_end_out) {
		  pixfrac = (pix_end_in  - pix_start_in)  / pix_size;
		    pixnorm = 1.0;
                }
                if (pix_start_in <= pix_start_out &&
                    pix_end_in   >= pix_end_out) {
		  pixfrac = (pix_end_out - pix_start_out) / pix_size;
		    pixnorm = (pix_end_out - pix_start_out) / pix_size_in;
                }
                //cpl_table_set_double(fspec_tmp, "PIXFRAC", j, pixfrac);
                cpl_table_set_double(fspec_tmp, "PIXFRAC", j, pixnorm);

                pix_snr_in  += pixnorm *
                    cpl_table_get_double(fspec_tmp, "FLUX",    j, NULL) /
                    cpl_table_get_double(fspec_tmp, "FLUXERR", j, NULL);
                pix_num_ave += pixnorm;
            }

            // Averaging
            espda_spec_ave(fspec_tmp, "FLUX", "FLUXERR", "PIXFRAC", 0,
			   sel_num - rej_num, &flux_out, &fluxerr_out,
			   &fluxrms);


	    // Propagating QUALDATA
	    //int qualdata = (int)(cpl_table_get_column_mean(fspec_tmp, "QUALDATA") \
	    //		   * (sel_num - rej_num));

	    // Identify unique elements in QUALDATA
	    // from https://stackoverflow.com/questions/30577716/find-distinct-numbers-from-an-array-in-c
	    int len = (int)sel_num - (int)rej_num;
	    //int w[len];
	    int *w = cpl_calloc(len, sizeof(int));
	    int qualdata = 0;
	    int m = 0;
      int *v = cpl_table_get_data_int(fspec_tmp, "QUALDATA");
	    for (int j = 0; j < len; j++) {
	        int k;
	        for(k = 0; k < m; k++) {
		    if(v[j]==w[k]) {
		        break;
		    }
		}
		if (k == m) {
                    w[m] = v[j];
		    qualdata = qualdata + v[j];
		    m++;
		}
	    }

            // Averaging failure printout
	    if (flux_out == INV_DOUBLE && fluxerr_out == INV_DOUBLE) {
	      /*
		    cpl_msg_warning(cpl_func, "Row %i of %i (approximate "
                                "wavelength %7.3f) has invalid flux and "
                                "flux error!", i + 1, (int)row_num_out,
                                (pix_start_out + pix_end_out) * 0.5);
	      */
                cpl_table_set_int(rspec, "MASK", i, 1);
                cpl_table_set_invalid(rspec, "WAVEL", i);
            }
	    else {
	      /*    cpl_msg_info(cpl_func, "Row %i of %i (approximate "
                                "wavelength %7.3f) has valid flux and "
                                "flux error!", i + 1, (int)row_num_out,
                                (pix_start_out + pix_end_out) * 0.5)*/;
	    }

            cpl_table_delete(fspec_tmp);

            // Fill the output spectrum table
            cpl_table_set_double(rspec, "FLUX",    i, flux_out);
            cpl_table_set_double(rspec, "FLUXERR", i, fluxerr_out);
            cpl_table_set_double(rspec, "FLUXRMS", i, fluxrms);
            cpl_table_set_int(rspec, "QUALDATA", i, qualdata);

            pix_snr_out += flux_out / fluxerr_out;
        }

        // If contributing rows are not found
        else {
            cpl_table_set_int    (rspec, "MASK",  i, 0);
            cpl_table_set_invalid(rspec, "WAVEL", i);
        }


        // Loop printout
        if (row_num_out < 10) {
            espda_print_verb(1, cpl_msg_info(cpl_func, " %-*i rows filled; %-*i "
                "left.", (int)ceil(log10(row_num_out)), i + 1,
                (int)ceil(log10(row_num_out)), (int)row_num_out - i - 1));
        }
        else {
            if ((i + 1) % (int)ceil(row_num_out/10.0) == 0 && i < row_num_out) {
                espda_print_verb(1, cpl_msg_info(cpl_func, " %-*i rows filled; "
                    "%-*i left.", (int)ceil(log10(row_num_out)), i + 1,
                    (int)ceil(log10(row_num_out)), (int)row_num_out - i - 1));
            }
        }
    }
    //cpl_table_dump(fspec, 3600, 50, NULL);

    if (cpl_table_get_nrow(rspec) == 1 &&
        !cpl_table_is_valid(rspec, "WAVEL", 0)) {
        rspec = NULL; //cpl_table_erase_window(rspec, 0, 1);
    }
    else {
        cpl_table_erase_invalid(rspec);
    }
    pix_snr_in  = pix_snr_in  / (row_num_in - rej_num_tot);
    pix_snr_out = pix_snr_out /  row_num_out;
    pix_num_ave = pix_num_ave /  row_num_out;
//    if (cpl_table_get_nrow(rspec) > 0) {
    if (rspec != NULL) {
//    if (rspec != NULL && cpl_table_get_nrow(rspec) > 0) {
        double fluxerr_ave = cpl_table_get_column_mean(rspec, "FLUXERR");
        double fluxrms_ave = cpl_table_get_column_mean(rspec, "FLUXRMS");

        // Rejection printout
        espda_print_verb(1,
	    cpl_msg_info(cpl_func, "%i/%i contributions rejected "
			 "(%1.0f percent).", (int)rej_num_tot, (int)sel_num_tot, //(int)row_num_in,
			 100.0 * (int)rej_num_tot / (int)sel_num_tot)); //(int)row_num_in));
        espda_print_verb(1,
	    cpl_msg_info(cpl_func, "Average input SNR per pixel: "
            "%f; average output SNR per pixel: %f. ", pix_snr_in, pix_snr_out));
        espda_print_verb(1,
	    cpl_msg_info(cpl_func, "Number of input pixels: %i, "
            "number of output pixels: %i, average contributions per output "
            "pixel: %f;", (int)row_num_in - (int)rej_num_tot, (int)row_num_out,
            pix_num_ave));
        espda_print_verb(1, cpl_msg_info(cpl_func, "average flux error per "
            "output pixel: %f, average flux RMS per output pixel: %f.",
            fluxerr_ave, fluxrms_ave));
    }
    else {
        cpl_msg_warning(cpl_func, "Rebinned spectrum is empty!");
    }

    //cpl_table_erase_column(fspec, "HPIXSIZE");
    cpl_table_erase_column(fspec, "VELSTEP");
    cpl_table_erase_column(fspec, "PIXSTART");
    cpl_table_erase_column(fspec, "PIXEND");

    return rspec;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Remove lines from a spectrum.
  @param   spec        A fitted spectrum (SPEC_FIT)
  @param   peak_thres  The threshold for truncating line peaks.
  @return  The spectrum with lines removed (SPEC_FIT)

  This function takes a spectrum @a spec and removes the fitted spectral lines
  from it by dividing the normalized flux in column @a NORM by the fitted
  profile in column @a FIT. Results are saved in column @a FLUX of the newly
  created spectrum. The flux density below @a peak_thres is truncated to avoid
  large fluctuations in division when @a FIT approaches 0.

  If @a spec is NULL, an error is issued.

  N.B. Currently the function is called before column @a NORM of @a spec is
       filled, so the normalized flux is computed by dividing column @a FLUX by
       column @a CONT.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_spec_remove(cpl_table *spec, double peak_thres) {
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_table *spec_temp = cpl_table_duplicate(spec);
    cpl_table_divide_columns  (spec_temp, "FLUX", "CONT");
    cpl_table_unselect_all(spec_temp);
    cpl_table_or_selected_double(spec_temp, "FIT", CPL_GREATER_THAN,
        peak_thres);
    cpl_table *spec_rem = cpl_table_extract_selected(spec_temp);

    // Line removal by division
    /*
    cpl_table_divide_columns  (spec_rem, "FLUX", "FIT" );
    */

    // Line removal by subtraction
    cpl_table_multiply_scalar(spec_rem, "FIT", -1);
    cpl_table_add_scalar(spec_rem, "FIT", 1);
    cpl_table_add_columns(spec_rem, "FLUX", "FIT");
    cpl_table_add_scalar(spec_rem, "FIT", -1);
    cpl_table_multiply_scalar(spec_rem, "FIT", -1);

    cpl_table_multiply_columns(spec_rem, "FLUX", "CONT");
    cpl_table_delete(spec_temp);
    return spec_rem;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Shift the wavelengths to the rest frame.
  @param   spec  A full spectrum (SPEC_FULL).
  @param   rv    Radial velocities for the exposures.
  @return  CPL error code.

  This function shifts the wavelengths of a stellar spectrum @a spec to the rest
  frame, using the radial velocities @a rv for the different exposures in the
  spectrum.

  For each exposure @f$ e @f$, the shifted wavelengths
  @f$ \lambda^\mathrm{RF} @f$ are computed from the original wavelengths
  @f$ \lambda @f$ as follows: @f[
    \lambda_e^\mathrm{RF}=\lambda_e\left(1+\frac{\mathit{rv}_e}{c}\right)
  @f]
  @f$ c @f$ being the speed of light.

  If @a spec or @a rv are NULL or the number of elements in @a rv is not equal
  to the number of exposures, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_spec_shift(cpl_table *spec, cpl_array *rv) {
    extern int v_level;
    espda_util_error(spec == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(rv   == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());


/*
 * Spectrum and RV array are loaded. If the number of RV values is not equal to
 * the number of observations in the spectrum table, the function is terminated.
 */

    cpl_size row_num = cpl_table_get_nrow(spec);
    cpl_size rv_num  = cpl_array_get_size(rv);
    espda_util_error((rv_num != (int)cpl_table_get_column_max(spec, "SPECID")),
        CPL_ERROR_INCOMPATIBLE_INPUT, cpl_error_get_code());

/*
 * Wavelengths are corrected. The RV array is scanned; each value is used to
 * correct the wavelengths of the corresponding observation, identified by the
 * column SPECID in the table.
 */

    for (int i = 0; i < (int)rv_num; i++) {
        double rv_i = cpl_array_get_double(rv, i, NULL);
        for (int k = 0; k < (int)row_num; k++) {
            int specid = cpl_table_get_int(spec, "SPECID", k, NULL);
            if (specid == i + 1) {
	      //printf("%f\n", rv_i);
                double corr  = (1.0 - rv_i / LIGHTSPEED);
                double wavel = cpl_table_get_double(spec, "WAVEL", k, NULL)
                    * corr;
                cpl_table_set_double(spec, "WAVEL", k, wavel);
            }
        }
        espda_print_verb(1, cpl_msg_info(cpl_func, " Observation %2i; RV: "
            "%3.1f.", i + 1, rv_i));
    }
    return cpl_error_get_code();
}

/**@}*/
