/* $Id: espda_synth_ccf.c 2018-09-08 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2018-09-08 14:15:10 +0000 (sat, 24 feb 2017) $
 * $Revision: 202524 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_synth_ccf.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"
#include <gsl/gsl_version.h>

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_synth_ccf
 *
 * Functions for cross-correlation of merged spectrum with synthetic spectrum
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Cross-correlate merged spectrum with synthetic spectrum
  @param    wobs     Array of wavelength of the observed spectrum
  @param    fobs     Array of flux of the observed spectrum
  @param    fsyn     Array of flux of the synthetic spectrum
  @param    fmsk     Mask values at the observed spectrum
  @param    wcor     Array of RV-corrected wavelength of the observed spectrum
  @param    fres     Flux residuals of observed minus synthetic spectra
  @param    xccf     x-axis values of CCF in RV [km/s]
  @param    yccf     y-axis values of CCF
  @param    pccf     parameters of the fit of CCF
  @param    eccf     errors on the parameters of the fit of the CCF
  @param    pccfc    parameters of the fit of CCF core
  @param    eccfc    errors on the parameters of the fit of the CCF core
  @param    ygfit    y-axis of the fitting function of the CCF
  @param    rvccf    First guess RV value [km/s] to center the CCF
  @param    vwccf    CCF window in RV [km/s]
  @return   CPL_ERROR_NONE iff OK

  @doc      
  This function cross-correlate the merged spectrum with the synthetic spectrum
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_ccf_fit_synth(cpl_array     *wobs,
			                       cpl_array     *fobs,
			                       cpl_array     *fsyn,
			                       cpl_array     *fmsk,
			                       cpl_array     *wcor,
			                       cpl_array     *fres,
			                       cpl_array    **xccf,
			                       cpl_array    **yccf,
				                   cpl_array     *pccf,
			                       cpl_array     *eccf,
				                   cpl_array    *pccfc,
			                       cpl_array    *eccfc,
				                   cpl_array   **ygfit,
				                   cpl_array **ygfcore,
                                   double        rvccf,
                                   double        vwccf,
                                   int            fitg,
                                   int             frv)

{
    cpl_errorstate prestate = cpl_errorstate_get();

/*
 * This function creates the cross-correlation function of the observed   *
 * merged spectrum with the interpolated synthetic spectrum both with the *
 * the same wavelength array in logaritmic scale. This function calls     *
 * the function espda_ccf_fit_gauss which fits using a gaussian model     *
 * function and computes the central RV of the CCF and its RV error       *
*/

    // Defining the number of elements of the arrays

    cpl_size nobs = cpl_array_get_size (wobs);

    //Deriving the number of mask points:

    int nrj= 0;

    for (int j = 0; j < (int)nobs; j++) {
       int cmsx = cpl_array_get_int (fmsk, j, NULL);
       if ((int)cmsx == 0) {
         nrj++;
       }
     }

    // We compute the initial shift in pixels of the synthetic spectrum
    // according to the central RV of the ccf provide as input

    // Dispersion in km/s
    double vlight = 299792.458 ;

    //    double wini = cpl_array_get_double (wobs, 0, NULL);
    //    int    nfin = (int)nobs
    //    double wfin = cpl_array_get_double (wobs, nfin, NULL);
    //    double wmid = SQRT (w2a * w1a) ;

    double w1 = cpl_array_get_double (wobs, 5, NULL);
    double w2 = cpl_array_get_double (wobs, 6, NULL);
    double vdw = (w2 - w1) / w1 * vlight ;

    printf("Number of points of observed spectrum: %i \n", (int)nobs);
    printf("Wavelength dispersion in km/s : %f \n", vdw);

    // We search for the pixel to center the ccf RV window
    //pixel location of the first guess RV
    //We choose -RV because the x-axis of CCF is inverted due to
    //the fact that we perform SUM [fobs(i)*fsyn(i+d)]
    double vp = -rvccf / vdw ;

    int vpi = (int)vp ;
    int vpt = vpi ;
    if (vp >= 0) {
      vpt = (int)vp ;
    } else {
      vpt = -(int)vp ;
    }
    
    //size of CCF window in pixels
    double wp = vwccf / vdw ;
    int wpi = (int)wp ;

    //double size of CCF window in pixels (+-window size)
    int wpt = (int)wpi * 2 ;

    // We redefine the observed and synthetic arrays for cross-correlation
    // using the window size

    //int nob = (int)nobs - (int)wpt - (int)wpt ;
    //Excluding the mask points
    int nob = (int)nobs - (int)wpt - (int)wpt -(int)nrj;
    cpl_array *fos = cpl_array_new (nob, CPL_TYPE_DOUBLE);
    for (int j = 0; j < (int)nob; j++) {
      int i = (int)j + (int)wpt ;
      double cfobs = cpl_array_get_double (fobs, i, NULL);
      cpl_array_set_double (fos, j, cfobs);
    }

    // We perform the cross-correlation
 
    //cpl_array *fss = cpl_array_new (nob, CPL_TYPE_DOUBLE);
    cpl_array *fcc = cpl_array_new (wpt, CPL_TYPE_DOUBLE);
    cpl_array *nfcc = cpl_array_new (wpt, CPL_TYPE_DOUBLE);
    cpl_array *xcc = cpl_array_new (wpt, CPL_TYPE_DOUBLE);
    cpl_array *yfit = cpl_array_new (wpt, CPL_TYPE_DOUBLE);
    cpl_array *bfit = cpl_array_new (wpt, CPL_TYPE_DOUBLE);

    double cfcc2 = 0 ;

    for (int i = 0 ; i < (int)wpt; i++) { 
      for (int k = 0  ; k < (int)nob; k++) {

	//Computation of index for the SYNTH spectrum:
	//- INDEX across CCF whole window (0,Npoint(CCF)=wpt)
        //this is twice the input RV window
	//- INDEX of the point in which the CCF is calculated
	//   (0,Npoint(region of Observed spectrum)=nob)
	//   nob=nobs-wpt
	//- Size of the RV window = wpi
	//- Twice the size RV window = wpt
	//- Initial point to center the CCF from the initial 
	//   guess RV = vpi

        int j = (int)i + (int)k - (int)wpi + (int)wpt + (int)vpi ;

        double cfo  = cpl_array_get_double (fos, k, NULL);
        if (cfo > 0) {
        double cfss = cpl_array_get_double (fsyn, j, NULL);
        double cfos  = cpl_array_get_double (fos, k, NULL);
        double cfcc1 = cpl_array_get_double (fcc, i, NULL);
        cfcc2 = cfcc1 + (cfos-1.) * (cfss-1.) ;
        cpl_array_set_double (fcc, i, cfcc2);
        }    
      }
      double ci = (double)i ;
      double cxcc = ci * vdw + ((double)vpi * vdw) - ((double)wpi * vdw)  ;

     //We change the sign of the x-axis of CCF because it appears inverted
     //since we perform SUM [fobs(i)*fsyn(i+d)]
      double cxccf = -cxcc  ;

      cpl_array_set_double (xcc, i, cxccf);
    }

    // Fitting a gaussian to the CCF

    int npar = 6;
    cpl_array *pfit = cpl_array_new (npar, CPL_TYPE_DOUBLE);
    cpl_array *efit = cpl_array_new (npar, CPL_TYPE_DOUBLE);

    //Instead of using the initial guess for gaussian fitting of the CCF
    //we use the RV at maximum value of the CCF

    cpl_size irvm = 0;
    cpl_array_get_maxpos (fcc, &irvm); 
    double rvmax =  cpl_array_get_double (xcc, (int)irvm, NULL); 
    double fccmax =  cpl_array_get_double (fcc, (int)irvm, NULL);
    double fccmin = cpl_array_get_min (fcc);
    printf("RV(MAX) iMAX FCC(iMAX) FCC(iMIN) : %f %i %f %f \n", rvmax, (int)irvm, fccmax, fccmin);

    //We initialize the rvinit
    double rvinit = 0 ;

    //We evaluate the reliability of the rvmax

    if (rvmax < -(double)vwccf+10. || rvmax > +(double)vwccf-10.) {

    	//We search for the peak in a short RV window
    	double rvw = vwccf/2. ;
    	int nm = 0 ;
    	int np = 0 ;

    	for (int i = 0 ; i < (int)wpt; i++) {
    	double vpp = (double)i * vdw ;
    	if (vpp < -(double)rvw) {
    		nm = i;
    	}
    	if (vpp < +(double)rvw) {
    		np = i;
    	}
    	}
    	int nv = np - nm ;
    	cpl_array *xccv = cpl_array_new (nv, CPL_TYPE_DOUBLE);
    	cpl_array *fccv = cpl_array_new (nv, CPL_TYPE_DOUBLE);

    	int j = 0;
    	for (int i = 0 ; i < (int)nv; i++) {
    	j = nm + i ;
    	double cxcc = cpl_array_get_double (xcc, j, NULL);
    	cpl_array_set_double (xccv, i, cxcc);
    	double cfcc = cpl_array_get_double (fcc, j, NULL);
    	cpl_array_set_double (fccv, i, cfcc);
    	}
    	cpl_size irvmv = 0;
    	cpl_array_get_maxpos (fccv, &irvmv);
    	double rvmaxv =  cpl_array_get_double (xccv, (int)irvmv, NULL);
    	rvinit = rvmaxv ;

        //We delete the new created arrays to find the maximum of the CCF

        cpl_array_delete(xccv);
        cpl_array_delete(fccv);

    } else {
        rvinit = rvmax ;
    }

    //double fccmax = cpl_array_get_max (fcc);
    //double fccmin = cpl_array_get_min (fcc);

    //printf("FCC(iMIN) & Npoints (unmasked): %f %i \n", fccmin, nob);

    ////We subtract its minimum value to the CCF
    //double fccmin = cpl_array_get_min (fcc);
    //for (int j = 0; j < (int)wpt; j++) {
    //  double cfcc = cpl_array_get_double (fcc, j, NULL);
    //  double bfcc = cfcc - fccmin ;
    //  cpl_array_set_double (nfcc, j, bfcc);
    //}
    ////We normalize the CCF to its maximum value
    //double bfccmax = cpl_array_get_max (nfcc);
    //for (int j = 0; j < (int)wpt; j++) {
    //  double cfcc = cpl_array_get_double (nfcc, j, NULL);
    //  double mfcc = cfcc / bfccmax ;
    //  cpl_array_set_double (nfcc, j, mfcc);
    //}

    //We normalize the CCF to its maximum value
    double bfccmax = cpl_array_get_max (fcc);
    for (int j = 0; j < (int)wpt; j++) {
      double cfcc = cpl_array_get_double (fcc, j, NULL);
      double mfcc = cfcc / bfccmax ;
      cpl_array_set_double (nfcc, j, mfcc);
    }

    double bfccmin = cpl_array_get_min (nfcc);
    //Condition to select the fitting function
    if (fitg == 1) {

    //Fit the CCF with a gaussian + 2nd order polynomia
    cpl_ensure_code(espda_ccf_fit_gauss(xcc, nfcc, yfit, bfit,
	        		pfit, efit, bfccmin, rvinit)
                    == CPL_ERROR_NONE, cpl_error_get_code());

    } else {
    //Fit the CCF with a lorentzian + 2nd order polynomia
    cpl_ensure_code(espda_ccf_fit_lorntz(xcc, nfcc, yfit, bfit,
    		        pfit, efit, bfccmin, rvinit)
                    == CPL_ERROR_NONE, cpl_error_get_code());
    }

    //Fit the CCF with a parabola to the CCF core
    cpl_size irv = 0;
    cpl_array_get_maxpos (nfcc, &irv);
	int npc = 7;
	int npcp = (int)npc*2 + 1;
	int npr = 3;

	cpl_array *nfcore = cpl_array_new (npcp, CPL_TYPE_DOUBLE);
	cpl_array *xfcore = cpl_array_new (npcp, CPL_TYPE_DOUBLE);
	cpl_array *yfcore = cpl_array_new (npcp, CPL_TYPE_DOUBLE);

	cpl_array *pfcore = cpl_array_new (npr, CPL_TYPE_DOUBLE);
    cpl_array *efcore = cpl_array_new (npr, CPL_TYPE_DOUBLE);

    int j = 0 ;
	for (int i = 0; i < (int)wpt; i++) {
        if (i > (int)irv-npc-1 && i < (int)irv+npc+1 ) {
	       double cfc  = cpl_array_get_double (nfcc, i, NULL);
           double cxc = cpl_array_get_double (xcc, i, NULL);
           cpl_array_set_double (nfcore, j, cfc);
           cpl_array_set_double (xfcore, j, cxc);
           cpl_array_set_double (yfcore, j, cfc);
           j = j + 1 ;
           //printf("core index : %i %i %i %i %i \n", i, j, npc, npcp, (int)irv);
        }
    }

	cpl_ensure_code(espda_ccf_fit_par(xfcore, nfcore, yfcore,
    		        pfcore, efcore, rvinit)
                    == CPL_ERROR_NONE, cpl_error_get_code());

    // Printing the fitting parameters

    //double ca0 =  cpl_array_get_double (pfit, 0, NULL);
    //double ca1 =  cpl_array_get_double (pfit, 1, NULL);
    //double ca2 =  cpl_array_get_double (pfit, 2, NULL);
    //double ca3 =  cpl_array_get_double (pfit, 3, NULL);
    //double ca4 =  cpl_array_get_double (pfit, 4, NULL);
    //double ca5 =  cpl_array_get_double (pfit, 5, NULL);
    //printf("CCF fit : %f %f %f %f %f %f \n", ca0, ca1, ca2, ca3, ca4, ca5);

    // Correcting the observed wavelength array for the measured RV

    double crv =  cpl_array_get_double (pfit, 1, NULL);
    double crvcore =  cpl_array_get_double (pfcore, 1, NULL);
    double ecrv =  cpl_array_get_double (efit, 1, NULL);
    double ecrvcore =  cpl_array_get_double (efcore, 1, NULL);

    //printf("RV eRV RV_CORE eRV_CORE: %f %f %f %f \n", crv, crvcore, ecrv, ecrvcore);

    for (int j = 0 ; j < (int)wpt; j++) {
      double cx =  cpl_array_get_double (xcc, j, NULL);
      double cy =  cpl_array_get_double (fcc, j, NULL);
      double ncy =  cpl_array_get_double (nfcc, j, NULL);
      double cf =  cpl_array_get_double (yfit, j, NULL);
      double bf =  cpl_array_get_double (bfit, j, NULL);
      //printf("CCF : %f %f %f %f %f \n", cx, cy, ncy, cf, bf);
    }

    for (int j = 0 ; j < (int)npcp; j++) {
      double cxc =  cpl_array_get_double (xfcore, j, NULL);
      double cyc =  cpl_array_get_double (nfcore, j, NULL);
      double cfc =  cpl_array_get_double (yfcore, j, NULL);
      //printf("CCF CORE : %f %f %f \n", cxc, cyc, cfc);
    }

    //Condition for wrong CCF fit
    if (crv >= vwccf | crv <= -vwccf) {
    	double cpf =  -9999. ;
    	double epf =  9999. ;
        cpl_array_set_double (pfit, 1, cpf);
        cpl_array_set_double (efit, 1, epf);
        double cy =  cpl_array_get_min (fcc) ;
        double cb =  0. ;
        //printf("CCF parfit: %f %f \n", cy, cb);
        for (int j = 0 ; j < (int)wpt; j++) {
            cpl_array_set_double (yfit, j, cy);
            cpl_array_set_double (bfit, j, cb);
        }
    }

    //Condition to choose the RV derived by fitting the CCF or the CCF core
    if (frv == 1) {
      //To use the RV from fitting the complete CCF
    	crv =  cpl_array_get_double (pfit, 1, NULL);
    } else {
        //To use the RV from fitting the CCF core
    	crv =  cpl_array_get_double (pfcore, 1, NULL);
    }

    //Saving the parameters of the CCF fit
    for (int j = 0 ; j < (int)npar; j++) {
      double cpf =  cpl_array_get_double (pfit, j, NULL);
      double ecpf =  cpl_array_get_double (efit, j, NULL);
      cpl_array_set_double (pccf, j, cpf);
      cpl_array_set_double (eccf, j, ecpf);
    }

    //Saving the parameters of the CCF fit core
    for (int j = 0 ; j < (int)npr; j++) {
      double cpfc =  cpl_array_get_double (pfcore, j, NULL);
      double ecpfc =  cpl_array_get_double (efcore, j, NULL);
      cpl_array_set_double (pccfc, j, cpfc);
      cpl_array_set_double (eccfc, j, ecpfc);
    }

    *xccf = cpl_array_new(wpt, CPL_TYPE_DOUBLE);
    *yccf = cpl_array_new(wpt, CPL_TYPE_DOUBLE);
    *ygfit = cpl_array_new(wpt, CPL_TYPE_DOUBLE);
    //*xccfcore = cpl_array_new(wpt, CPL_TYPE_DOUBLE);
    //*yccfcore = cpl_array_new(wpt, CPL_TYPE_DOUBLE);
    *ygfcore = cpl_array_new(wpt, CPL_TYPE_DOUBLE);

    int k = 0;
    for (int j = 0 ; j < (int)wpt; j++) {
       //We save the values of the ccf fit
      double cx =  cpl_array_get_double (xcc, j, NULL);
      double cy =  cpl_array_get_double (nfcc, j, NULL);
      double cf =  cpl_array_get_double (yfit, j, NULL);
      double bf =  cpl_array_get_double (bfit, j, NULL);
      // printf("CCF : %f %f %f %f \n", cx, cy, cf, bf);
      cy = cy - bf ;
      cf = cf - bf ;
      cpl_array_set_double (*xccf, j, cx);
      cpl_array_set_double (*yccf, j, cy);
      cpl_array_set_double (*ygfit, j, cf);
      //We save the values of the core fit
      double yfc = 0. ;
      cpl_array_set_double (*ygfcore, j, yfc);
      if (j > (int)irv-npc-1 && j < (int)irv+npc+1 ) {
	      yfc  = cpl_array_get_double (yfcore, k, NULL);
	      yfc = yfc - bf ;
          cpl_array_set_double (*ygfcore, j, yfc);
          k = k + 1 ;
          printf("CCF CORE FIT : %i %i %f \n", j, k, yfc);
      }
    }
    
    double vshift = -crv ;
    for (int j = 0 ; j < (int)nobs; j++) {
      double cwav = cpl_array_get_double (wobs, j, NULL);
      double cwcr = cwav * ( sqrt (( 1. + (vshift / vlight)) / 
                                   ( 1. - (vshift / vlight)) )) ;
      cpl_array_set_double (wcor, j, cwcr);
    }

    // We interpolate the flux array of the synthetic spectrum
    // in the new RV-corrected wavelength array of the observed spectrum

    //The interpolation function for a given dataset is stored in a gsl_interp object.
    //These are created by the following functions.

    //Function: gsl_interp * gsl_interp_alloc (const gsl_interp_type * T, size_t size)
    //This function returns a pointer to a newly allocated interpolation object
    //of type T for size data-points.

    //Function: int gsl_interp_init (gsl_interp * interp, const double xa[],
    //const double ya[], size_t size)
    //This function initializes the interpolation object interp for the data
    //(xa,ya) where xa and ya are arrays of size size. The interpolation object
    //(gsl_interp) does not save the data arrays xa and ya and only stores the
    //static state computed from the data. The xa data array is always
	//assumed to be strictly ordered, with increasing x values; the behavior
    //for other arrangements is not defined.

	//Function: void gsl_interp_free (gsl_interp * interp)
	//This function frees the interpolation object interp.

    //Interpolation Type: gsl_interp_cspline
    //Cubic spline with natural boundary conditions.
    //The resulting curve is piecewise cubic on each interval,
    //with matching first and second derivatives at the supplied
    //data-points. The second derivative is chosen to be zero at
    //the first point and last point.

    //We define the boundary of the arrays to compute the residuals
    double vpn = vshift / vdw ;
    int nvp = 0 ;
    if (vpn >= 0) {
      nvp = (int)vpn + 1 ;
    } else {
      nvp = -(int)vpn + 1 ;
    }
    double dvpn = nvp * 2.0 ;
    int invp = (int)dvpn ;
    cpl_size nos = (int)nobs - (int)invp  ;

    //printf(" RV [km/s] : %f \n", vshift);
    //printf(" Discarded number of boundary points in array of residuals : %f \n", vpn);
    //printf(" Discarded number of boundary points in array of residuals : %i \n", (int)dvpn);
    //printf(" Nobs : %i \n", (int)nobs);
    //printf(" Nos : %i \n", (int)nos);

    //printf("Initiate the interpolation \n");

    cpl_array *fsyc = cpl_array_duplicate (fsyn);

    gsl_interp_accel *acc = gsl_interp_accel_alloc ();
    gsl_spline *spline = gsl_spline_alloc (gsl_interp_cspline, nobs);

    double xsyn[nobs] ;
    double ysyn[nobs] ;
    for (int j = 0; j < (int)nobs; j++) {
    //   int k = (int)nvp + j ;
       double cxs = cpl_array_get_double (wobs, j, NULL);
       double cys = cpl_array_get_double (fsyn, j, NULL);
       xsyn[j] = cxs ;
       ysyn[j] = cys ;
    }

    //printf("Initiate the interpolation 2 \n");
    //double xs_max= cpl_array_get_max (wobs);
    //double xs_min= cpl_array_get_min (wobs);
    //double xo_max= cpl_array_get_max (wcor);
    //double xo_min= cpl_array_get_min (wcor);
    //printf("xsyn_max = %f xsyn_min= %f \n",xs_max,xs_min);
    //printf("xobs_max = %f xobs_min= %f \n",xo_max,xo_min);

    //for (int j = 0; j < 10; j++) {
    //    double cwcor = cpl_array_get_double (wcor, j, NULL);
    //      printf("wcor [i] = %f \n",cwcor);
    //    double cwobs = cpl_array_get_double (wobs, j, NULL);
    //      printf("wobs [i] = %f \n",cwobs);
    //    printf("xsyn [i] = %f \n",xsyn[j]);
    //  }
    //  printf("xsyn [0] = %f %i \n",xsyn[0], 0);
    //  for (int j = (int)nos-10; j < (int)nos; j++) {
    //  printf("xsyn [nos] = %f %i\n",xsyn[j], j);
    //  }

    gsl_spline_init (spline, xsyn, ysyn, nobs);

    //printf("Ok spline interp \n");

    //Creating the new array of synthetic spectrum
    for (int j = 0; j < (int)nos; j++) {
       int k = (int)nvp + j ;
       double cwcor = cpl_array_get_double (wcor, k, NULL);
       double cfcor = gsl_spline_eval (spline, cwcor, acc);
       cpl_array_set_double (fsyc, k, cfcor);
    }
    //printf("Finalize the interpolation \n");

    gsl_spline_free (spline);
    gsl_interp_accel_free (acc);

    //Filling with flux=1 for boundary points
    double cfcor = 1 ;
    for (int j = 0; j < (int)nvp; j++) {
       int k = (int)nos + j ;
       cpl_array_set_double (fsyc, j, cfcor);
       cpl_array_set_double (fsyc, k, cfcor);
    }

    //We compute the flux residuals between the observed-synthetic spectra
    for (int j = 0; j < (int)nobs; j++) {
       double cfobs = cpl_array_get_double (fobs, j, NULL);
       cpl_array_set_double (fres, j, cfobs);
    }

    cpl_array_subtract (fres, fsyc);

    // Closing the function

    if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to perform the cross-correlation");
    }
    //Deleting arrays

    //gsl_vector_free (xsyn);
    //gsl_vector_free (ysyn);

    cpl_array_delete(fsyc);
    cpl_array_delete(fos);
    cpl_array_delete(fcc);
    cpl_array_delete(xcc);
    cpl_array_delete(yfit);
    cpl_array_delete(bfit);
    cpl_array_delete(pfit);
    cpl_array_delete(efit);
    cpl_array_delete(nfcc);
    cpl_array_delete(pfcore);
    cpl_array_delete(efcore);
    cpl_array_delete(nfcore);
    cpl_array_delete(xfcore);
    cpl_array_delete(yfcore);

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Fit a gaussian+polynomial to the cross-correlation function
  @param    xarr    x-axis array of the CCF
  @param    yarr    y-axis array of the CCF
  @param    yfit    Array of the gaussian fitting function
  @param    bfit    Array of the background 2nd-order polynomial fitting function
  @param    pfit    Array of the fitting parameters
  @param    efit    Array of the error on fitting parameters
  @param    rvi     Initial guess of the RV value in km/s
  @return   CPL_ERROR_NONE  iff OK
        
  @doc      
  This function fit a gaussian to a function defined by two arrays plus a polynomial 
  that can be of order 0, 1 and 2. The fitting function array is given by yfit as 
  output as well as the parameters.
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_ccf_fit_gauss(cpl_array     *xarr,
 			                       cpl_array     *yarr,
 			                       cpl_array     *yfit,
 			                       cpl_array     *bfit,
 			                       cpl_array     *pfit,
 			                       cpl_array     *efit,
				                   double         ymin,
								   double          rvi)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (xarr);
    cpl_size   ncoef = cpl_array_get_size (pfit);
    int        npt = (int)npoint ;

    // defining input cpl arrays as just C++ arrays and 
    // extract the values from cpl arrays

    double xa[npt] ;
    double ya[npt] ;
    double sa[npt] ;
    for (int j = 0 ; j < (int)npt; j++) {
      double cxa = cpl_array_get_double (xarr, j, NULL);
      double cya = cpl_array_get_double (yarr, j, NULL);
        xa[j] = cxa ;
        ya[j] = cya ;
        sa[j] = cya * 0.001 ;
    }

    // Performing single gaussian + polynomial fit
 
    int npar = (int)ncoef ;
    int ig = 0 ;
    double gcf[npar] ;
    double gcf_er[npar] ;
    //Gauss == f(x)=(1/(sigma*SQRT(2*pi)))*exp[-(x-x0)^2/(2*sigma^2)]
    //FWHM = 2 * SQRT(2*log(2)) * sigma = 2.3548 * sigma
    //y=A0*exp(-z^2./2.)+A3+A4*x+A5*x^2. with z=(x-A1)/A2
    gcf[ig] = 1. - ymin ;
    gcf[ig+1] = rvi ;
    gcf[ig+2] = 1. ;
    //gcf[ig+3] = 0. ;
    gcf[ig+3] = ymin ;
    gcf[ig+4] = 0. ;
    gcf[ig+5] = 0. ;

    int stat2;

    gaussfit(xa,ya,sa,npt,gcf,gcf_er,npar,&stat2);

    // Saving the result of single gaussian + polynomial fit into cpl arrays

    for (int i = 0; i < npt; i++)
     {
       double ti = xa[i] ;
       double yf = gcf[0] * exp (- ( (ti-gcf[1])*(ti-gcf[1]) ) /
				    ( 2.*gcf[2]*gcf[2]) ) +
                     gcf[3] + gcf[4] * ti + gcf[5] * ti * ti ;
       double bf = gcf[3] + gcf[4] * ti + gcf[5] * ti * ti ;
       cpl_array_set_double (yfit, i, yf);
       cpl_array_set_double (bfit, i, bf);
     }

    for (int i = 0; i < npar; i++)
     {
       double cf = gcf[i] ;
       cpl_array_set_double (pfit, i, cf);
       double ef = gcf_er[i] ;
       cpl_array_set_double (efit, i, ef);
     }

    // We release the memory by deleting the remaining arrays
    //      cpl_array_delete(ctx_out);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to fit the cross-correlation function...");
    }

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Fit a lorentzian+polynomial to the cross-correlation function
  @param    xarr    x-axis array of the CCF
  @param    yarr    y-axis array of the CCF
  @param    yfit    Array of the lorentzian fitting function
  @param    bfit    Array of the background 2nd-order polynomial fitting function
  @param    pfit    Array of the fitting parameters
  @param    efit    Array of the error on fitting parameters
  @param    rvi     Initial guess of the RV value in km/s
  @return   CPL_ERROR_NONE  iff OK

  @doc
  This function fit a lorentzian to a function defined by two arrays plus a polynomial
  that can be of order 0, 1 and 2. The fitting function array is given by yfit as
  output as well as the parameters.
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_ccf_fit_lorntz(cpl_array     *xarr,
 			                        cpl_array     *yarr,
 			                        cpl_array     *yfit,
 			                        cpl_array     *bfit,
 			                        cpl_array     *pfit,
 			                        cpl_array     *efit,
					                double         ymin,
				                    double          rvi)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (xarr);
    cpl_size   ncoef = cpl_array_get_size (pfit);
    int        npt = (int)npoint ;

    // defining input cpl arrays as just C++ arrays and
    // extract the values from cpl arrays

    double xa[npt] ;
    double ya[npt] ;
    double sa[npt] ;
    for (int j = 0 ; j < (int)npt; j++) {
      double cxa = cpl_array_get_double (xarr, j, NULL);
      double cya = cpl_array_get_double (yarr, j, NULL);
        xa[j] = cxa ;
        ya[j] = cya ;
        sa[j] = cya * 0.001 ;
    }

    // Performing single lorentzian + polynomial fit

    int npar = (int)ncoef ;
    int ig = 0 ;
    double gcf[npar] ;
    double gcf_er[npar] ;
    //Lorentz == f(x)=((1/pi)*(1/2)*gamma)*(1/((x-x0)^2+(gamma/2)^2))
    //FWHM = gamma = 2 * SQRT(2*log(2)) * sigma = 2.3548 * sigma
    //y=A0 * (1. / ( (x-A1)^2.+( (2.*SQRT(2.*log(2.))*A2)/2.)^2.) ) +A3+A4*x+A5*x^2.
    gcf[ig] = 1. - ymin ;
    gcf[ig+1] = rvi ;
    gcf[ig+2] = 1. ;
    //gcf[ig+3] = 0. ;
    gcf[ig+3] = ymin ;
    gcf[ig+4] = 0. ;
    gcf[ig+5] = 0. ;

    int stat2;

    lorntzfit(xa,ya,sa,npt,gcf,gcf_er,npar,&stat2);

    // Saving the result of single lorentzian + polynomial fit into cpl arrays

    for (int i = 0; i < npt; i++)
     {
       double ti = xa[i] ;
       double yf = gcf[0] * (1. / ( ( (ti-gcf[1])*(ti-gcf[1]) ) +
		                         ( ( (CPL_MATH_FWHM_SIG * gcf[2]) *
		                   (CPL_MATH_FWHM_SIG * gcf[2]) ) / 4.) ) ) +
						   gcf[3] + gcf[4] * ti + gcf[5] * ti * ti ;
       double bf = gcf[3] + gcf[4] * ti + gcf[5] * ti * ti ;
       cpl_array_set_double (yfit, i, yf);
       cpl_array_set_double (bfit, i, bf);
     }

    for (int i = 0; i < npar; i++)
     {
       double cf = gcf[i] ;
       cpl_array_set_double (pfit, i, cf);
       double ef = gcf_er[i] ;
       cpl_array_set_double (efit, i, ef);
     }

    // We release the memory by deleting the remaining arrays
    //      cpl_array_delete(ctx_out);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to fit the cross-correlation function...");
    }

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Fit a parabola to core of the cross-correlation function
  @param    xarr    x-axis array of the CCF
  @param    yarr    y-axis array of the CCF
  @param    yfit    Array of the parabolic fitting function
  @param    pfit    Array of the fitting parameters
  @param    efit    Array of the error on fitting parameters
  @param    rvi     Initial guess of the RV value in km/s
  @return   CPL_ERROR_NONE  iff OK

  @doc
  This function fit a lorentzian to a function defined by two arrays plus a polynomial
  that can be of order 0, 1 and 2. The fitting function array is given by yfit as
  output as well as the parameters.
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_ccf_fit_par(cpl_array     *xarr,
 			                     cpl_array     *yarr,
 			                     cpl_array     *yfit,
 			                     cpl_array     *pfit,
 			                     cpl_array     *efit,
				                 double          rvi)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (xarr);
    cpl_size   ncoef = cpl_array_get_size (pfit);
    int        npt = (int)npoint ;

    // defining input cpl arrays as just C++ arrays and
    // extract the values from cpl arrays

    double xa[npt] ;
    double ya[npt] ;
    double sa[npt] ;
    for (int j = 0 ; j < (int)npt; j++) {
      double cxa = cpl_array_get_double (xarr, j, NULL);
      double cya = cpl_array_get_double (yarr, j, NULL);
        xa[j] = cxa ;
        ya[j] = cya ;
        sa[j] = cya * 0.001 ;
    }

    // Performing single lorentzian + polynomial fit

    int npar = (int)ncoef ;
    int ig = 0 ;
    double gcf[npar] ;
    double gcf_er[npar] ;
    //y=A0+((x-A1)^2.)*A2
    gcf[ig] = 1. ;
    gcf[ig+1] = rvi ;
    gcf[ig+2] = 1. ;

    int stat2;

    parfit(xa,ya,sa,npt,gcf,gcf_er,npar,&stat2);

    // Saving the result of single parabolic fit into cpl arrays

    for (int i = 0; i < npt; i++)
     {
       double ti = xa[i] ;
       double yf = gcf[0] + ( ( ti - gcf[1] ) * ( ti - gcf[1] ) * gcf[2] ) ;
       cpl_array_set_double (yfit, i, yf);
     }

    for (int i = 0; i < npar; i++)
     {
       double cf = gcf[i] ;
       cpl_array_set_double (pfit, i, cf);
       double ef = gcf_er[i] ;
       cpl_array_set_double (efit, i, ef);
     }

    // We release the memory by deleting the remaining arrays
    //      cpl_array_delete(ctx_out);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to fit the cross-correlation function...");
    }

    return CPL_ERROR_NONE;
}

//---------------------------------------------
// Gaussian fitting function
//---------------------------------------------

  //void gaussfit(double t[], double y[], double sig[], long npt,
  void gaussfit(double t[], double y[], double sig[], int npt,
               double gcf[], double gcf_er[], int npar, int *stat2)
  {
  const gsl_multifit_fdfsolver_type *T;
  gsl_multifit_fdfsolver *s;
  #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
  gsl_matrix *J;
  #endif
  int stat;
  //  unsigned int i, iter = 0;
  //size_t i, iter = 0;
     //  long N = npt;
     //  const size_t n = N;
  //const size_t n = npt;
  //const size_t p = npar;

  int i = 0;
  int iter = 0;
  int n = (int)npt;
  int p = (int)npar;

  gsl_matrix *covar = gsl_matrix_alloc (p, p);
  struct data d = {n, npar, t, y, sig};
  gsl_multifit_function_fdf f;

  double x_init[npar];
  for (int i=0; i< (int)npar; i++)
	x_init[i]=gcf[i];

  f.f = &gaussb_f;
  f.df = &gaussb_df;
  f.fdf = &gaussb_fdf;
  f.n = n;
  f.p = p;
  f.params = &d;

  gsl_vector_view x = gsl_vector_view_array (x_init, p);

  T = gsl_multifit_fdfsolver_lmder;

  s = gsl_multifit_fdfsolver_alloc (T, n, p);

  /* initialize solver with starting point and weights */
     //  gsl_multifit_fdfsolver_wset (s, &f, &x.vector, &w.vector);
  gsl_multifit_fdfsolver_set (s, &f, &x.vector);

  //      printf(" Ok6 \n");
  //      print_state (iter, s);

  do
    {
      iter++;
      stat = gsl_multifit_fdfsolver_iterate (s);
      
      //      printf ("status = %s\n", gsl_strerror (stat));

      //      print_state (iter, s);

      if (stat)
        break;

      stat = gsl_multifit_test_delta (s->dx, s->x,
                                      1e-6, 1e-6);
      //                                1e-4, 1e-4);
    }
  while (stat == GSL_CONTINUE && iter < 5000);

  //  printf(" Iter %i \n", iter);
  #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
  J = gsl_matrix_alloc(n, p);
  gsl_multifit_fdfsolver_jac(s, J);
  gsl_multifit_covar (J, 0.0, covar);
  gsl_matrix_free (J);
  #else
  gsl_multifit_covar (s->J, 0.0, covar);
  #endif
  #define FIT(i) gsl_vector_get(s->x, i)
  #define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

  {

    double chi = gsl_blas_dnrm2(s->f);
    double dof = n - p;
    double c = GSL_MAX_DBL(1, chi / sqrt(dof));

    for (i=0; i < (int)npar; i++)
	{
	    gcf[i]=FIT(i);
        gcf_er[i]=c*ERR(i);
	}

   *stat2=stat;
   //  *stat2=0;   //sometimes we have bad fits but the result is perfectably acceptable
   }
   gsl_multifit_fdfsolver_free (s);
   gsl_matrix_free (covar);
  }

  //---------------------------------------------
  //---------------------------------------------
//  //void print_state (size_t iter, gsl_multifit_fdfsolver * s)
//  void print_state (int iter, gsl_multifit_fdfsolver * s)
//  {
//   printf ("Iter: %3u x = % 15.8f % 15.8f % 15.8f % 15.8f % 15.8f % 15.8f "
//  //  printf ("Iter: %3i x = % 15.8f % 15.8f % 15.8f % 15.8f % 15.8f % 15.8f "
//           "|f(x)| = %g\n",
//  //espda_synth_ccf.c:476:11: warning: format specifies type 'unsigned int'
//  //        but the argument has type 'size_t' (aka 'unsigned long') [-Wformat]
//  //          iter,
     
//          iter,
//          gsl_vector_get (s->x, 0),
//          gsl_vector_get (s->x, 1),
//          gsl_vector_get (s->x, 2),
//          gsl_vector_get (s->x, 3),
//          gsl_vector_get (s->x, 4),
//          gsl_vector_get (s->x, 5),
//          gsl_blas_dnrm2 (s->f));
//   }


 //---------------------------------------------
 // Lorentzian fitting function
 //---------------------------------------------

  //void lorntzfit(double t[], double y[], double sig[], long npt,
   void lorntzfit(double t[], double y[], double sig[], int npt,
                double gcf[], double gcf_er[], int npar, int *stat2)
   {
   const gsl_multifit_fdfsolver_type *T;
   gsl_multifit_fdfsolver *s;
   #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
   gsl_matrix *J;
   #endif
   int stat;
   //  unsigned int i, iter = 0;
   //size_t i, iter = 0;
      //  long N = npt;
      //  const size_t n = N;
   //const size_t n = npt;
   //const size_t p = npar;

   int i = 0;
   int iter = 0;
   int n = (int)npt;
   int p = (int)npar;

   gsl_matrix *covar = gsl_matrix_alloc (p, p);
   struct data d = {n, npar, t, y, sig};
   gsl_multifit_function_fdf f;

   double x_init[npar];
   for (int i=0; i< (int)npar; i++)
 	x_init[i]=gcf[i];

   f.f = &lorntzb_f;
   f.df = &lorntzb_df;
   f.fdf = &lorntzb_fdf;
   f.n = n;
   f.p = p;
   f.params = &d;

   gsl_vector_view x = gsl_vector_view_array (x_init, p);

   T = gsl_multifit_fdfsolver_lmder;

   s = gsl_multifit_fdfsolver_alloc (T, n, p);

   /* initialize solver with starting point and weights */
      //  gsl_multifit_fdfsolver_wset (s, &f, &x.vector, &w.vector);
   gsl_multifit_fdfsolver_set (s, &f, &x.vector);

   //      printf(" Ok6 \n");
   //      print_state (iter, s);

   do
     {
       iter++;
       stat = gsl_multifit_fdfsolver_iterate (s);

       //      printf ("status = %s\n", gsl_strerror (stat));

       //      print_state (iter, s);

       if (stat)
         break;

       stat = gsl_multifit_test_delta (s->dx, s->x,
                                       1e-6, 1e-6);
       //                                1e-4, 1e-4);
     }
   while (stat == GSL_CONTINUE && iter < 5000);

   //  printf(" Iter %i \n", iter);
   #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
   J = gsl_matrix_alloc(n, p);
   gsl_multifit_fdfsolver_jac(s, J);
   gsl_multifit_covar (J, 0.0, covar);
   gsl_matrix_free (J);
   #else
   gsl_multifit_covar (s->J, 0.0, covar);
   #endif
   #define FIT(i) gsl_vector_get(s->x, i)
   #define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

   {

     double chi = gsl_blas_dnrm2(s->f);
     double dof = n - p;
     double c = GSL_MAX_DBL(1, chi / sqrt(dof));

     for (i=0; i < (int)npar; i++)
 	{
 	    gcf[i]=FIT(i);
        gcf_er[i]=c*ERR(i);
 	}

    *stat2=stat;
    //  *stat2=0;   //sometimes we have bad fits but the result is perfectably acceptable
    }
    gsl_multifit_fdfsolver_free (s);
    gsl_matrix_free (covar);
   }


  //---------------------------------------------
  // Parabolic function
  //---------------------------------------------

    //void parfit(double t[], double y[], double sig[], long npt,
     void parfit(double t[], double y[], double sig[], int npt,
                  double gcf[], double gcf_er[], int npar, int *stat2)
     {
     const gsl_multifit_fdfsolver_type *T;
     gsl_multifit_fdfsolver *s;
     #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
     gsl_matrix *J;
     #endif
     int stat;
     //  unsigned int i, iter = 0;
     //size_t i, iter = 0;
        //  long N = npt;
        //  const size_t n = N;
     //const size_t n = npt;
     //const size_t p = npar;

     int i = 0;
     int iter = 0;
     int n = (int)npt;
     int p = (int)npar;

     gsl_matrix *covar = gsl_matrix_alloc (p, p);
     struct data d = {n, npar, t, y, sig};
     gsl_multifit_function_fdf f;

     double x_init[npar];
     for (int i=0; i< (int)npar; i++)
   	 x_init[i]=gcf[i];

     f.f = &parb_f;
     f.df = &parb_df;
     f.fdf = &parb_fdf;
     f.n = n;
     f.p = p;
     f.params = &d;

     gsl_vector_view x = gsl_vector_view_array (x_init, p);

     T = gsl_multifit_fdfsolver_lmder;

     s = gsl_multifit_fdfsolver_alloc (T, n, p);

     /* initialize solver with starting point and weights */
        //  gsl_multifit_fdfsolver_wset (s, &f, &x.vector, &w.vector);
     gsl_multifit_fdfsolver_set (s, &f, &x.vector);

     //      printf(" Ok6 \n");
     //      print_state (iter, s);

     do
       {
         iter++;
         stat = gsl_multifit_fdfsolver_iterate (s);

         //      printf ("status = %s\n", gsl_strerror (stat));

         //      print_state (iter, s);

         if (stat)
           break;

         stat = gsl_multifit_test_delta (s->dx, s->x,
                                         1e-6, 1e-6);
         //                                1e-4, 1e-4);
       }
     while (stat == GSL_CONTINUE && iter < 5000);

     //  printf(" Iter %i \n", iter);
     #if defined GSL_MAJOR_VERSION && GSL_MAJOR_VERSION >= 2
     J = gsl_matrix_alloc(n, p);
     gsl_multifit_fdfsolver_jac(s, J);
     gsl_multifit_covar (J, 0.0, covar);
     gsl_matrix_free (J);
     #else
     gsl_multifit_covar (s->J, 0.0, covar);
     #endif
     #define FIT(i) gsl_vector_get(s->x, i)
     #define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

     {

       double chi = gsl_blas_dnrm2(s->f);
       double dof = n - p;
       double c = GSL_MAX_DBL(1, chi / sqrt(dof));

       for (i=0; i < (int)npar; i++)
   	{
   	    gcf[i]=FIT(i);
          gcf_er[i]=c*ERR(i);
   	}

      *stat2=stat;
      //  *stat2=0;   //sometimes we have bad fits but the result is perfectably acceptable
      }
      gsl_multifit_fdfsolver_free (s);
      gsl_matrix_free (covar);
     }

/**@}*/
