/* $Id: espda_fit_starcont.c 247663 2019-11-26 12:10:14Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2019-11-26 12:10:14 +0000 (Tue, 26 Nov 2019) $
 * $Revision: 247663 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include "cpl_plot.h"

#include "espda_cont_ord.h"
#include "espda_util_cont.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"
#include "espda_param.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -------------------------------------------------------------------------y----*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_fit_starcont_create(cpl_plugin *);
static int espda_fit_starcont_exec(cpl_plugin *);
static int espda_fit_starcont_destroy(cpl_plugin *);
static int espda_fit_starcont(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_fit_starcont_description[] = 
"This recipe fits a (low-order) polynomial to a set of order spectra given \n"
"as a CPL FITS Table input file. The continuum polynomial is determine \n"
"by iteratively fitting the points defined by removing outliers using a \n"
"sigma clipping with a lower and upper threshold. Once the fitting function \n"
"procedure converges, the recipe generates a new normalized spectrum \n"
"dividing the observed flux by the fitting function, order by order, and \n"
"then merging all spectral orders.\n"
"\n"
"INPUT FRAMES                          FORMAT  # EXPECTED  TAGS          "
"NOTES: Relevant Table Columns:\n" 
"order spectrum                        SPEC    1           ORD_SPEC      "
"WAVEL, FLUX, FLUXERR\n"
"\n"
"OUTPUT FRAMES                         FORMAT  # PRODUCED  TAGS          "
"NOTES: Relevant Table Columns:\n"
"normalized order spectrum             SPEC    1           NORD_SPEC     "
"WAVEL, FLUX, FLUXERR, CONT, CONTERR, MASK\n"
"normalized merged rebinned spectrum   SPEC    1           NMERG_SPEC    "
"WAVEL, FLUX, FLUXERR\n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_fit_starcont", 
                        "Normalize to unity the continuum of a set "
                        "of stellar spectra", 
                        espda_fit_starcont_description, 
                        "Jonay Gonzalez Hernandez",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_fit_starcont_create, espda_fit_starcont_exec,
                        espda_fit_starcont_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_starcont_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */
    /* Parameters of espda_cont_ord */
    /* Parameter ord_lrejt */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_lrejt",
                                    CPL_TYPE_DOUBLE, "Multiplication factor "
                                    "for low rejection threshold: [0., 1.5]",
                                    "espda.espda_fit_starcont", 0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-lrejt");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_hrejt */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_hrejt",
                                    CPL_TYPE_DOUBLE, "Multiplication factor "
                                    "for high rejection threshold: [1.5, 3.0]",
                                    "espda.espda_fit_starcont", 2.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-hrejt");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_niter */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_niter",
                                    CPL_TYPE_INT, "Number of iterations "
                                    "of continuum fitting: [1, 10]",
                                    "espda.espda_fit_starcont", 5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-niter");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_forder1 */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_ford410",
                                    CPL_TYPE_INT, "Order of the fitting for  "
                                    "function (Wavelength < 410nm):  [0, 9]",
                                    "espda.espda_fit_starcont", 0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-ford410");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_forder2 */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_ford450",
                                    CPL_TYPE_INT, "Order of the fitting for  "
                                    "function (Wavelength < 450nm):  [0, 9]",
                                    "espda.espda_fit_starcont", 1);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-ford450");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_forder3 */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_ford510",
                                    CPL_TYPE_INT, "Order of the fitting for  "
                                    "function (Wavelength < 510nm):  [0, 9]",
                                    "espda.espda_fit_starcont", 2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-ford510");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ord_forder4 */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.ord_ford800",
                                    CPL_TYPE_INT, "Order of the fitting for  "
                                    "function (Wavelength < 800nm):  [0, 9]",
                                    "espda.espda_fit_starcont", 3);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ord-ford800");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

   /* Parameters of espda_spec_rebin */
   /* Parameter flux_kappa */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.flux_kappa",
                                    CPL_TYPE_INT, "Threshold for sigma-"
			                        "clipping for rebinning:  [3, 5]",
                                    "espda.espda_fit_starcont", 3);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "flux-kappa");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);


   /* Parameter flux_velstep */
    param = cpl_parameter_new_value("espda.espda_fit_starcont.flux_velstep",
				 CPL_TYPE_DOUBLE, "Step on velocity "
                                   "binning [km/s]: [0.3, 100.]",
			                       "espda.espda_fit_starcont", 0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "flux-velstep");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_starcont_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_fit_starcont(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_starcont_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_starcont(cpl_frameset            *frameset,
			   const cpl_parameterlist *paramlist)
{
    cpl_propertylist *proplist1   = NULL;
    cpl_propertylist *proplist2   = NULL;
    cpl_table        *ord_tab   = NULL;

    cpl_msg_info(cpl_func, "*** Starting Fit StarCont ***");

/* 
 * Frames are loaded and checked.
 */

    cpl_msg_info(cpl_func, "*** Loading and checking input FITS frames ***");
    cpl_ensure_code(espda_frame_import(frameset, FSPEC_ORIG, &ord_tab)
                    == CPL_ERROR_NONE, cpl_error_get_code());
    if (ord_tab == NULL) {
        return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, 
               "Wrong input frames or tag not set to FSPEC_ORIG");
    }

    cpl_msg_info(cpl_func, "*** 2D SPEC - ORDER SPECTRA table LOADED***");

   /* Getting parameters from the list into variables */
    espda_param_double("ord-lrejt",         lrejt,
        cpl_parameterlist_get_first_const(paramlist), 0.,  1.5);

    espda_param_double("ord-hrejt",     hrejt,
        cpl_parameterlist_get_next_const(paramlist),  1.5,  3.0); 

    espda_param_int("ord-niter",     niter,
        cpl_parameterlist_get_next_const(paramlist),  1,  10); 

    espda_param_int("ord-ford410",     ford410,
        cpl_parameterlist_get_next_const(paramlist),  0,  9);

    espda_param_int("ord-ford450",     ford450,
        cpl_parameterlist_get_next_const(paramlist),  0,  9);

    espda_param_int("ord-ford510",     ford510,
        cpl_parameterlist_get_next_const(paramlist),  0,  9);

    espda_param_int("ord-ford800",     ford800,
        cpl_parameterlist_get_next_const(paramlist),  0,  9);

    espda_param_int("flux-kappa",     kappa,
        cpl_parameterlist_get_next_const(paramlist),  3,  5); 

    espda_param_double("flux-velstep",     velstep,
        cpl_parameterlist_get_next_const(paramlist),  0.3,  1e2);
   
    /* DFS groups are set */
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());

    /* Fitting stellar continuum */
    cpl_msg_info(cpl_func, "*** Fitting the continuum of stellar spectra ***");

    /* Sorting the table ORD_SPEC by WAVELENGTH, then ORDID, then SPECID */
        cpl_propertylist *spec_sort   = NULL;
        spec_sort = cpl_propertylist_new();
        cpl_propertylist_append_bool(spec_sort, "WAVEL", FALSE);
        cpl_propertylist_append_bool(spec_sort, "ORDID", FALSE);
        cpl_propertylist_append_bool(spec_sort, "SPECID", FALSE);
        cpl_table_sort(ord_tab, spec_sort);

    /* Selecting the order to fit:
       First defining the numbers of orders and number of spectra */
    int norders = (int)cpl_table_get_column_max (ord_tab, "ORDID");
    int nspec = (int)cpl_table_get_column_max (ord_tab, "SPECID");
    printf("Number Orders and Spectra: %i and %i \n", norders, nspec);

    /* defining an array to save all values of the wavelength of all orders */
    cpl_array *wave_fin = cpl_array_new (cpl_table_get_nrow(ord_tab),
                                         CPL_TYPE_DOUBLE);

    /* defining an array to save all values of the rescaled flux of all orders */
    cpl_array *flux_fin = cpl_array_new (cpl_table_get_nrow(ord_tab),
                                         CPL_TYPE_DOUBLE);
    cpl_array *flux_ferr = cpl_array_new (cpl_table_get_nrow(ord_tab),
                                         CPL_TYPE_DOUBLE);

    /* defining an array to save all values of the continuum of all orders */
    cpl_array *cont_fin = cpl_array_new (cpl_table_get_nrow(ord_tab), 
                                         CPL_TYPE_DOUBLE);
    cpl_array *cont_ferr = cpl_array_new (cpl_table_get_nrow(ord_tab), 
                                         CPL_TYPE_DOUBLE);

    /* defining an array to save all values of the mask of all orders */
    cpl_array *mask_fin = cpl_array_new (cpl_table_get_nrow(ord_tab), 
                                         CPL_TYPE_INT);

    /* defining an array to save all values of the wavelength of all orders */
    cpl_array *spec_fin = cpl_array_new (cpl_table_get_nrow(ord_tab),
                                         CPL_TYPE_INT);
    cpl_array *ord_fin = cpl_array_new (cpl_table_get_nrow(ord_tab),
                                         CPL_TYPE_INT);

    /* Size of the number of points */
    int ford = (int)ford800 ;
    cpl_size nrows = 0; 

    /* Starting the loop of all orders */
    int nord= 0;
    int nspc= 0;
    for (int i = 0; i < norders; i++) {
        for (int j = 0; j < nspec; j++) {

    /* extracting a given order of the ORD_SPEC table */
    nord = i+1;
    nspc = j+1;
    cpl_table_unselect_all (ord_tab);
    cpl_table_or_selected_int (ord_tab, "ORDID", CPL_EQUAL_TO, i+1);
    cpl_table_and_selected_int (ord_tab, "SPECID", CPL_EQUAL_TO, j+1);
    cpl_table *order_tab = cpl_table_extract_selected (ord_tab);

    const double *wavep = cpl_table_get_data_double_const (order_tab, "WAVEL");
    cpl_array *wave_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (wave_arr, wavep);

    const double *fluxp = cpl_table_get_data_double_const (order_tab, "FLUX");
    cpl_array *flux_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (flux_arr, fluxp);

    const double *efluxp = cpl_table_get_data_double_const (order_tab, "FLUXERR");
    cpl_array *flux_err = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (flux_err, efluxp);

    const int *specp = cpl_table_get_data_int_const (order_tab, "SPECID");
    cpl_array *spec_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_INT);
    cpl_array_copy_data_int (spec_arr, specp);

    const int *ordp = cpl_table_get_data_int_const (order_tab, "ORDID");
    cpl_array *ord_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_INT);
    cpl_array_copy_data_int (ord_arr, ordp);

    /* Calling the function to perform the fit of the stellar continuum */
    cpl_array *cont_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_DOUBLE);
    cpl_array *cont_err = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_DOUBLE);
    cpl_array *mask_arr = cpl_array_new(cpl_table_get_nrow(order_tab), CPL_TYPE_INT);

    /* Initiating the order fit */
    cpl_size nlength = cpl_array_get_size (wave_arr);

    //printf("Ord: %i FORDER: %i %i %i %i FORD: %i \n", i, ford410, ford450, ford510, ford800, ford);

    int nford = 5;
    cpl_array *xford = cpl_array_new(nford, CPL_TYPE_INT);
    cpl_array_set_int(xford, 0, ford410);
    cpl_array_set_int(xford, 1, ford450);
    cpl_array_set_int(xford, 2, ford510);
    cpl_array_set_int(xford, 3, ford800);
    cpl_array_set_int(xford, 4, ford);

    //printf("FORDER: %i %i %i %i FORD: %i Ord: %i Spec: %i \n", ford410, ford450, ford510, ford800, ford, i, j);

    cpl_ensure_code(espda_cont_ford(wave_arr,
    		            xford) == CPL_ERROR_NONE, cpl_error_get_code());

    ford = cpl_array_get_int(xford, 4, NULL);
    double cwavea = cpl_array_get_double (wave_arr, 0, NULL);
    double cwaveb = cpl_array_get_double (wave_arr, (int)nlength-1, NULL);
    //printf("FORDER: %i FORD: %i Ord: %i Spec: %i Wmin: %f Wmax: %f \n", ford410, ford, i, j, cwavea, cwaveb);

    //cpl_ensure_code(espda_cont_ord(wave_arr, flux_arr, flux_err, cont_arr,
    //		            cont_err, mask_arr, lrejt, hrejt, niter, nord, nspc,
	//                    ford) == CPL_ERROR_NONE, cpl_error_get_code());

    cpl_ensure_code(espda_cont_ord(wave_arr, flux_arr, cont_arr,
    		            cont_err, mask_arr, lrejt, hrejt, niter,
	                    ford) == CPL_ERROR_NONE, cpl_error_get_code());

    /* Save the output wave into an array, by appending order by order */

    cpl_array_insert (wave_fin, wave_arr, nrows);

    /* Save the output flux and error into an array, by appending order by order */
    //cpl_size nlength = cpl_array_get_size (flux_arr);

    cpl_array_insert (flux_fin, flux_arr, nrows);
    cpl_array_insert (flux_ferr, flux_err, nrows);

    /* Save the output continuum fit into an array, by appending order by order */
    // cpl_size nlength = cpl_array_get_size (cont_arr);

    cpl_array_insert (cont_fin, cont_arr, nrows);
    cpl_array_insert (cont_ferr, cont_err, nrows);

    /* Save the output mask into an array, by appending order by order */
    cpl_array_insert (mask_fin, mask_arr, nrows);

    /* Save the output mask into an array, by appending order by order */
    cpl_array_insert (spec_fin, spec_arr, nrows);
    cpl_array_insert (ord_fin, ord_arr, nrows);

    /* We delete all the arrays and tables to release and clean the memory */
    cpl_array_delete(xford);
    cpl_array_delete(wave_arr);
    cpl_array_delete(flux_arr);
    cpl_array_delete(flux_err);
    cpl_array_delete(cont_arr);
    cpl_array_delete(cont_err);
    cpl_array_delete(mask_arr);
    cpl_array_delete(spec_arr);
    cpl_array_delete(ord_arr);

    cpl_table_delete(order_tab);

    /* Accounting for the number of points of this order when appending arrays */

    nrows = nrows + nlength ;

    /* Closing the loop of all spec */

    }

    /* Closing the loop of all orders */

    }

    /* Creating new CONT and CONTERR columns */
    cpl_table_new_column(ord_tab, "CONT", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ord_tab, "CONTERR", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ord_tab, "NFLUX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ord_tab, "NFLUXERR", CPL_TYPE_DOUBLE);
    //cpl_table_new_column(ord_tab, "SNR", CPL_TYPE_DOUBLE);
    //cpl_table_new_column(ord_tab, "NSNR", CPL_TYPE_DOUBLE);

    /* Copy the output contiuum data into a table DOUBLE column labelled as "WAVE" */
    const double *wavef = cpl_array_get_data_double_const (wave_fin);
    cpl_table_copy_data_double (ord_tab, "WAVEL", wavef);

    /* Copy the output contiuum data into a table DOUBLE column labelled as "FLUX" */
    const double *fluxf = cpl_array_get_data_double_const (flux_fin);
    cpl_table_copy_data_double (ord_tab, "FLUX", fluxf);
    cpl_table_copy_data_double (ord_tab, "NFLUX", fluxf);
    //cpl_table_copy_data_double (ord_tab, "SNR", fluxf);

    const double *efluxf = cpl_array_get_data_double_const (flux_ferr);
    cpl_table_copy_data_double (ord_tab, "FLUXERR", efluxf);
    cpl_table_copy_data_double (ord_tab, "NFLUXERR", efluxf);

    /* Copy the output continuum data into a table DOUBLE column labelled as "CONT" */
    const double *contf = cpl_array_get_data_double_const (cont_fin);
    cpl_table_copy_data_double (ord_tab, "CONT", contf);

    const double *econtf = cpl_array_get_data_double_const (cont_ferr);
    cpl_table_copy_data_double (ord_tab, "CONTERR", econtf);

    /* Copy the output mask data into a table INT column labelled as "MASK" */
    const int *maskf = cpl_array_get_data_int_const (mask_fin);
    cpl_table_copy_data_int (ord_tab, "MASK", maskf);

    /* Copy the output SPECID data into a table INT column labelled as "SPECID" */
    const int *specf = cpl_array_get_data_int_const (spec_fin);
    cpl_table_copy_data_int (ord_tab, "SPECID", specf);

    /* Copy the output ORDID data into a table INT column labelled as "ORDID" */
    const int *ordf = cpl_array_get_data_int_const (ord_fin);
    cpl_table_copy_data_int (ord_tab, "ORDID", ordf);

    /* We delete the remaining arrays to release and clean the memory */
    cpl_array_delete(wave_fin);
    cpl_array_delete(flux_fin);
    cpl_array_delete(flux_ferr);
    cpl_array_delete(cont_fin);
    cpl_array_delete(cont_ferr);
    cpl_array_delete(mask_fin);
    cpl_array_delete(spec_fin);
    cpl_array_delete(ord_fin);

    /* Sorting again the table ORD_SPEC by WAVELENGTH */
    cpl_propertylist *spec_wave   = NULL;
    spec_wave = cpl_propertylist_new();
    cpl_propertylist_append_bool(spec_wave, "WAVEL", FALSE);
    cpl_table_sort(ord_tab, spec_wave);

    /* We divide the flux in column "NFLUX" by the fit in column "CONT" within "ord_tab" */
    cpl_table_divide_columns (ord_tab, "NFLUX", "CONT");
    /* We divide the flux in column "NFLUXERR" by the fit in column "CONT" within "ord_tab" */
    cpl_table_divide_columns (ord_tab, "NFLUXERR", "CONT");
    /* We divide the flux in column "SNR" by the fit in column "FLUXERR" within "ord_tab" */
    //cpl_table_divide_columns (ord_tab, "SNR", "FLUXERR");
    /* We divide the flux in column "NSNR" by the fit in column "FLUXERR" within "ord_tab" */
    //cpl_table_duplicate_column (ord_tab, "NSNR", ord_tab, "NFLUX");
    //cpl_table_divide_columns (ord_tab, "NSNR", "NFLUXERR");

    /* We duplicate the table */
    cpl_table *nord_tab = cpl_table_duplicate (ord_tab);

    /* We divide the flux in column "FLUX" by the fit in column "CONT" within "nord_tab" */
    cpl_table_divide_columns (nord_tab, "FLUX", "CONT");
    /* We divide the flux in column "FLUXERR" by the fit in column "CONT" within "nord_tab" */
    cpl_table_divide_columns (nord_tab, "FLUXERR", "CONT");

// We perform the merging and rebinning of table "nord_tab"

    cpl_table *nmerg_tab = NULL;

    cpl_msg_info(cpl_func, "*** Rebinning spectrum ***");

    //double wav0 = 380.0 ;
    //double wav1 = 780.0 ;
    double wav0 = cpl_table_get_column_min(nord_tab, "WAVEL") ;
    double wav1 = cpl_table_get_column_max(nord_tab, "WAVEL") ;
    //cpl_ensure_code(espda_spec_rebin(nord_tab, kappa, wav0,
    //                            wav1, velstep, &nmerg_tab)
    //                        == CPL_ERROR_NONE, cpl_error_get_code());
    nmerg_tab = espda_spec_rebin(nord_tab, kappa, wav0, wav1, velstep);

    /* Erasing columns in the new tables */

      //    cpl_table_erase_column(ord_tab, "EQFACT");
      //    cpl_table_erase_column(ord_tab, "PIXSIZE");

    /* cpl_table_dump_structure(nmerg_tab, stdout); 
    exit(0); */

    cpl_table_erase_column(nmerg_tab, "CONT"); 
    cpl_table_erase_column(nmerg_tab, "NORM"); 
    cpl_table_erase_column(nmerg_tab, "NORMERR"); 
    cpl_table_erase_column(nmerg_tab, "SPECID"); 
    cpl_table_erase_column(nmerg_tab, "ORDID"); 
    //cpl_table_erase_column(nmerg_tab, "PIXSIZE");
    cpl_table_erase_column(nmerg_tab, "EQFACT");
    cpl_table_erase_column(nmerg_tab, "FLUXRMS");
    //cpl_table_erase_column(nmerg_tab, "SNR");
    //cpl_table_erase_column(nmerg_tab, "NSNR");

    /* Creating new MERGED SNR columns */
    //cpl_table_new_column(nmerg_tab, "MSNR", CPL_TYPE_DOUBLE);
    /* We divide the flux in column "MSNR" by the fit in column "FLUXERR" within "nmerg_tab" */
    cpl_table_duplicate_column (nmerg_tab, "MSNR", nmerg_tab, "FLUX");
    cpl_table_divide_columns (nmerg_tab, "MSNR", "FLUXERR");

/*
 * Products are exported into frames. 
 */ 

    cpl_msg_info(cpl_func, "*** Export products into frames ***");
    cpl_table_save(nmerg_tab,NULL,NULL,"nmerg_tab.fits",CPL_IO_DEFAULT);
    cpl_table_save(ord_tab,NULL,NULL,"ord_tab.fits",CPL_IO_DEFAULT);
    cpl_table* ext_tab=NULL;
    cpl_table_and_selected_int(ord_tab,"MASK",CPL_EQUAL_TO,0);
    ext_tab = cpl_table_extract_selected(ord_tab);
    proplist1 = cpl_propertylist_new();
    double mnflx= cpl_table_get_column_mean(ext_tab, "NFLUX") ;
    double menflx= cpl_table_get_column_mean(ext_tab, "NFLUXERR") ;
    cpl_free(ext_tab);
    //double msnr= cpl_table_get_column_mean(ord_tab, "SNR") ;
    //double mnsnr= cpl_table_get_column_mean(ord_tab, "NSNR") ;
    
    cpl_propertylist_append_double(proplist1, "ESO DAS QC STARCONT MEAN NFLUX", mnflx);
    cpl_propertylist_append_double(proplist1, "ESO DAS QC STARCONT MEAN NFLUXERR", menflx);
    //cpl_propertylist_append_double(proplist1, "ESO DAS QC MEAN SNR", msnr);
    //cpl_propertylist_append_double(proplist1, "ESO DAS QC MEAN NSNR", mnsnr);
    if (ord_tab != NULL) {
        cpl_ensure_code(espda_frame_export(frameset, paramlist, ord_tab, 
                        proplist1, "espda_fit_starcont", NORD_SPEC, 
                        "NORD_SPEC.fits")  == CPL_ERROR_NONE,
                        cpl_error_get_code());
    }
    else {
        cpl_msg_warning(cpl_func, "No spectrum normalized! No frame exported.");
    }

    proplist2 = cpl_propertylist_new();
    double mmflx= cpl_table_get_column_mean(nmerg_tab, "FLUX") ;
    double memflx= cpl_table_get_column_mean(nmerg_tab, "FLUXERR") ;
    double mmsnr= cpl_table_get_column_mean(nmerg_tab, "MSNR") ;
    cpl_propertylist_append_double(proplist2, "ESO DAS QC STARCONT MEAN MFLUX", mmflx);
    cpl_propertylist_append_double(proplist2, "ESO DAS QC STARCONT MEAN MFLUXERR", memflx);
    cpl_propertylist_append_double(proplist2, "ESO DAS QC STARCONT MEAN MSNR", mmsnr);
    if (nmerg_tab != NULL) {
        cpl_ensure_code(espda_frame_export(frameset, paramlist, nmerg_tab, 
                        proplist2, "espda_fit_starcont", NMERG_SPEC, 
                        "NMERG_SPEC.fits") == CPL_ERROR_NONE, 
                        cpl_error_get_code());
    }
    else {
        cpl_msg_warning(cpl_func, "No spectrum merged! No frame exported.");
    }

/*    
*Finishing the recipe
*/

    /* We delete all the property lists to release and clean the memory */
    cpl_propertylist_delete(spec_wave);
    cpl_propertylist_delete(spec_sort);

    /* We delete all the tables to release and clean the memory */
    //cpl_table_delete(nord_tab);

    cpl_msg_info(cpl_func, "*** Successful Recipe Execution!! ***");

    return (int)cpl_error_get_code();

}

/**@}*/
/*--------------------------------------------------------------------------*/
